Package: CovEsts
Title: Nonparametric Estimators for Covariance Functions
Version: 1.1.0
Authors@R: 
    c(person(given = "Adam",
             family = "Bilchouris",
             role = c("cre", "aut"),
             email = "adam.bilchouris@gmail.com",
             comment = c(ORCID = "0009-0002-4649-0247")),
      person(given = "Andriy",
             family = "Olenko",
             role = "aut",
             email = "a.olenko@latrobe.edu.au",
             comment = c(ORCID = "0000-0002-0917-7000"))
     )
Description: Several nonparametric estimators of autocovariance functions. Procedures for constructing their confidence regions by using bootstrap techniques. Methods to correct autocovariance estimators and several tools for analysing and comparing them. Supplementary functions, including kernel computations and discrete cosine Fourier transforms. For more details see Bilchouris and Olenko (2025) <doi:10.17713/ajs.v54i1.1975>.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.2
Suggests: testthat (>= 3.0.0)
Imports: graphics, grDevices, parallel, stats, utils
Config/testthat/edition: 3
URL: https://github.com/AdamBilchouris/CovEsts
BugReports: https://github.com/AdamBilchouris/CovEsts/issues
NeedsCompilation: no
Packaged: 2026-04-19 09:55:40 UTC; Adam
Author: Adam Bilchouris [cre, aut] (ORCID:
    <https://orcid.org/0009-0002-4649-0247>),
  Andriy Olenko [aut] (ORCID: <https://orcid.org/0000-0002-0917-7000>)
Maintainer: Adam Bilchouris <adam.bilchouris@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-19 10:10:02 UTC
Built: R 4.6.0; ; 2026-04-25 17:54:56 UTC; unix

Package: CovEsts
Title: Nonparametric Estimators for Covariance Functions
Version: 1.1.0
Authors@R: 
    c(person(given = "Adam",
             family = "Bilchouris",
             role = c("cre", "aut"),
             email = "adam.bilchouris@gmail.com",
             comment = c(ORCID = "0009-0002-4649-0247")),
      person(given = "Andriy",
             family = "Olenko",
             role = "aut",
             email = "a.olenko@latrobe.edu.au",
             comment = c(ORCID = "0000-0002-0917-7000"))
     )
Description: Several nonparametric estimators of autocovariance functions. Procedures for constructing their confidence regions by using bootstrap techniques. Methods to correct autocovariance estimators and several tools for analysing and comparing them. Supplementary functions, including kernel computations and discrete cosine Fourier transforms. For more details see Bilchouris and Olenko (2025) <doi:10.17713/ajs.v54i1.1975>.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.2
Suggests: testthat (>= 3.0.0)
Imports: graphics, grDevices, parallel, stats, utils
Config/testthat/edition: 3
URL: https://github.com/AdamBilchouris/CovEsts
BugReports: https://github.com/AdamBilchouris/CovEsts/issues
NeedsCompilation: no
Packaged: 2026-04-19 09:55:40 UTC; Adam
Author: Adam Bilchouris [cre, aut] (ORCID:
    <https://orcid.org/0009-0002-4649-0247>),
  Andriy Olenko [aut] (ORCID: <https://orcid.org/0000-0002-0917-7000>)
Maintainer: Adam Bilchouris <adam.bilchouris@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-19 10:10:02 UTC
Built: R 4.5.2; ; 2026-04-20 23:21:08 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(area_between,CovEsts)
S3method(area_between,default)
S3method(as.double,CovEsts)
S3method(check_pd,CovEsts)
S3method(check_pd,default)
S3method(hilbert_schmidt,CovEsts)
S3method(hilbert_schmidt,default)
S3method(kernel_est,CovEsts)
S3method(kernel_est,default)
S3method(lines,BootEsts)
S3method(lines,CovEsts)
S3method(lines,VarioEsts)
S3method(make_pd,CovEsts)
S3method(make_pd,default)
S3method(max_distance,CovEsts)
S3method(max_distance,default)
S3method(mse,CovEsts)
S3method(mse,default)
S3method(nearest_pd,CovEsts)
S3method(nearest_pd,default)
S3method(normalise_acf,CovEsts)
S3method(normalise_acf,default)
S3method(plot,BootEsts)
S3method(plot,CovEsts)
S3method(plot,VarioEsts)
S3method(print,BootEsts)
S3method(print,CovEsts)
S3method(print,VarioEsts)
S3method(shrinking,CovEsts)
S3method(shrinking,default)
S3method(spectral_norm,CovEsts)
S3method(spectral_norm,default)
S3method(splines_est,CovEsts)
S3method(splines_est,default)
S3method(to_pacf,CovEsts)
S3method(to_pacf,default)
S3method(to_vario,CovEsts)
S3method(to_vario,default)
export(adjusted_est)
export(area_between)
export(block_bootstrap)
export(bootstrap_sample)
export(check_pd)
export(corrected_est)
export(dct_1d)
export(hilbert_schmidt)
export(idct_1d)
export(kernel)
export(kernel_ec)
export(kernel_est)
export(kernel_symm)
export(kernel_symm_ec)
export(make_pd)
export(max_distance)
export(mse)
export(nearest_pd)
export(normalise_acf)
export(shrinking)
export(spectral_norm)
export(splines_est)
export(standard_est)
export(starting_locs)
export(taper)
export(tapered_est)
export(to_pacf)
export(to_vario)
export(truncated_est)
export(window)
export(window_ec)
export(window_symm)
export(window_symm_ec)
importFrom(grDevices,rgb)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,polygon)
importFrom(parallel,clusterExport)
importFrom(parallel,clusterSetRNGStream)
importFrom(parallel,makePSOCKcluster)
importFrom(parallel,parLapply)
importFrom(parallel,stopCluster)
importFrom(stats,acf)
importFrom(stats,optim)
importFrom(stats,quantile)
importFrom(stats,runif)
importFrom(stats,toeplitz)
importFrom(utils,head)
importFrom(utils,tail)

# This function sets variables such use in parameters.
test_x <- c(1, 2, 3)
test_X <- c(1, 2, 3)
test_meanX <- mean(test_X)
test_xij <- Xij_mat(test_X, test_meanX)
test_rhoT1 <- rho_T1(test_x, test_meanX, 2, 0.01, test_xij)

.gaussian_kernel <- function(x, params = 1) {
  return(exp(-x^2 / params))
}

.exponential_kernel <- function(x, params = 1) {
  return(exp(-x / params))
}

.wave_kernel <- function(x, params = 1) {
  returnVal <- x
  returnVal_indices <- which(returnVal != 0 & returnVal != Inf)
  returnVal[returnVal == 0] <- 1
  returnVal[returnVal == Inf] <- 0
  returnVal[returnVal_indices] <- (params[1] / returnVal[returnVal_indices]) * sin(returnVal[returnVal_indices] / params[1])
  return(returnVal)
}

.rational_quadratic_kernel <- function(x, params = 1) {
  returnVal <- x
  returnVal_indices <- which(returnVal != Inf)
  returnVal[returnVal == Inf] <- 0
  returnVal[returnVal_indices] <- (params[1] / (returnVal[returnVal_indices]^2 + params[1]))
  return(returnVal)
}

.circular_kernel <- function(x, params = 1) {
  returnVal <- x
  returnVal_indices <- which(returnVal < params[1])
  returnVal[returnVal_indices] <- ((2 / pi) * acos(returnVal[returnVal_indices]  / params[1])) - ((2 / pi) * (returnVal[returnVal_indices]  / params[1]) * sqrt(1 - (returnVal[returnVal_indices]  / params[1])^2))
  returnVal[-returnVal_indices] <- 0
  return(returnVal)
}

.spherical_kernel <- function(x, params = 1) {
  returnVal <- x
  returnVal_indices <- which(returnVal < params[1])
  returnVal[returnVal_indices] <- 1 - ((3/2) * (returnVal[returnVal_indices] / params[1])) + ((1/2) * (returnVal[returnVal_indices] / params[1])^3)
  returnVal[-returnVal_indices] <- 0
  return(returnVal)
}

.matern_kernel <- function(x, params = c(1, 1)) {
  returnVal <- x
  returnVal_indices <- which(returnVal != 0 & returnVal != Inf)
  returnVal[returnVal == 0] <- 1
  returnVal[returnVal == Inf] <- 0
  returnVal[returnVal_indices] <- (((sqrt(2 * params[2]) * returnVal[returnVal_indices] / params[1])^params[2]) / (2^(params[2] - 1) * gamma(params[2]))) * besselK(sqrt(2 * params[2]) * returnVal[returnVal_indices] / params[1], params[2])
  return(returnVal)
}

.bessel_j_kernel <- function(x, params = c(1, 1, 1)) {
  returnVal <- x
  returnVal_indices <- which(returnVal != 0 & returnVal != Inf)
  returnVal[returnVal == 0] <- 1
  returnVal[returnVal == Inf] <- 0
  returnVal[returnVal_indices] <- (2^params[2]) * gamma(params[2] + 1) * (besselJ(returnVal[returnVal_indices] / params[1], params[2]) / ((returnVal[returnVal_indices] / params[1])^params[2]))
  return(returnVal)
}

.cauchy_kernel <- function(x, params = c(1, 1, 1)) {
  ((1 + (x / params[1])^params[2])^(-(params[3] / params[2])))
}

.validate_kernel_params <- function(name = c("gaussian", "exponential", "wave", "rational_quadratic", "spherical", "circular", "bessel_j", "matern", "cauchy"), params) {
  name <- match.arg(name)

  switch(name,
         gaussian = TRUE,
         exponential = TRUE,
         wave = TRUE,
         rational_quadratic = TRUE,
         spherical = TRUE,
         circular = TRUE,
         bessel_j = ifelse(params[3] >= 1 & params[2] >= (params[3] / 2) - 1, TRUE, FALSE),
         matern = ifelse(params[2] > 0, TRUE, FALSE),
         cauchy = ifelse(params[2] > 0 & params[2] <= 2 & params[3] >= 0, TRUE, FALSE)
  )
}


# == Isotropic kernels ==
# These are taken from the following paper and book:
# Genton, Marc. (2001). Classes of Kernels for Machine Learning: A Statistics Perspective.
# Journal of Machine Learning Research. 2. 299-312. 10.1162/15324430260185646.
#
# Hristopulos, D. T. (2020). Random Fields for Spatial Data Modeling:  A Primer for Scientists and Engineers.
# In Advances in Geographic Information Science. Springer Netherlands. https://doi.org/10.1007/978-94-024-1918-4

#' 1D Isotropic Kernels.
#'
#' This function computes one of the isotropic kernels listed below. Note that `kernel()` is **deprecated**, please use `kernel_ec()` instead.
#' Unlike [kernel_symm_ec], these kernels are only defined when \eqn{x \geq 0.} They are used as
#' kernel multipliers in estimators [corrected_est] and [kernel_est].
#'
#' \strong{Gaussian Kernel}.
#' The isotropic Gaussian kernel, which is positive-definite for \eqn{{R}^{d}, d \in N,} is defined as
#' \deqn{a(x;\theta) = \exp(-x^{2} / \theta), \theta > 0 .}
#' The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}.
#'
#' \strong{Exponential Kernel}.
#' The isotropic exponential kernel, which is positive-definite for \eqn{{R}^{d}, d \in N,} is defined as
#' \deqn{a(x;\theta) = \exp(-x / \theta) , \theta > 0 .}
#' The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}.
#'
#' \strong{Isotropic Wave (Cardinal Sine) Kernel}.
#' The isotropic wave (cardinal sine) kernel, which is positive-definite for \eqn{{R}^{d}, d \leq 3,} is given by
#' \deqn{a(x;\theta) = \left\{ \begin{array}{ll}
#' \frac{\theta}{x} \sin\left( \frac{x}{\theta} \right), & x \neq 0 \\
#' 1, & x = 0
#' \end{array} \right.
#' }
#' where \eqn{\theta > 0 .}
#' The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}.
#'
#' \strong{Isotropic Rational Quadratic Kernel}.
#' The isotropic rational quadratic kernel, which is positive-definite for \eqn{{R}^{d}, d \in N,} is defined as
#' \deqn{a(x;\theta) = \frac{\theta}{x^{2} + \theta} , \theta >  0 .}
#' The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}.
#'
#' \strong{Isotropic Spherical Kernel}.
#' The isotropic spherical kernel, which is positive-definite for \eqn{{R}^{3}, d \leq 3,} is given by
#' \deqn{a(x;\theta) = \left\{
#' \begin{array}{ll}
#' 1 - \frac{3}{2}\frac{x}{\theta} + \frac{1}{2}\left( \frac{x}{\theta} \right)^{3}, & x < \theta \\
#' 0, & \mbox{otherwise}
#' \end{array} \right.
#' }
#' where \eqn{\theta > 0 .}
#' The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}.
#'
#' \strong{Isotropic Circular Kernel}.
#' The isotropic circular kernel, which is positive-definite for \eqn{{R}^{d}, d \leq 2,} is given by
#' \deqn{a(x;\theta) = \left\{
#' \begin{array}{ll}
#' \frac{2}{\pi}\arccos\left( \frac{x}{\theta} \right) - \frac{2}{\pi}\frac{x}{\theta} \sqrt{ 1 - \left( \frac{x}{\theta} \right)^{2} }, & x < \theta \\
#' 0, & \mbox{otherwise}
#' \end{array} \right.
#' }
#' where \eqn{\theta > 0 .}
#' The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}.
#'
#' \strong{Isotropic Matérn Kernel}.
#' The isotropic Matérn kernel, which is positive-definite for \eqn{{R}^{d}, d \in N,} and when \eqn{\nu > 0,} is defined as
#' \deqn{a(x; \theta, \nu) = \left(\sqrt{2\nu} \frac{x}{\theta} \right)^{\nu} \left(2^{\nu - 1} \Gamma(\nu) \right)^{-1} K_{\nu}\left( \sqrt{2\nu} \frac{x}{\theta} \right) ,}
#' where \eqn{\theta > 0} and \eqn{K_{\nu}(\cdot)} is the modified Bessel function of the second kind.
#' The \code{params} argument is of the form \code{c(}\eqn{\theta, \nu}\code{)}.
#'
#' \strong{Isotropic Bessel Kernel}.
#' The isotropic Bessel kernel, which is positive-definite for \eqn{{R}^{d}, d \in N,} and when \eqn{\nu \geq \frac{d}{2} - 1,} is given by
#' \deqn{a(x; \theta, \nu) = 2^{\nu} \Gamma(\nu + 1) J_{\nu}(x / \theta) (x / \theta)^{-\nu} ,}
#' where \eqn{\theta > 0} and \eqn{J_{\nu}(\cdot)} is the Bessel function of the first kind.
#' The \code{params} argument is of the form \code{c(}\eqn{\theta, \nu, d} \code{)}.
#'
#' \strong{Isotropic Cauchy Kernel}.
#' The isotropic Cauchy kernel, which is positive-definite for \eqn{{R}^{d}, d \in N,} and when \eqn{0 < \alpha \leq 2} and \eqn{\beta \geq 0,} is defined by
#' \deqn{a(x ; \theta, \alpha, \beta) = (1 + (x / \theta)^{\alpha})^{-(\beta / \alpha)} , \theta > 0 .}
#' The \code{params} argument is of the form \code{c(}\eqn{\theta, \alpha, \beta} \code{)}.
#'
#' @references
#' Genton, M. (2001). Classes of Kernels for Machine Learning: A Statistics Perspective. Journal of Machine Learning Research. 2, 299-312. https://doi.org/10.1162/15324430260185646
#'
#' Table 4.2 in Hristopulos, D. T. (2020). Random Fields for Spatial Data Modeling:  A Primer for Scientists and Engineers. Springer. https://doi.org/10.1007/978-94-024-1918-4
#'
#' @param x A vector or matrix of arguments of at least length 1 for which the kernel is computed at.
#' @param name The name of the kernel. Options are: gaussian, exponential, wave, rational_quadratic, spherical, circular, bessel_j, matern, and cauchy.
#' @param params A vector of parameters for the kernel. See the documentation below for the position of the parameters. All kernels have a scale parameter as the first value in the vector.
#'
#' @return A vector or matrix of kernel values.
#' @export
#'
#' @rdname kernel_ec
#'
#' @examples
#' x <- c(0.2, 0.4, 0.6)
#' theta <- 0.9
#' kernel_ec(x, "gaussian", c(theta))
#' kernel_ec(x, "exponential", c(theta))
#' kernel_ec(x, "wave", c(theta))
#' kernel_ec(x, "rational_quadratic", c(theta))
#' kernel_ec(x, "spherical", c(theta))
#' kernel_ec(x, "circular", c(theta))
#' nu <- 1
#' kernel_ec(x, "matern", c(theta, nu))
#' dim <- 1
#' kernel_ec(x, "bessel_j", c(theta, nu, dim))
#' alpha <- 1
#' beta <- 2
#' kernel_ec(x, "cauchy", c(theta, alpha, beta))
#' curve(kernel_ec(x, "gaussian", c(theta)), from = 0, to = 5)
#' curve(kernel_ec(x, "exponential", c(theta)), from = 0, to = 5)
#' curve(kernel_ec(x, "wave", c(theta)), from = 0, to = 5)
#' curve(kernel_ec(x, "rational_quadratic", c(theta)), from = 0, to = 5)
#' curve(kernel_ec(x, "spherical", c(theta)), from = 0, to = 5)
#' curve(kernel_ec(x, "circular", c(theta)), from = 0, to = 5)
#' curve(kernel_ec(x, "matern", c(theta, nu)), from = 0, to = 5)
#' curve(kernel_ec(x, "bessel_j", c(theta, nu, dim)), from = 0, to = 5)
#' curve(kernel_ec(x, "cauchy", c(theta, alpha, beta)), from = 0, to = 5)
kernel_ec <- function(x, name = c("gaussian", "exponential", "wave", "rational_quadratic", "spherical", "circular", "bessel_j", "matern", "cauchy"), params=c(1)) {
  stopifnot(length(x) >= 1, all(x >= 0), is.numeric(x), !any(is.na(x)), length(params) > 0, is.numeric(params[1]), params[1] > 0)
  name <- match.arg(name)

  if(!.validate_kernel_params(name, params)) {
    stop("Parameter choice `(", paste0(params, collapse = ", "), ")` is not appropriate for `", name, "` kernel")
  }

  returnVal <- switch(name,
         gaussian = .gaussian_kernel(x, params),
         exponential = .exponential_kernel(x, params),
         wave = .wave_kernel(x, params),
         rational_quadratic = .rational_quadratic_kernel(x, params),
         spherical = .spherical_kernel(x, params),
         circular = .circular_kernel(x, params),
         matern = .matern_kernel(x, params),
         bessel_j = .bessel_j_kernel(x, params),
         cauchy = .cauchy_kernel(x, params)
  )

  return(returnVal)
}

# == Symmetric kernels ==
# These are to be used with Hall's estimator.
# Unlike the isotropic case, these are symmetric probability distributions.

#' 1D Isotropic Symmetric Kernels.
#'
#' These functions computes values of kernels that have the properties of symmetric probability distributions. Note that `kernel_symm()` is **deprecated**, please use `kernel_symm_ec()` instead.
#' For a kernel \eqn{a(x)}, the standardised version is \eqn{a(x) / \int_{-\infty}^{\infty} a(x) dx}, so that the integral is 1.
#' The symmetric kernels are different to [kernel] and are used in the functions [adjusted_est] and [truncated_est].
#'
#' \strong{Symmetric Gaussian Kernel}.
#' The symmetric Gaussian kernel is defined as
#' \deqn{a(x;\theta) = \sqrt{\pi \theta} \exp(-x^{2} / \theta) , \theta > 0.}
#' The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}.
#'
#' \strong{Symmetric Wave Kernel}.
#' The wave (cardinal sine) kernel is given by
#' \deqn{a(x;\theta) = \left\{ \begin{array}{ll}
#' (\sqrt{\theta^{2}} \pi)^{-1} \frac{\theta}{x} \sin\left( \frac{x}{\theta} \right), & x \neq 0 \\
#' 1, & x = 0
#' \end{array} \right.}
#' where \eqn{\theta > 0.}
#' The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}
#'
#' \strong{Symmetric Rational Quadratic Kernel}.
#' The symmetric rational quadratic kernel is given by
#' \deqn{a(x;\theta) = (\pi \sqrt{\theta})^{-1} \left(\frac{\theta}{x^{2} + \theta}\right), \theta > 0.}
#' The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}
#'
#' \strong{Symmetric Besesel Kernel}.
#' The symmetric Bessel kernel, which is valid when \eqn{\nu \geq \frac{d}{2} - 1,} is given by
#' \deqn{a(x; \theta, \nu) = \left(\Gamma\left(\frac{1}{2} + \nu\right)/(2 \sqrt{\pi} \theta \Gamma(1 + \nu))\right) ( 2^{\nu} \Gamma(\nu + 1) J_{\nu}(x / \theta) (x / \theta)^{-\nu}), \,\theta > 0, \nu \geq \frac{d}{2} - 1,}
#' where \eqn{J_{\nu}(\cdot)} is the Bessel function of the first kind and \eqn{d} is the dimension.
#' The \code{params} argument is of the form \code{c(}\eqn{\theta, \nu, d}\code{)}.
#'
#' @param x A vector or matrix of arguments of at least length 1 for which the kernel is computed at. Each value can be negative as well as positive.
#' @param name The name of the kernel. Options are: gaussian, wave, rational_quadratic, bessel_j.
#' @param params A vector of parameters for the kernel. See the documentation below for the position of the parameters. All kernels will have a scale parameter as the first value in the vector.
#'
#' @return A vector or matrix of values.
#' @export
#'
#' @rdname kernel_symm_ec
#'
#' @examples
#' x <- c(-2, -1, 0, 1, 2)
#' theta <- 1
#' kernel_symm_ec(x, "gaussian", c(theta))
#' kernel_symm_ec(x, "wave", c(theta))
#' kernel_symm_ec(x, "rational_quadratic", c(theta))
#' dim <- 1
#' nu <- 1
#' kernel_symm_ec(x, "bessel_j", c(theta, nu, dim))
#' curve(kernel_symm_ec(x, "gaussian", c(theta)), from = -5, to = 5)
#' curve(kernel_symm_ec(x, "wave", c(theta)), from = -5, to = 5)
#' curve(kernel_symm_ec(x, "rational_quadratic", c(theta)), from = -5, to = 5)
#' curve(kernel_symm_ec(x, "bessel_j", c(theta, nu, dim)), from = -5, to = 5)
kernel_symm_ec <- function(x, name = c("gaussian", "wave", "rational_quadratic", "bessel_j"), params=c(1)) {
  stopifnot(length(x) >= 1, is.numeric(x), !any(is.na(x)), length(params) > 0, is.numeric(params[1]), params[1] > 0)

  name <- match.arg(name)

  if(!.validate_kernel_params(name, params)) {
    stop("Parameter choice `(", paste0(params, collapse = ", "), ")` is not appropriate for `", name, "` symmetric kernel")
  }

  returnVal <- switch(name,
         gaussian = sqrt(pi * params[1])^(-1) * .gaussian_kernel(abs(x), params),
         wave = (sqrt(params[1]^(2)) * pi)^(-1) * .wave_kernel(abs(x), params),
         rational_quadratic = ((pi * sqrt(params[1]))^(-1)) * .rational_quadratic_kernel(abs(x), params),
         bessel_j = (gamma((1/2 + params[2])) / (2*sqrt(pi) * params[1] * gamma(1 + params[2]))) * (2^params[2]) * gamma(params[2] + 1) * .bessel_j_kernel(abs(x), params)
  )

  return(returnVal)
}

# Gaussian kernel
test_that("kernel_ec(, \"gaussian\", ) works", {
  expect_equal(kernel_ec(c(0, 1, Inf), "gaussian", c(1)), c(1, exp(-(1^2)), 0))
})

test_that("kernel_ec(, \"gaussian\", ) fails for empty x, theta=1", {
  expect_error(kernel_ec(c(), "gaussian", c(1)))
})

test_that("kernel_ec(, \"gaussian\", ) fails for theta <= 0", {
  expect_error(kernel_ec(c(0, 1, Inf), "gaussian", c(0)))
  expect_error(kernel_ec(c(0, 1, Inf), "gaussian", c(-0.1)))
})

test_that("kernel_ec(, \"exponential\", ) fails for any negative x, theta=1", {
  expect_error(kernel_ec(c(0, 1, -1, -2), "gaussian", c(1)))
})

# Exponential kernel
test_that("kernel_ec(, \"exponential\", ) works", {
  expect_equal(kernel_ec(c(0, 1, Inf), "exponential", c(1)), c(1, exp(-1), 0))
})

test_that("kernel_ec(, \"exponential\", ) fails for empty x, theta=1", {
  expect_error(kernel_ec(c(), "exponential", c(1)))
})

test_that("kernel_ec(, \"exponential\", ) fails for theta <= 0", {
  expect_error(kernel_ec(c(0, 1, Inf), "exponential", c(0)))
  expect_error(kernel_ec(c(0, 1, Inf), "exponential", c(-0.1)))
})

test_that("kernel_ec(, \"exponential\", ) fails for any negative x, theta=1", {
  expect_error(kernel_ec(c(0, 1, -1, -2), "exponential", c(1)))
})

# Wave kernel
test_that("kernel_ec(, \"wave\", ) works", {
  expect_equal(kernel_ec(c(0, 1, Inf), "wave", c(1)), c(1, (1/1) * sin(1 / 1), 0))
})

test_that("kernel_ec(, \"wave\", )  fails for empty x, theta=1", {
  expect_error(kernel_ec(c(), "wave", c(1)))
})

test_that("kernel_ec(, \"wave\", )  fails for theta <= 0", {
  expect_error(kernel_ec(c(0, 1, Inf), "wave", c(0)))
  expect_error(kernel_ec(c(0, 1, Inf), "wave", c(-0.1)))
})

test_that("kernel_ec(, \"wave\", ) fails for any negative x, theta=1", {
  expect_error(kernel_ec(c(0, 1, -1, -2), "wave", c(1)))
})

# Rational Quadratic kernel
test_that("kernel_ec(, \"rational_quadratic\", ) works", {
  expect_equal(kernel_ec(c(0, 1, Inf), "rational_quadratic", c(1)), c((1 - (0^2 / (0^2 + 1))), (1 - (1^2 / (1^2 + 1))), 0))
})

test_that("kernel_ec(, \"rational_quadratic\", ) fails for empty x, theta=1", {
  expect_error(kernel_ec(c(), "rational_quadratic", c(1)))
})

test_that("kernel_ec(, \"rational_quadratic\", ) fails for theta <= 0", {
  expect_error(kernel_ec(c(0, 1, Inf), "rational_quadratic", c(0)))
  expect_error(kernel_ec(c(0, 1, Inf), "rational_quadratic", (-0.1)))
})

test_that("kernel_ec(, \"rational_quadratic\", ) fails for any negative x, theta=1", {
  expect_error(kernel_ec(c(0, 1, -1, -2), "rational_quadratic", c(1)))
})

# Spherical kernel
test_that("kernel_ec(, \"spherical\", ) works", {
  expect_equal(kernel_ec(c(0, 1, Inf), "spherical", c(2)), c(1, 1 - ((3/2) * (1 / 2)) + ((1/2) * (1 / 2)^3), 0))
})

test_that("kernel_ec(, \"spherical\", ) fails for empty x, theta=1", {
  expect_error(kernel_ec(c(), "spherical", c(1)))
})

test_that("kernel_ec(, \"spherical\", ) fails for theta <= 0", {
  expect_error(kernel_ec(c(0, 1, Inf), "spherical", c(0)))
  expect_error(kernel_ec(c(0, 1, Inf), "spherical", c(-0.1)))
})

test_that("kernel_ec(, \"spherical\", ) fails for any negative x, theta=1", {
  expect_error(kernel_ec(c(0, 1, -1, -2), "spherical", c(1)))
})

# Circular kernel
test_that("kernel_ec(, \"circular\", ) works", {
  expect_equal(kernel_ec(c(0, 1, Inf), "circular", c(2)), c(1, ((2 / pi) * acos(1 / 2)) - ((2 / pi) * (1 / 2) * sqrt(1 - (1 / 2)^2)), 0))
})

test_that("kernel_ec(, \"circular\", ) fails for empty x, theta=1", {
  expect_error(kernel_ec(c(), "circular", c(1)))
})

test_that("kernel_ec(, \"circular\", ) fails for theta <= 0", {
  expect_error(kernel_ec(c(0, 1, Inf), "circular", c(0)))
  expect_error(kernel_ec(c(0, 1, Inf), "circular", c(-0.1)))
})

test_that("kernel_ec(, \"circular\", )fails for any negative x, theta=1", {
  expect_error(kernel_ec(c(0, 1, -1, -2), "circular", c(1)))
})

# Bessel kernel
test_that("kernel_ec(, \"bessel_j\", ) works", {
  expect_equal(kernel_ec(c(0, 1, Inf), "bessel_j", c(1, 0, 2)), c(1, (2^0) * gamma(0 + 1) * (besselJ(1 / 1, 0) / ((1 / 1)^0)), 0))
})

test_that("kernel_ec(, \"bessel_j\", ) fails for empty x, theta=1", {
  expect_error(kernel_ec(c(), 1, 0, 2))
})

test_that("kernel_ec(, \"bessel_j\", ) fails for theta <= 0", {
  expect_error(kernel_ec(c(0, 1, Inf), "bessel_j",  c(0, 0, 2)))
  expect_error(kernel_ec(c(0, 1, Inf), "bessel_j",  c(-0.1, 0, 2)))
})

test_that("kernel_ec(, \"bessel_j\", ) fails for any negative x, theta=1", {
  expect_error(kernel_ec(c(0, 1, -1, -2), "bessel_j",  c(1, 0, 2)))
})

test_that("kernel_ec(, \"bessel_j\", ) fails for nu <= (dim / 2) - 1, theta=1, dim=2", {
  expect_error(kernel_ec(c(0, 1, Inf), "bessel_j", c(1, -0.5, 2)))
})

# Matérn kernel
test_that("kernel_ec(, \"matern\", ) works", {
  expect_equal(kernel_ec(c(0, 1, Inf), "matern", c(1, 1)), c(1, (((sqrt(2 * 1) * 1 / 1)^1) / (2^(1 - 1) * gamma(1))) * besselK(sqrt(2 * 1) * 1 / 1, 1), 0))
})

test_that("kernel_ec(, \"matern\", ) fails for empty x, theta=1", {
  expect_error(kernel_ec(c(), "matern", c(1, 1)))
})

test_that("kernel_ec(, \"matern\", ) fails for theta <= 0", {
  expect_error(kernel_ec(c(0, 1, Inf), "matern", c(0, 1)))
  expect_error(kernel_ec(c(0, 1, Inf), "matern", c(-0.1, 1)))
})

test_that("kernel_ec(, \"matern\", ) fails for any negative x, theta=1", {
  expect_error(kernel_ec(c(0, 1, -1, -2), "matern", c(1, 0)))
})

test_that("kernel_ec(, \"matern\", ) fails for nu <= 0, theta=1", {
  expect_error(kernel_ec(c(0, 1, Inf), "matern", c(1, 0)))
  expect_error(kernel_ec(c(0, 1, Inf), "matern", c(1, -0.5)))
})

# Cauchy kernel
test_that("kernel_ec(, \"cauchy\", ) works", {
  expect_equal(kernel_ec(c(0, 1, Inf), "cauchy", c(1, 1, 1)), c(1, 0.5, 0))
})

test_that("kernel_ec(, \"cauchy\", ) fails for empty x, theta=1", {
  expect_error(kernel_ec(c(), "cauchy", c(1, 1, 1)))
})

test_that("kernel_ec(, \"cauchy\", ) fails for theta < 0", {
  expect_error(kernel_ec(c(0, 1, Inf), "cauchy", c(0, 1, 1)))
  expect_error(kernel_ec(c(0, 1, Inf), "cauchy", c(-0.1, 1, 1)))
})

test_that("kernel_ec(, \"cauchy\", ) fails for any negative x, theta=1", {
  expect_error(kernel_ec(c(0, 1, -1, -2), "cauchy", c(1, 1, 1)))
})

test_that("kernel_ec(, \"cauchy\", ) fails for alpha <= 0, theta=1, beta=1", {
  expect_error(kernel_ec(c(0, 1, Inf), "cauchy", c(1, 0, 1)))
  expect_error(kernel_ec(c(0, 1, Inf), "cauchy", c(1, -0.1, 1)))
})

test_that("kernel_ec(, \"cauchy\", ) fails for alpha > 2, theta=1, beta=1", {
  expect_error(kernel_ec(c(0, 1, Inf), "cauchy", c(1, 3, 1)))
})

test_that("kernel_ec(, \"cauchy\", ) fails for beta < 0, theta=1, alpha=1", {
  expect_error(kernel_ec(c(0, 1, Inf), "cauchy", c(1, 1, -0.1)))
})

test_that('.validate_kernel_params() works', {
  expect_equal(.validate_kernel_params("gaussian", c(1)), TRUE)
})

test_that('.validate_kernel_params() fails for name not in "gaussian", "exponential", "wave", "rational_quadratic", "spherical", "circular", "bessel_j", "matern", "cauchy"', {
  expect_error(.validate_kernel_params("test", c(1)))
})

# Gaussian kernel
test_that("kernel_symm_ec(, \"gaussian\", ) works", {
  expect_equal(kernel_symm_ec(c(0, 1, Inf), "gaussian", c(1)), c(sqrt(pi)^(-1)*1, sqrt(pi)^(-1)*exp(-(1^2)), 0))
})

test_that("kernel_symm_ec(, \"gaussian\", ) fails for empty x, theta=1", {
  expect_error(kernel_symm_ec(c(), "gaussian", c(1)))
})

test_that("kernel_gaussian(, \"gaussian\", ) fails for theta <= 0", {
  expect_error(kernel_symm_ec(c(0, 1, Inf), "gaussian", c(0)))
  expect_error(kernel_symm_ec(c(0, 1, Inf), "gaussian", c(-0.1)))
})

test_that("kernel_symm_ec(, \"gaussian\", ) fails for any NA x, theta=1", {
  expect_error(kernel_symm_ec(c(0, 1, NA), "gaussian", c(1)))
})

# Wave kernel
test_that("kernel_symm_ec(, \"wave\", ) works", {
  expect_equal(kernel_symm_ec(c(0, 1, Inf), "wave", c(1)), c((sqrt(1^(-2)) / pi) * 1, (sqrt(1^(-2)) / pi) * (1/1) * sin(1 / 1), 0))
})

test_that("kernel_symm_ec(, \"wave\", ) fails for empty x, theta=1", {
  expect_error(kernel_symm_ec(c(), "wave", c(1)))
})

test_that("kernel_symm_ec(, \"wave\", ) fails for theta <= 0", {
  expect_error(kernel_symm_ec(c(0, 1, Inf), "wave", c(0)))
  expect_error(kernel_symm_ec(c(0, 1, Inf), "wave", c(-0.1)))
})

test_that("kernel_symm_ec(, \"wave\", ) fails for any NA x, theta=1", {
  expect_error(kernel_symm_ec(c(0, 1, NA), "wave", c(1)))
})

# Rational Quadratic kernel
test_that("kernel_symm_ec(, \"rational_quadratic\", ) works", {
  expect_equal(kernel_symm_ec(c(0, 1, Inf), "rational_quadratic", c(1)), c(((pi * sqrt(1))^(-1)) * (1 - (0^2 / (0^2 + 1))), ((pi * sqrt(1))^(-1)) * (1 - (1^2 / (1^2 + 1))), 0))
})

test_that("kernel_symm_ec(, \"rational_quadratic\", ) fails for empty x, theta=1", {
  expect_error(kernel_symm_ec(c(), "rational_quadratic", c(1)))
})

test_that("kernel_symm_ec(, \"rational_quadratic\", ) fails for theta <= 0", {
  expect_error(kernel_symm_ec(c(0, 1, Inf), "rational_quadratic", c(0)))
  expect_error(kernel_symm_ec(c(0, 1, Inf), "rational_quadratic",  c(-0.1)))
})

test_that("kernel_symm_ec(, \"rational_quadratic\", ) fails for any NA x, theta=1", {
  expect_error(kernel_symm_ec(c(0, 1, NA), "rational_quadratic", c(1)))
})

# Bessel kernel
test_that("kernel_symm_ec(, \"bessel_j\", ) works", {
  expect_equal(kernel_symm_ec(c(0, 1, Inf), "bessel_j", c(1, 0, 2)), c((gamma((1/2) + 0) / (2*sqrt(pi) * 1 * gamma(1 + 0))) * 1,
              (gamma((1/2) + 0) / (2*sqrt(pi) * 1 * gamma(1 + 0))) * ((2^0) * gamma(0 + 1) * (besselJ(1 / 1, 0) / ((1 / 1)^0))), 0))
})

test_that("kernel_symm_ec(, \"bessel_j\", ) fails for empty x, theta=1", {
  expect_error(kernel_symm_ec(c(), "bessel_j", c(1, 0, 2)))
})

test_that("kernel_symm_ec(, \"bessel_j\", ) fails for theta <= 0", {
  expect_error(kernel_symm_ec(c(0, 1, Inf), "bessel_j", c(0, 0, 2)))
  expect_error(kernel_symm_ec(c(0, 1, Inf), "bessel_j", c(-0.1, 0, 2)))
})

test_that("kernel_symm_ec(, \"bessel_j\", ) fails for any NA x, theta=1", {
  expect_error(kernel_symm_ec(c(0, 1, NA), "bessel_j", c(1)))
})

test_that("kernel_symm_ec(, \"gaussian\") works", {
  expect_equal(kernel_symm_ec(c(0, 1, 2), "gaussian"), c(0.56418958, 0.20755375, 0.01033349))
})

test_that("kernel_symm_ec(, \"gaussian\") fails for at least one NA in x", {
  expect_error(kernel_symm_ec(c(0, NA, 2), "gaussian"))
})

test_that("kernel_symm_ec(, \"gaussian\") fails for nonnumeric x", {
  expect_error(kernel_symm_ec(c(0, 'a', 2), "gaussian"))
  expect_error(kernel_symm_ec(c(0, 1i, 2), "gaussian"))
})

test_that("kernel_symm_ec(, \"gaussian\") fails empty x", {
  expect_error(kernel_symm_ec(c(), "gaussian"))
})

#' Area Between Estimated Autocovariance Functions.
#'
#' This function estimates the area between two estimated autocovariance functions.
#'
#' @details
#' This function estimates the area between two estimated autocovariance functions over a set of lags, from 0 up to \eqn{h_{n}} defined by
#' \deqn{
#' \int_{0}^{h_{n}} \left| \hat{C}_{1}(h) - \hat{C}_{2}(h) \right| dh ,
#' }
#' where \eqn{\hat{C}_{1}(\cdot)} and \eqn{\hat{C}_{2}(\cdot)} are estimated autocovariance functions.
#'
#' To approximate this integral the trapezoidal rule is used.
#'
#' If \code{lags} is empty, a uniform time grid with a step of 1 will be used which may result in a different area than if lags is specified.
#'
#' @param est1 A numeric vector representing the first estimated autocovariance function.
#' @param est2 A numeric vector of the same length as \code{est1} representing the second estimated autocovariance function
#' @param lags An optional vector of lags starting from 0 up until some other lag. If empty, a vector of lags is created starting from 0 until \code{len(est1) - 1}, by 1.
#' @param plot A boolean determining whether a plot should be created. By default, no plot is created.
#'
#' @return A numeric value representing the estimated area between two estimated autocovariance functions.
#' @export
#'
#' @importFrom graphics legend
#' @importFrom graphics lines
#' @importFrom graphics polygon
#'
#' @examples
#' x <- seq(0, 5, by=0.1)
#' estCov1 <- exp(-x^2)
#' estCov2 <- exp(-x^2.1)
#' area_between(estCov1, estCov2, lags=x)
#' area_between(estCov1, estCov2, lags=x, plot = TRUE)
area_between <- function(est1, est2, lags=c(), plot = FALSE) {
  UseMethod("area_between")
}

#' @describeIn area_between Method for `CovEsts` objects.
#' @export
area_between.CovEsts <- function(est1, est2, lags=c(), plot = FALSE) {
  if(length(lags)) {
    return(area_between.default(est1$acf, est2$acf, lags = lags, plot = plot))
  }
  return(area_between.default(est1$acf, est2$acf, lags = est1$lags, plot = plot))
}

#' @describeIn area_between Method for numeric vectors.
#' @export
area_between.default <- function(est1, est2, lags=c(), plot = FALSE) {
  stopifnot(is.numeric(est1), is.numeric(est2), length(est1) >= 1, length(est2) >= 1, !any(is.na(est1)), !any(is.na(est2)),
            length(est1) == length(est2), is.logical(plot))

  if(length(lags) != 0) {
    stopifnot(is.numeric(lags), length(est1) == length(lags), length(est2) == length(lags))
  }
  # Generate a vector of lags, {0, ..., length(est1) - 1}
  else {
    lags <- seq(0, length(est1) - 1)
  }

  yDiff <- abs(est1 - est2)
  integral_val <- 0
  for(i in 2:length(yDiff)) {
    integral_val_temp <-  ((yDiff[i - 1] + yDiff[i]) / 2) * (lags[i] - lags[i - 1])
    integral_val <- integral_val + integral_val_temp
  }

  if(plot) {
      y_min <- ifelse(min(est1) < min(est2), min(est1), min(est2))
      y_max <- ifelse(max(est1) > max(est2), max(est1), max(est2))
      plot(lags, est1, lwd=2, lty=1, col=2, type='l', ylim=c(y_min, y_max), xlab='Lag (h)', ylab='Estimated Autocovariance')
      graphics::lines(lags, est2, lwd=2, lty=1, col=3)
      # Create a region between the two curves.
      graphics::polygon(c(lags, rev(lags)), c(est2, rev(est1)), col="#AAAAAAAA", border=2, lty=0)
      graphics::legend("topright", inset=0, title="Estimated Autocovariance", c("1", "2", "Area Between"),
             col=c(2, 3, "#AAAAAAAA"), lwd=c(2, 2, NA), lty=c(1, 1, NA), density=c(0, 0, NA), fill=c(NA, NA, "#AAAAAAAA"), border=c(NA, NA, "#000000FF"))
  }

  return(integral_val)
}

#' Maximum Vertical Distance Between Estimated Functions.
#'
#' This function computes the maximum vertical distance between functions.
#'
#' @details
#' This function computes the maximum vertical distance between functions:
#' \deqn{D(\hat{C}_{1}(h), \hat{C}_{2}(h)) = \displaystyle \max_{h} \left| \hat{C}_{1}(h) - \hat{C}_{2}(h) \right| ,
#' }
#' where \eqn{\hat{C}_{1}(\cdot)} and \eqn{\hat{C}_{2}(\cdot)} are estimated autocovariance functions.
#' It assumes that the estimated values are given for the same set of lags.
#' The vectors of the function values must be of the same length.
#'
#' @param est1 A numeric vector representing the first estimated autocovariance function.
#' @param est2 A numeric vector of the same length as \code{est1} representing the second estimated autocovariance function
#' @param lags An optional vector of lags starting from 0 up until some other lag. If empty, a vector of lags is created starting from 0 until \code{len(est1) - 1}, by 1.
#' @param plot A boolean as to whether a plot should be created. By default, no plot is created.
#'
#' @return A numeric value representing the maximum vertical distance between the two estimated functions.
#' @export
#'
#' @examples
#' x <- seq(0, 5, by=0.1)
#' estCov1 <- exp(-x^2)
#' estCov2 <- exp(-x^2.1)
#' max_distance(estCov1, estCov2, lags=x)
#' max_distance(estCov1, estCov2, lags=x, plot = TRUE)
max_distance <- function(est1, est2, lags=c(), plot = FALSE) {
  UseMethod("max_distance")
}

#' @describeIn max_distance Method for `CovEsts` objects.
#' @export
max_distance.CovEsts <- function(est1, est2, lags=c(), plot = FALSE) {
  if(length(lags)) {
    return(max_distance.default(est1$acf, est2$acf, lags = lags, plot = plot))
  }
  return(max_distance.default(est1$acf, est2$acf, lags = est1$lags, plot = plot))
}

#' @describeIn max_distance Method for numeric vectors.
#' @export
max_distance.default <- function(est1, est2, lags=c(), plot = FALSE) {
  stopifnot(is.numeric(est1), is.numeric(est2), length(est1) >= 1, length(est2) >= 1, !any(is.na(est1)), !any(is.na(est2)),
            length(est1) == length(est2), is.logical(plot))

  if(length(lags) != 0) {
    stopifnot(is.numeric(lags), length(est1) == length(lags), length(est2) == length(lags))
  }
  # Generate a vector of lags, {0, ..., length(est1) - 1}
  else {
    lags <- seq(0, length(est1) - 1)
  }

  maxDist <- -Inf

  distVec <- rep(Inf, length(est1))

  for(i in 1:length(est1)) {
    d <- abs(est1[i] - est2[i])
    if(is.numeric(d)) {
      distVec[i] <- d
      if(d > maxDist) {
        maxDist <- d
      }
    }
    else {
      distVec[i] <- 0
    }
  }

  if(plot) {
    plot(lags, distVec, type='l', xlab='Lag (h)', ylab='Distance')
  }

  return(maxDist)
}

#' Create a Cyclic Matrix for a Given Vector.
#'
#' This helper function creates a symmetric matrix from a given vector \eqn{v}.
#'
#' @details
#' This function creates a symmetric matrix for a given vector \eqn{v}.
#' If \eqn{v = \{v_{0}, v_{1} , \dots , v_{N-1} , v_{N} \},} then the symmetric matrix will has the form
#' \deqn{\left[ {\begin{array}{ccccc}
#' v_{0}     & v_{1}     & \cdots & v_{N - 1} & v_{N}     \\
#' v_{1}     & v_{0}     & \cdots & v_{N - 2} & v_{N - 1} \\
#' \vdots    & \vdots    & \ddots & \vdots    & \vdots    \\
#' v_{N - 1} & v_{N- 2}  & \cdots & v_{0}     & v_{1}     \\
#' v_{N}     & v_{N - 1} & \cdots & v_{1}     & v_{0}     \\
#' \end{array}} \right]
#' }
#'
#' @param v A numeric vector.
#'
#' @return A symmetric matrix.
#'
#' @importFrom stats toeplitz
#'
#' @examples
#' \dontrun{
#' v <- c(1, 2, 3)
#' cyclic_matrix(v)
#' }
cyclic_matrix <- function(v) {
  stopifnot(is.numeric(v), !any(is.na(v)), length(v) >= 1)
  return(toeplitz(v))
}

#' Compute the Spectral Norm Between Estimated Functions.
#'
#' This function computes the spectral norm of the difference of two estimated autocovariance functions.
#' This function is intended for estimates over lags with a constant difference.
#'
#' @details
#' This function computes the spectral norm of the difference of two estimated autocovariance functions.
#' Let \eqn{D(h) = \hat{C}_{1}(h) - \hat{C}_{2}(h),}
#' where \eqn{\hat{C}_{1}(\cdot)} and \eqn{\hat{C}_{2}(\cdot)} are estimated autocovariance functions.
#'
#' A matrix \eqn{D} is created from \eqn{D(\cdot)},
#' \deqn{\left[ {\begin{array}{ccccc}
#' D(h_{0})     & D(h_{1})     & \cdots & D(h_{n - 1}) & D(h_{n})     \\
#' D(h_{1})     & D(h_{0})     & \cdots & D(h_{n - 2}) & D(h_{n - 1}) \\
#' \vdots       & \vdots       & \ddots & \vdots       & \vdots       \\
#' D(h_{n - 1}) & D(h_{n - 2}) & \cdots & D(h_{0})     & D(h_{1})     \\
#' D(h_{n})     & D(h_{n - 1}) & \cdots & D(h_{1})     & D(h_{0})     \\
#' \end{array}} \right] ,
#' }
#' over a set of lags \eqn{\{h_{0}, h_{1}, \dots , h_{N} \}.}
#' This matrix is created by [cyclic_matrix].
#'
#' The spectral norm is defined as the largest eigenvalue of \eqn{D.}
#'
#' @param est1 A numeric vector representing the first estimated autocovariance function.
#' @param est2 A numeric vector of the same length as \code{est1} representing the second estimated autocovariance function
#'
#' @return The spectral norm of the differences between the two functions.
#' @export
#'
#' @examples
#' x <- seq(0, 5, by=0.1)
#' estCov1 <- exp(-x^2)
#' estCov2 <- exp(-x^2.1)
#' spectral_norm(estCov1, estCov2)
spectral_norm <- function(est1, est2) {
  UseMethod("spectral_norm")
}

#' @describeIn spectral_norm Method for `CovEsts` objects.
#' @export
spectral_norm.CovEsts <- function(est1, est2) {
  return(spectral_norm.default(est1$acf, est2$acf))
}

#' @describeIn spectral_norm Method for numeric vectors.
#' @export
spectral_norm.default <- function(est1, est2) {
  stopifnot(is.numeric(est1), is.numeric(est2), length(est1) >= 1, length(est2) >= 1, !any(is.na(est1)), !any(is.na(est2)),
            length(est1) == length(est2))
  det1 <- norm(cyclic_matrix(est1 - est2), type = "2")
  if(is.numeric(det1)) {
    return(det1)
  }
  return(NA)
}

#' Check if an Autocovariance Function Estimate is Positive-Definite or Not.
#'
#' This function checks if an autocovariance function estimate is positive-definite or not by determining if the eigenvalues of the corresponding matrix (see the Details section) are all positive.
#'
#' @details
#' For an autocovariance function estimate \eqn{\hat{C}(\cdot)} over a set of lags separated by a constant difference \eqn{\{h_{0}, h_{1} , h_{2} , \dots , h_{n} \},}
#' construct the symmetric matrix
#' \deqn{\left[ {\begin{array}{ccccc}
#' \hat{C}(h_{0})     & \hat{C}(h_{1})     & \cdots & \hat{C}(h_{n - 1}) & \hat{C}(h_{n})     \\
#' \hat{C}(h_{1})     & \hat{C}(h_{0})     & \cdots & \hat{C}(h_{n - 2}) & \hat{C}(h_{n - 1}) \\
#' \vdots             & \vdots             & \ddots & \vdots             & \vdots             \\
#' \hat{C}(h_{n - 1}) & \hat{C}(h_{n - 2}) & \cdots & \hat{C}(h_{0})     & \hat{C}(h_{1})     \\
#' \hat{C}(h_{n})     & \hat{C}(h_{n - 1}) & \cdots & \hat{C}(h_{1})     & \hat{C}(h_{0})     \\
#' \end{array}} \right] .
#' }
#'
#' The eigendecomposition of this matrix is computed to determine if all eigenvalues are positive. If so, the estimated autocovariance function is assumed to be positive-definite.
#'
#' @param est A numeric vector, corresponding cyclic matrix representing an estimated autocovariance function, or a \code{CovEsts} S3 object.
#'
#' @return A boolean where \code{TRUE} denotes a positive-definite autocovariance function estimate and \code{FALSE} for an estimate that is not positive-definite.
#' @export
#'
#' @examples
#' x <- seq(0, 5, by=0.1)
#' estCov <- exp(-x^2)
#' check_pd(estCov)
check_pd <- function(est) {
  UseMethod("check_pd")
}

#' @describeIn check_pd Method for `CovEsts` objects.
#' @export
check_pd.CovEsts <- function(est) {
  return(check_pd.default(est$acf))
}

#' @describeIn check_pd Method for numeric vectors.
#' @export
check_pd.default <- function(est) {
  stopifnot(is.numeric(est), length(est) >= 1, !any(is.na(est)))

  if(is.matrix(est)) {
    eigens <- zapsmall(eigen(est)$values)
    return(!any(eigens < 0) && !all(eigens == 0))
  }

  cyclic_mat <- cyclic_matrix(est)
  eigens <- zapsmall(eigen(cyclic_mat)$values)
  return(!any(eigens < 0) && !all(eigens == 0))
}

#' MSE Between Estimated Autocovariance Functions.
#'
#' This function computes the mean-square difference/error between two autocovariance functions (estimated or theoretical).
#'
#' @details
#' This function computes the mean-square difference/error (MSE) between two estimated autocovariance functions (estimated or theoretical).
#' The MSE is defined as
#' \deqn{\frac{1}{n} \sum_{i=0}^{n} \left(\widehat{C}_{1}(h_{i}) - \widehat{C}_{2}(h_{i})\right)^{2}}
#' over a set of lags \eqn{\{h_{0}, h_{1} , h_{2} , \dots , h_{n} \}.}
#'
#' @param est1 A numeric vector representing the first estimated autocovariance function.
#' @param est2 A numeric vector of the same length as \code{est1} representing the second estimated (or theoretical) autocovariance function
#'
#' @return A numeric value representing the MSE between two autocovariance functions (estimated or theoretical).
#' @export
#'
#' @examples
#' x <- seq(0, 5, by=0.1)
#' estCov1 <- exp(-x^2)
#' estCov2 <- exp(-x^2.1)
#' mse(estCov1, estCov2)
mse <- function(est1, est2) {
  UseMethod("mse")
}

#' @describeIn mse Method for `CovEsts` object.
#' @export
mse.CovEsts <- function(est1, est2) {
  return(mse.default(est1$acf, est2$acf))
}

#' @describeIn mse Method for numeric vectors.
#' @export
mse.default <- function(est1, est2) {
  stopifnot(is.numeric(est1), is.numeric(est2), length(est1) >= 1, length(est2) >= 1, !any(is.na(est1)), !any(is.na(est2)),
            length(est1) == length(est2))
  return(mean((est1 - est2)^2))
}

#' Hilbert-Schmidt Norm Between Estimated Autocovariance Functions.
#'
#' This function computes the Hilbert-Schidmt norm between two estimated autocovariance functions.
#'
#' @details
#' This function computes the Hilbert-Schidmt norm between two estimated autocovariance functions.
#' The Hilbert-Schmidt norm of a matrix
#' \deqn{D = \left[(d_{i,j})_{1 \le i,j \le n}\right] = \left[ {\begin{array}{ccccc}
#' D(h_{0})     & D(h_{1})     & \cdots & D(h_{n - 1}) & D(h_{n})     \\
#' D(h_{1})     & D(h_{0})     & \cdots & D(h_{n - 2}) & D(h_{n - 1}) \\
#' \vdots       & \vdots       & \ddots & \vdots       & \vdots       \\
#' D(h_{n - 1}) & D(h_{n - 2}) & \cdots & D(h_{0})     & D(h_{1})     \\
#' D(h_{n})     & D(h_{n - 1}) & \cdots & D(h_{1})     & D(h_{0})     \\
#' \end{array}} \right] ,
#' }
#' over a set of lags \eqn{\{h_{0}, h_{1}, \dots , h_{N} \},} where \eqn{D(h) = \hat{C}_{1}(h) - \hat{C}_{2}(h),}
#' is defined as \deqn{{\left\Vert D \right\Vert}_{HS} = \sqrt{\sum_{i,j} d_{i, j}^{2}}.}
#'
#' @param est1 A numeric vector representing the first estimated autocovariance function.
#' @param est2 A numeric vector of the same length as \code{est1} representing the second estimated autocovariance function
#'
#' @return A numeric value representing the estimated Hilbert-Schmidt norm between two estimated autocovariance functions.
#' @export
#'
#' @examples
#' x <- seq(0, 5, by=0.1)
#' estCov1 <- exp(-x^2)
#' estCov2 <- exp(-x^2.1)
#' hilbert_schmidt(estCov1, estCov2)
hilbert_schmidt <- function(est1, est2) {
  UseMethod("hilbert_schmidt")
}

#' @describeIn hilbert_schmidt Method for `CovEsts` objects.
#' @export
hilbert_schmidt.CovEsts <- function(est1, est2) {
  return(hilbert_schmidt.default(est1$acf, est2$acf))
}

#' @describeIn hilbert_schmidt Method for numeric vectors.
#' @export
hilbert_schmidt.default <- function(est1, est2) {
  stopifnot(is.numeric(est1), is.numeric(est2), length(est1) >= 1, length(est2) >= 1, !any(is.na(est1)), !any(is.na(est2)),
            length(est1) == length(est2))
  return(sqrt(sum(cyclic_matrix(est1 - est2)^2)))
}

#' Compute the Nearest Positive-Definite Matrix.
#'
#' This function computes the nearest positive-definite matrix to some matrix \eqn{A}.
#'
#' @details
#' This function computes the nearest positive-definite matrix to some matrix \eqn{A}.
#'
#' The procedure to do so is as follows
#' For a matrix \eqn{X}, compute the symmetric matrix \eqn{B = (A + A^{T}) / 2.}
#' Let \eqn{B = UH} be the polar decomposition of B.
#' The nearest positive-definite matrix to \eqn{X} is \eqn{X_{F} = (B + H) / 2.}
#'
#' Unlike [shrinking], only an autocorrelation matrix can be returned, not an autocovariance function.
#'
#' The implementation is a translation of https://au.mathworks.com/matlabcentral/fileexchange/42885-nearestspd#functions_tab .
#'
#' @references
#' Higham, N. J. (1988). Computing a nearest symmetric positive semidefinite matrix. Linear Algebra and its Applications, 103, 103-118. https://doi.org/10.1016/0024-3795(88)90223-6
#'
#' D'Errico, J. (2025). nearestSPD (https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd), MATLAB Central File Exchange. Retrieved August 2, 2025.
#'
#' @param X Either a numeric square matrix. If a vector is provided, a matrix will be created of the form found in [cyclic_matrix].
#'
#' @return The closest positive-definite autocorrelation matrix.
#' @export
#'
#' @examples
#' X <- c(1, 0, -1.1)
#' nearest_pd(X)
#' check_pd(nearest_pd(X))
nearest_pd <- function(X) {
  UseMethod("nearest_pd")
}

#' @describeIn nearest_pd Method for `CovEsts` objects.
#' @export
nearest_pd.CovEsts <- function(X) {
  return(nearest_pd.default(X$acf))
}

#' @describeIn nearest_pd Method for numeric vectors.
#' @export
nearest_pd.default <- function(X) {
  stopifnot(is.numeric(X), !any(is.na(X)), length(X) > 0)
  if(!is.matrix(X)) {
    X <- cyclic_matrix(X)
  }
  # https://au.mathworks.com/matlabcentral/fileexchange/42885-nearestspd#functions_tab
  B <- (X + t(X)) / 2

  svd_mat <- svd(X)
  # https://en.wikipedia.org/wiki/Polar_decomposition#Relation_to_the_SVD
  H <- svd_mat$v %*% diag(svd_mat$d) %*% t(svd_mat$v)

  X_F <- (B + H) / 2
  X_F <- (X_F + t(X_F)) / 2

  ident_mat <- diag(nrow(X_F))
  k <- 1
  while(!all(zapsmall(eigen(X_F)$values) >= 0)) {
    min_eig <- min(eigen(X_F)$values)
    X_F <- X_F + (ident_mat * (min_eig * k^2 + abs(min_eig - (min_eig + .Machine$double.eps * min_eig))))
  }
  return(X_F)
}

# area_between
test_that("area_between() works", {
  expect_equal(area_between(test_estCov1, test_estCov2), 2.10491826, tolerance = sqrt(.Machine$double.eps))
})

test_that("area_between() works with a lag vector", {
  expect_equal(area_between(test_estCov1, test_estCov2, test_x_metrics), 0.210491826, tolerance = sqrt(.Machine$double.eps))
})

test_that("area_between() fails for nonnumeric est1", {
  expect_error(area_between(c('a', test_estCov1[-1]), test_estCov2))
  expect_error(area_between(c(1i, test_estCov1[-1]), test_estCov2))
})

test_that("area_between() fails for nonnumeric est2", {
  expect_error(area_between(test_estCov1, c('a', test_estCov2[-1])))
  expect_error(area_between(test_estCov1, c(1i, test_estCov2[-1])))
})

test_that("area_between() fails for empty est1", {
  expect_error(area_between(c(), test_estCov2))
})

test_that("area_between() fails for empty est2", {
  expect_error(area_between(test_estCov1, c()))
})

test_that("area_between() fails for at least on NA in est1", {
  expect_error(area_between(c(NA, test_estCov1[-1]), test_estCov2))
})

test_that("area_between() fails for at least on NA in est2", {
  expect_error(area_between(test_estCov1, c(NA, test_estCov2[-1])))
})

test_that("area_between() fails if est1 is of a different length", {
  expect_error(area_between(test_estCov1[-1], test_estCov2))
  expect_error(area_between(test_estCov1[-1], test_estCov2, test_x_metrics))
})

test_that("area_between() fails if est2 is of a different length", {
  expect_error(area_between(test_estCov1, test_estCov2[-1]))
  expect_error(area_between(test_estCov1, test_estCov2[-1], test_x_metrics))
})

test_that("area_between() fails if plot is not a boolean", {
  expect_error(area_between(test_estCov1, test_estCov2, plot = 1))
  expect_error(area_between(test_estCov1, test_estCov2, plot = 'TRUE'))
})

test_that("area_between() fails if lags is not a vector", {
  expect_error(area_between(test_estCov1, test_estCov2, matrix(c(1,2,3,4), 2)))
})

test_that("area_between() fails if lags is of a different length", {
  expect_error(area_between(test_estCov1, test_estCov2, test_x_metrics[-1]))
})

# max_distance
test_that("max_distance() works", {
  expect_equal(max_distance(test_estCov1, test_estCov2), 0.1818237429, tolerance = sqrt(.Machine$double.eps))
})

test_that("max_distance() fails for nonnumeric est1", {
  expect_error(max_distance(c('a', test_estCov1[-1]), test_estCov2))
  expect_error(max_distance(c(1i, test_estCov1[-1]), test_estCov2))
})

test_that("max_distance() fails for nonnumeric est2", {
  expect_error(max_distance(test_estCov1, c('a', test_estCov2[-1])))
  expect_error(max_distance(test_estCov1, c(1i, test_estCov2[-1])))
})

test_that("max_distance() fails for empty est1", {
  expect_error(max_distance(c(), test_estCov2))
})

test_that("max_distance() fails for empty est2", {
  expect_error(max_distance(test_estCov1, c()))
})

test_that("max_distance() fails for at least on NA in est1", {
  expect_error(max_distance(c(NA, test_estCov1[-1]), test_estCov2))
})

test_that("max_distance() fails for at least on NA in est2", {
  expect_error(max_distance(test_estCov1, c(NA, test_estCov2[-1])))
})

test_that("max_distance() fails if est1 is of a different length", {
  expect_error(max_distance(test_estCov1[-1], test_estCov2))
})

test_that("max_distance() fails if est2 is of a different length", {
  expect_error(max_distance(test_estCov1, test_estCov2[-1]))
})

test_that("max_distance() fails if plot is not a boolean", {
  expect_error(max_distance(test_estCov1, test_estCov2, plot = 1))
  expect_error(max_distance(test_estCov1, test_estCov2, plot = 'TRUE'))
})

# cyclic_matrix
test_that("cyclic_matrix() works", {
  expect_equal(cyclic_matrix(c(1, 2)), matrix(c(1, 2, 2, 1), 2))
})

test_that("cyclic_matrix() fails for nonnumeric v", {
  expect_error(cyclic_matrix(c(1, 'a', 3)))
  expect_error(cyclic_matrix(c(1, 1i, 3)))
})

test_that("cyclic_matrix() fails for at least one NA in v", {
  expect_error(cyclic_matrix(c(1, NA, 3)))
})


test_that("cyclic_matrix() fails if v is empty", {
  expect_error(cyclic_matrix(c()))
})

# spectral_norm
test_that("spectral_norm() works", {
  expect_equal(spectral_norm(test_estCov1, test_estCov2), 1.661283524, tolerance = sqrt(.Machine$double.eps))
})

test_that("spectral_norm() fails for nonnumeric est1", {
  expect_error(spectral_norm(c('a', test_estCov1[-1]), test_estCov2))
  expect_error(spectral_norm(c(1i, test_estCov1[-1]), test_estCov2))
})

test_that("spectral_norm() fails for nonnumeric est2", {
  expect_error(spectral_norm(test_estCov1, c('a', test_estCov2[-1])))
  expect_error(spectral_norm(test_estCov1, c(1i, test_estCov2[-1])))
})

test_that("spectral_norm() fails for empty est1", {
  expect_error(spectral_norm(c(), test_estCov2))
})

test_that("spectral_norm() fails for empty est2", {
  expect_error(spectral_norm(test_estCov1, c()))
})

test_that("spectral_norm() fails for at least on NA in est1", {
  expect_error(spectral_norm(c(NA, test_estCov1[-1]), test_estCov2))
})

test_that("spectral_norm() fails for at least on NA in est2", {
  expect_error(spectral_norm(test_estCov1, c(NA, test_estCov2[-1])))
})

test_that("spectral_norm() fails if est1 is of a different length", {
  expect_error(spectral_norm(test_estCov1[-1], test_estCov2))
})

test_that("spectral_norm() fails if est2 is of a different length", {
  expect_error(spectral_norm(test_estCov1, test_estCov2[-1]))
})

# check_pd
test_that("check_pd() works", {
  expect_equal(check_pd(test_estCov1), TRUE)
})

test_that("check_pd() fails for nonnumeric est", {
  expect_error(check_pd(c(1, 'a', 3)))
  expect_error(check_pd(c(1, 1i, 3)))
})

test_that("check_pd() fails for empty est", {
  expect_error(check_pd(c()))
})

test_that("check_pd() fails for at least on NA in est", {
  expect_error(check_pd(c(1, NA, 3)))
})

# mse
test_that("mse() works", {
  expect_equal(mse(test_estCov1, test_estCov2), 0.013225263, tolerance = sqrt(.Machine$double.eps))
})

test_that("mse() fails for nonnumeric est1", {
  expect_error(mse(c('a', test_estCov1[-1]), test_estCov2))
  expect_error(mse(c(1i, test_estCov1[-1]), test_estCov2))
})

test_that("mse() fails for nonnumeric est2", {
  expect_error(mse(test_estCov1, c('a', test_estCov2[-1])))
  expect_error(mse(test_estCov1, c(1i, test_estCov2[-1])))
})

test_that("mse() fails for empty est1", {
  expect_error(mse(c(), test_estCov2))
})

test_that("mse() fails for empty est2", {
  expect_error(mse(test_estCov1, c()))
})

test_that("mse() fails for at least on NA in est1", {
  expect_error(mse(c(NA, test_estCov1[-1]), test_estCov2))
})

test_that("mse() fails for at least on NA in est2", {
  expect_error(mse(test_estCov1, c(NA, test_estCov2[-1])))
})

test_that("mse() fails if est1 is of a different length", {
  expect_error(mse(test_estCov1[-1], test_estCov2))
})

test_that("mse() fails if est2 is of a different length", {
  expect_error(mse(test_estCov1, test_estCov2[-1]))
})

# hilbert_schmidt
test_that("hilbert_schmidt() works", {
  expect_equal(hilbert_schmidt(test_estCov1, test_estCov2), 2.5411595, tolerance = sqrt(.Machine$double.eps))
})

test_that("hilbert_schmidt() fails for nonnumeric est1", {
  expect_error(hilbert_schmidt(c('a', test_estCov1[-1]), test_estCov2))
  expect_error(hilbert_schmidt(c(1i, test_estCov1[-1]), test_estCov2))
})

test_that("hilbert_schmidt() fails for nonnumeric est2", {
  expect_error(hilbert_schmidt(test_estCov1, c('a', test_estCov2[-1])))
  expect_error(hilbert_schmidt(test_estCov1, c(1i, test_estCov2[-1])))
})

test_that("hilbert_schmidt() fails for empty est1", {
  expect_error(hilbert_schmidt(c(), test_estCov2))
})

test_that("hilbert_schmidt() fails for empty est2", {
  expect_error(hilbert_schmidt(test_estCov1, c()))
})

test_that("hilbert_schmidt() fails for at least on NA in est1", {
  expect_error(hilbert_schmidt(c(NA, test_estCov1[-1]), test_estCov2))
})

test_that("hilbert_schmidt() fails for at least one NA in est2", {
  expect_error(hilbert_schmidt(test_estCov1, c(NA, test_estCov2[-1])))
})

test_that("hilbert_schmidt() fails if est1 is of a different length", {
  expect_error(hilbert_schmidt(test_estCov1[-1], test_estCov2))
})

test_that("hilbert_schmidt() fails if est2 is of a different length", {
  expect_error(hilbert_schmidt(test_estCov1, test_estCov2[-1]))
})

# nearest_pd
test_that("nearest_pd() fails for nonnumeric X", {
  expect_error(nearest_pd(c('a', test_estCov1[-1])))
  expect_error(nearest_pd(c(1i, test_estCov1[-1])))
})

test_that("nearest_pd() fails for at least on NA in X", {
  expect_error(nearest_pd(c(NA, test_estCov1[-1])))
})

test_that("nearest_pd() fails for empty X", {
  expect_error(nearest_pd(c()))
})

#' Generate Spline Knots.
#'
#' A helper function that generates \eqn{m + 2} spline knots of the form:
#' \deqn{
#' \kappa_{0} = 0 , \kappa_{1} = 1 / (m + 1) , \dots , \kappa_{m} = m / (m + 1) , \kappa_{m + 1} = 1 .
#' }
#' The knots are equally spaced with boundary knots \eqn{\kappa_{0} = 0} and \eqn{\kappa_{m + 1} = 1 .}
#'
#' @references Choi, I., Li, B. & Wang, X. (2013). Nonparametric Estimation of Spatial and Space-Time Covariance Function. JABES 18, 611-630. https://doi.org/10.1007/s13253-013-0152-z
#'
#' @param m The number of nonboundary knots.
#'
#' @return A numeric vector representing the knots, including the boundary knots.
#'
#' @examples
#' \dontrun{
#' generate_knots(3)
#' }
generate_knots <- function(m) {
  stopifnot(is.numeric(m), m > 0, m %% 1 == 0)
  knotVec <- seq(0, 1, length.out=m + 2)
  return(knotVec)
}

#' Get all \eqn{\tau}.
#'
#' A helper function to obtain all \eqn{\tau_{i}, i=-p, \dots, m + p + 1,} where each \eqn{\tau_{i}} is as follows.
#' For \eqn{i = -p , -p + 1, \dots , -2, -1 , m + 2, m + 3, \dots , m + p , m + p + 1,} it is equal to \eqn{\tau_{i} = i / (m + 1)}, and for \eqn{i = 0, \dots , m + 1,} it is \eqn{\tau_{i} = \kappa_{i}.}
#' See Choi, Li & Wang (2013, p. 615) for details.
#'
#' @references Choi, I., Li, B. & Wang, X. (2013). Nonparametric Estimation of Spatial and Space-Time Covariance Function. JABES 18, 611-630. https://doi.org/10.1007/s13253-013-0152-z
#'
#' @param p The order of the splines.
#' @param m The number of nonboundary knots.
#'
#' @return A numeric vector of all \eqn{\tau_{i}, i = -p, \dots, m + p + 1.}
#'
#' @examples
#' \dontrun{
#' get_taus(3, 2)
#' }
get_taus <- function(p, m) {
  stopifnot(is.numeric(p), p >= 0, p %% 1 == 0, is.numeric(m), m > 0, m %% 1 == 0)
  tauVec_indices <- -p:(m + p + 1)
  tauVec <- tauVec_indices / (m + 1)
  tauVec[(p + 1):(p + m + 2)] <- generate_knots(m)

  names(tauVec) <- as.character(tauVec_indices)

  return(as.list(tauVec))
}

#' Compute Adjusted Splines.
#'
#' A helper function that is an implementation of the formula from Choi, Li & Wang (2013, p. 616),
#' \deqn{
#' f_{j}^{(l)}(x) = \frac{m + 1}{l} \left( f_{j}^{(l - 1)}(x + 1) - \tau_{j - p} f_{j}^{(l - 1)}(x) + \tau_{j - p + l + 1} f_{j + 1}^{(l - 1)}(x) -  f_{j + 1}^{(l - 1)}(x + 1)  \right) ,
#' }
#' where \eqn{m} is the number of nonboundary knots, \eqn{p} is the order of the spline, \eqn{l} is the order of the adjusted spline (the function \eqn{f_{j}^{(l)}(\cdot)}) and \eqn{j = 1, 2, \dots , m + p.}
#'
#' @references Choi, I., Li, B. & Wang, X. (2013). Nonparametric Estimation of Spatial and Space-Time Covariance Function. JABES 18, 611-630. https://doi.org/10.1007/s13253-013-0152-z
#'
#' @param x Argument of the function.
#' @param j Index of basis function of order \eqn{l}.
#' @param l Order of function.
#' @param p The order of the splines.
#' @param m The number of nonboundary knots.
#' @param taus Vector of \eqn{\tau}s, see [get_taus].
#'
#' @return A numeric value of the adjusted spline \eqn{f_{j}^{(l)}(x).}
#'
#' @examples
#' \dontrun{
#' taus <- get_taus(3, 2)
#' adjusted_spline(1, 2, 1, 3, 2, taus)
#' }
adjusted_spline <- function(x, j, l, p, m, taus) {
  stopifnot(is.numeric(x), j %% 1 == 0, is.numeric(j), j > 0, l %% 1 == 0, is.numeric(l), is.numeric(p), p >= 0, p %% 1 == 0,
            is.numeric(m), m > 0, m %% 1 == 0, length(taus) > 0, all(!is.na(taus)))
  # base case
  if(l == 0) {
    tau1 <- taus[[paste0("", j - p, "")]]
    tau2 <- taus[[paste0("", j - p + 1, "")]]
    if((tau1 >= 0 & tau1 <= 1) & (tau2 >= 0 & tau2 <= 1) ) {
      constant <- (m + 1) / (x + 1)
      inner <- tau2^(x + 1) - tau1^(x + 1)
      return(constant * inner)
    }
    return(numeric(length(x)))
  }
  else {
    constant <- (m + 1) / l
    firstInner <- adjusted_spline(x + 1, j, l - 1, p, m, taus)
    tau_jp <- taus[[paste0("", j - p, "")]]
    secondInner <- tau_jp * adjusted_spline(x, j, l - 1, p, m, taus)

    tau_jpl1 <- taus[[paste0("", j - p + l + 1, "")]]

    thirdInner <- tau_jpl1 * adjusted_spline(x, j + 1, l - 1, p, m, taus)
    fourthInner <- adjusted_spline(x + 1, j + 1, l - 1, p, m, taus)

    retVal <- constant * (firstInner - secondInner + thirdInner - fourthInner)
    return(retVal)
  }
}

#' Construct Data Frame of Basis Functions.
#'
#' This helper function constructs a data frame with the following structure:
#' * One column for the x-values
#' * m + p columns values of squared basis functions evaluated at the corresponding x.
#'
#' @param x A vector of lags.
#' @param p The order of the splines.
#' @param m The number of nonboundary knots.
#' @param taus Vector of \eqn{\tau}s, see [get_taus].
#'
#' @return A data frame of the above structure.
#'
#' @examples
#' \dontrun{
#' taus <- get_taus(3, 2)
#' splines_df(seq(0, 2, by=0.25), 3, 2, taus)
#' }
splines_df <- function(x, p, m, taus) {
  stopifnot(is.numeric(x), length(x) > 0, all(!is.na(x)), is.numeric(p), p >= 0, p %% 1 == 0, is.numeric(m), m > 0, m %% 1 == 0,
            length(taus) > 0, all(!is.na(taus)))

  l <- p - 1
  x_squared <- x * x

  cols <- sapply(1:(m + p), function(j) adjusted_spline(x_squared, j, l, p, m, taus))
  cols <- matrix(cols, nrow = length(x), ncol = (m + p))
  colnames(cols) <- paste0('j', 1:(m + p))

  valsDf <- cbind(data.frame(lags = x), cols)

  return(valsDf)
}

#' Objective Function for WLS.
#'
#' This is the objective function to find the weights for each basis function in the minimising spline, see Choi, Li & Wang (2013, p. 617).
#' The parameters must be nonnegative, so a penalty of \eqn{10^{12}} is given if any parameters are negative.
#' The weights are chosen as per Choi, Li & Wang (2013, p. 617).
#'
#' @details
#' Let \eqn{\mathbf{\beta} = (\beta_{0}, \dots, \beta_{m + p})^{\prime}} be a vector of model coefficients, \eqn{\{f_{1}^{(p - 1)} , \dots , f_{m + p}^{(p - 1)} \}}
#' be a set of completely monotone basis functions, and \eqn{\widehat{C}(\cdot)} be an estimated covariance function.
#' As per Choi, Li & Wang (2013, p. 617), \eqn{\mathbf{\beta}} can be estimated via weighted-least squares,
#' \deqn{\
#' \hat{\mathbf{\beta}}_{WLS} = {\arg\min}_{\beta_{j} \ge 0} \sum_{i=1}^{L} w_{i} \left(\widehat{C}(h_{i}) - \sum_{j = 1}^{m + p} \beta_{j} f_{j}^{(p - 1)}(h_{i}^{2})  \right)^{2} ,
#' }
#' where \eqn{\{h_{1} , \dots , h_{L} \}} is a set of lags and \eqn{\{w_{1}, \dots , w_{L} \}} is a set of weights.
#' The set of weights is calculated in [splines_est], and they are of the form \eqn{w_{i} = (N - h_{i}) / ((1 - \widehat{C}(h_{i}))^{2}).}
#'
#' @references Choi, I., Li, B. & Wang, X. (2013). Nonparametric Estimation of Spatial and Space-Time Covariance Function. JABES 18, 611-630. https://doi.org/10.1007/s13253-013-0152-z
#'
#' @param par A vector of initial parameters to used in the minimisation process.
#' @param splines_df A data frame whose structure is defined in [splines_df].
#' @param weights A vector of weights, see the description.
#'
#' @return The value of the objective function at those parameters.
#'
#' @examples
#' \dontrun{
#' taus <- get_taus(3, 2)
#' x <- seq(0, 2, by=0.25)
#' maxLag <- 4
#' splines_df <- splines_df(x[1:maxLag], 3, 2, taus)
#' splines_df$estCov <- exp(-splines_df$lags^2) + 0.001
#' # pars are the initial parameters used in the minimisation process.
#' pars <- c(0.5, 0.5, 0.5, 0.5, 0.5)
#' weights <- c()
#' X <- rnorm(50)
#' for(i in 0:(maxLag - 1)) {
#'   weights <- c(weights, (length(X) - i) / ( (1 - splines_df$estCov[i + 1])^2 ))
#' }
#' solve_spline(pars, splines_df, weights)
#' }
solve_spline <- function(par, splines_df, weights) {
  if(any(par < 0)) {
    return(10^12)
  }

  splineParVal <- as.numeric(as.matrix(splines_df[, 2:(ncol(splines_df) - 1)]) %*% par)
  return( sum(weights * (splines_df$estCov - splineParVal)^2) )
}

#' Compute the Splines Estimator.
#'
#' Compute the estimated covariance function by using the method from Choi, Li & Wang (2013, pp. 614-617).
#' \deqn{C(\tau) = \sum_{j = 1}^{m + p} \beta_{j} f_{j}^{(p-1)}(\tau^{2}),}
#' where \eqn{m} is the number of nonboundary knots, \eqn{p} is the order of the splines, \eqn{\tau} is the isotropic distance, \eqn{\beta_{j}} are nonnegative weights and \eqn{f_{j}^{(p)}} are basis functions of order \eqn{p.}
#' For optimisation, the Nelder-Mead and L-BFGS-B methods are used, the one which selects parameters which minimises the objective function is chosen.
#'
#' Due to the weighting scheme, the autocovariance at lag zero cannot be 1,
#' \deqn{
#' w_{i} = \frac{N - i}{1 - C(i)}
#' }
#'
#' @references Choi, I., Li, B. & Wang, X. (2013). Nonparametric Estimation of Spatial and Space-Time Covariance Function. JABES 18, 611-630. https://doi.org/10.1007/s13253-013-0152-z
#'
#' @param X A vector representing observed values of the time series.
#' @param x A vector of lag indices.
#' @param estCov An estimated autocovariance function to fit to (a vector).
#' @param p The order of the splines.
#' @param m The number of nonboundary knots.
#' @param maxLag An optional parameter that determines the maximum lag to compute the estimated autocovariance function at. Defaults to \code{length(X) - 1}.
#' @param initial_pars An optional vector of parameters - can be used to fine tune the fit. By default, it is a vector of 0.5 whose length is \eqn{m+p.}
#' @param control An optional list of optimisation parameters used in the optimisation process, see \code{control} in [stats::optim].
#' @param type Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.
#'
#' @return A vector whose values are the spline autocovariance estimates or a \code{CovEsts} S3 object (list) with the following values
#' \describe{
#'  \item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
#'  \item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
#'  \item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the \code{type} parameter.}
#'  \item{\code{est_used}}{The estimator function used, in this case, 'splines_est'.}
#' }
#' If a numeric vector is given for the argument \code{estCov}, then a numeric vector output is given, and if a \code{CovEsts} S3 object is given, a \code{CovEsts} object is given as output.
#'
#' @export
#'
#' @importFrom stats optim
#' @importFrom stats runif
#'
#' @examples
#' X <- rnorm(100)
#' x <- seq(0, 5, by = 0.25)
#' maxLag <- 5
#' estCov <- standard_est(X, maxLag = maxLag)
#' estimated <- splines_est(X, x, estCov, 3, 2, maxLag = maxLag)
#' estimated
splines_est <- function(X, x, estCov, p, m, maxLag = length(X) - 1, type = c("autocovariance", "autocorrelation"), initial_pars = c(), control = list('maxit' = 1000)) {
  UseMethod("splines_est", estCov)
}

#' @describeIn splines_est Method for 'CovEsts' objects.
#' @export
splines_est.CovEsts <- function(X, x, estCov, p, m, maxLag = length(X) - 1, type =  c("autocovariance", "autocorrelation"), initial_pars = c(), control = list('maxit' = 1000)) {
  est <- splines_est.default(X, x, estCov$acf, p, m, maxLag = maxLag, type = type, initial_pars = initial_pars, control = control)

  res <- list(acf = est, lags = x[1:(maxLag + 1)], est_type = type, est_used = 'splines_est')
  return(structure(res, class = "CovEsts"))
}

#' @describeIn splines_est Method for numeric vectors.
#' @export
splines_est.default <- function(X, x, estCov, p, m, maxLag = length(X) - 1, type =  c("autocovariance", "autocorrelation"), initial_pars = c(), control = list('maxit' = 1000)) {
  stopifnot(is.numeric(X), all(!is.na(X)), is.numeric(x), all(!is.na(x)), is.numeric(maxLag),
            maxLag >= 0, maxLag <= (length(estCov) - 1), is.numeric(estCov),
            all(!is.na(estCov)), is.numeric(p), p >= 0, p %% 1 == 0,
            is.numeric(m), m > 0, m %% 1 == 0)

  type <- match.arg(type)

  if(is.numeric(initial_pars) && is.atomic(initial_pars)) {
    stopifnot(length(initial_pars) == p + m)
  }
  taus <- get_taus(p, m)

  splines_df <- splines_df(x[1:(maxLag + 1)], p, m, taus)
  splines_df[, 'estCov'] <- estCov

  weights_indices <- 0:maxLag
  weights <- (length(X) - weights_indices) / ( (1 - estCov[weights_indices + 1])^2 )

  pars <- rep(0.5, ncol(splines_df) - 2)
  if(length(initial_pars) != 0 && length(initial_pars) == ncol(splines_df - 2)) {
    pars <- initial_pars
  }
  # Add two optim results, one for Nelder-Mead and one for BFGS, mimicking what is done in the optimx package.
  optim_splines <- NA
  optim_splines_bfgs <- NA

  optim_splines <- optim(par=pars, fn = solve_spline, splines_df = splines_df, weights = weights, method = 'Nelder-Mead', control = control)
  converge <- optim_splines$convergence == 0

  optim_splines_bfgs <- optim(par=pars, fn = solve_spline, splines_df = splines_df, weights = weights, method = 'L-BFGS-B',
                         lower=rep(0, ncol(splines_df) - 2), control = control)
  converge_bfgs <- optim_splines_bfgs$convergence == 0

  optim_splines_betas <- c()
  # Check if both converge
  if(converge && converge_bfgs) {
    isNelder <- ifelse(optim_splines$value < optim_splines_bfgs$value, TRUE, FALSE)
    if(isNelder) {
      optim_splines_betas <- optim_splines$par
    }
    else {
      optim_splines_betas <- optim_splines_bfgs$par
    }
  }

  # if Nelder-Mead converges
  else if(converge && !converge_bfgs) {
    optim_splines_betas <- optim_splines$par
  }

  # if BFGS converges
  else if(!converge && converge_bfgs) {
    optim_splines_betas <- optim_splines_bfgs$par
  }

  # Neither converge
  else {
    # stop("Failed to converge after five tries.")
    stop("Failed to converge. Try new parameters or increasing the number of iterations.")
  }

  optim_vals <- as.numeric(as.matrix(splines_df[, 2:(ncol(splines_df) - 1)]) %*% optim_splines_betas)

  if(length(optim_vals) != length(estCov)) {
    stop("Length of output differs from input")
  }

  if(type == 'autocorrelation') {
    optim_vals <- optim_vals / optim_vals[1]
  }

  return(optim_vals)
}

# generate_knots()
test_that("generate_knots() works", {
  expect_equal(generate_knots(3), c(0, 1/4, 1/2, 3/4, 1))
})

test_that("generate_knots() fails for nonnumeric m", {
  expect_error(generate_knots('a'))
  expect_error(generate_knots(1i))
})

test_that("generate_knots() fails for m <= 0", {
  expect_error(generate_knots(0))
})

test_that("generate_knots() fails for noninteger m", {
  expect_error(generate_knots(1.1))
})


test_that("get_taus() works", {
  expect_equal(get_taus(test_p, test_m), list('-2' = -1/2, '-1' = -1/4, '0' = 0, '1' = 1/4, '2'= 1/2, '3' = 3/4, '4' = 1, '5' = 5/4, '6' = 3/2))
})

test_that("get_taus() fails for nonnumeric p", {
  expect_error(get_taus('a', test_m))
  expect_error(get_taus(1i, test_m))
})

test_that("get_taus() fails for p < 0", {
  expect_error(get_taus(-0.1, test_m))
})

test_that("get_taus() fails for noninteger p", {
  expect_error(get_taus(1.1, test_m))
})

test_that("get_taus() fails for nonnumeric m", {
  expect_error(get_taus(test_p, 'a'))
  expect_error(get_taus(test_p, 1i))
})

test_that("get_taus() fails for m < 0", {
  expect_error(get_taus(test_p, -0.1))
})

test_that("get_taus() fails for noninteger m", {
  expect_error(get_taus(test_p, 1.1))
})

# adjusted_spline
test_that("adjusted_spline() works", {
  expect_equal(adjusted_spline(1, 1, 1, test_p, test_m, test_taus), 1/24)
})

test_that("adjusted_spline() fails for nonnumeric x", {
  expect_error(adjusted_spline('a', 1, 1, test_p, test_m, test_taus))
  expect_error(adjusted_spline(1i, 1, 1, test_p, test_m, test_taus))
})

test_that("adjusted_spline() fails for noninteger j", {
  expect_error(adjusted_spline(1, 1.1, 1, test_p, test_m, test_taus))
})

test_that("adjusted_spline() fails for nonnumeric j", {
  expect_error(adjusted_spline(1, 'a', 1, test_p, test_m, test_taus))
  expect_error(adjusted_spline(1, 1i, 1, test_p, test_m, test_taus))
})


test_that("adjusted_spline() fails for j <= 0", {
  expect_error(adjusted_spline(1, -1, 1, test_p, test_m, test_taus))
})

test_that("adjusted_spline() fails for noninteger l", {
  expect_error(adjusted_spline(1, 1, 1.1, test_p, test_m, test_taus))
})

test_that("adjusted_spline() fails for nonnumeric l", {
  expect_error(adjusted_spline(1, 1, 'a', test_p, test_m, test_taus))
  expect_error(adjusted_spline(1, 1, 1i, test_p, test_m, test_taus))
})

test_that("adjusted_spline() fails for nonnumeric p", {
  expect_error(adjusted_spline(1, 1, 1, 'a', test_m, test_taus))
  expect_error(adjusted_spline(1, 1, 1, 1i, test_m, test_taus))
})

test_that("adjusted_spline() fails for p < 0", {
  expect_error(adjusted_spline(1, 1, 1, -1, test_m, test_taus))
})

test_that("adjusted_spline() fails for noninteger p", {
  expect_error(adjusted_spline(1, 1, 1, 1.1, test_m, test_taus))
})

test_that("adjusted_spline() fails for nonnumeric m", {
  expect_error(adjusted_spline(1, 1, 1, test_p, 'a', test_taus))
  expect_error(adjusted_spline(1, 1, 1, test_p, 1i, test_taus))
})

test_that("adjusted_spline() fails for m < 0", {
  expect_error(adjusted_spline(1, 1, 1, test_p, -1, test_taus))
})

test_that("adjusted_spline() fails for noninteger m", {
  expect_error(adjusted_spline(1, 1, 1, test_p, 1.1, test_taus))
})

test_that("adjusted_spline() fails for empty taus", {
  expect_error(adjusted_spline(1, 1, 1, test_p, test_m, c()))
})

test_that("adjusted_spline() fails for at least one NA in taus", {
  expect_error(adjusted_spline(1, 1, 1, test_p, test_m, c(1, NA, 2)))
})

# splines_df
test_that("splines_df() fails for nonnumeric x", {
  expect_equal(splines_df(1, 1, 1, list('-1' = -0.5, '0' = 0, '1' = 0.5, '2' = 1, '3' = 1.5)),
               data.frame(lags = 1, j1 = 1/4, j2 = 3/4))
})
test_that("splines_df() fails for nonnumeric x", {
  expect_error(splines_df(c(1, 'a', 3), test_p, test_m, test_taus))
  expect_error(splines_df(c(1, 1i, 3), test_p, test_m, test_taus))
})

test_that("splines_df() fails for empty x", {
  expect_error(splines_df(c(), test_p, test_m, test_taus))
})

test_that("splines_df() fails for at least one NA in x", {
  expect_error(splines_df(c(1, NA, 3), test_p, test_m, test_taus))
})

test_that("splines_df() fails for nonnumeric p", {
  expect_error(splines_df(test_x, 'a', test_m, test_taus))
  expect_error(splines_df(test_x, 1i, test_m, test_taus))
})

test_that("splines_df() fails for p < 0", {
  expect_error(splines_df(test_x, -1, test_m, test_taus))
})

test_that("splines_df() fails for nonninteger p", {
  expect_error(splines_df(test_x, 1.1, test_m, test_taus))
})

test_that("splines_df() fails for nonnumeric m", {
  expect_error(splines_df(test_x, test_p, 'a', test_m, test_taus))
  expect_error(splines_df(test_x, test_p, 1i, test_taus))
})

test_that("splines_df() fails for m < 0", {
  expect_error(splines_df(test_x, test_p, -1, test_taus))
})

test_that("splines_df() fails for nonninteger m", {
  expect_error(splines_df(test_x, test_p, 1.1, test_taus))
})

test_that("splines_df() fails for empty taus", {
  expect_error(splines_df(test_x, test_p, test_m, c()))
})

test_that("splines_df() fails for at least one NA in taus", {
  expect_error(splines_df(test_x, test_p, test_m, c(1, NA, 2)))
})

# splines_est
#test_that("splines_est() works", {
#  expect_equal(splines_est(test_X, test_x, 2, test_estCov, test_p, test_m), c(0.666666184, 0.002083332), tolerance = sqrt(.Machine$double.eps))
#})
# hard to test this function as it is random.

test_that("splines_est() fails for nonnumeric X", {
  expect_error(splines_est(c(1, 'a', 3), test_x, test_estCov, test_p, test_m, maxLag = 2))
  expect_error(splines_est(c(1, 1i, 3), test_x, test_estCov, test_p, test_m, maxLag = 2))
})

test_that("splines_est() fails for at least one NA in X", {
  expect_error(splines_est(c(1, NA, 3), test_x, test_estCov, test_p, test_m, maxLag = 2))
})

test_that("splines_est() fails for nonnumeric x", {
  expect_error(splines_est(test_X, c(1, 'a', 3), test_estCov, test_p, test_m, maxLag = 2))
  expect_error(splines_est(test_X, c(1, 1i, 3), test_estCov, test_p, test_m, maxLag = 2))
})

test_that("splines_est() fails for at least one NA in x", {
  expect_error(splines_est(test_X, c(1, NA, 3), test_estCov, test_p, test_m, maxLag = 2))
})

test_that("splines_est() fails for nonnumeric maxLag", {
  expect_error(splines_est(test_X, test_x, test_estCov, test_p, test_m, maxLag = 'a'))
  expect_error(splines_est(test_X, test_x, test_estCov, test_p, test_m, maxLag = 1i))
})

test_that("splines_est() fails for maxLag < 0", {
  expect_error(splines_est(test_X, test_x, test_estCov, test_p, test_m, maxLag =-1))
})

test_that("splines_est() fails for maxLag >= length(X)", {
  expect_error(splines_est(test_X, test_x, test_estCov, test_p, test_m, maxLag = length(test_X)))
})

test_that("splines_est() fails for nonnumeric estCov", {
  expect_error(splines_est(test_X, test_x, c(1, 'a', 3), test_p, test_m, maxLag = 2))
  expect_error(splines_est(test_X, test_x, c(1, 1i, 3), test_p, test_m, maxLag = 2))
})

test_that("splines_est() fails for at least one NA in estCov", {
  expect_error(splines_est(test_X, test_x, c(1, NA, 3), test_p, test_m, maxLag = 2))
})

test_that("splines_est() fails when length(estCov) != maxLag", {
  expect_error(splines_est(test_X, test_x, c(1, 2, 3, 4), test_p, test_m, maxLag = 2))
})

test_that("splines_est() fails for nonnumeric p", {
  expect_error(splines_est(test_X, test_x, test_estCov, 'a', test_m, maxLag = 2))
  expect_error(splines_est(test_X, test_x, test_estCov, 1i, test_m, maxLag = 2))
})

test_that("splines_est() fails for p < 0", {
  expect_error(splines_est(test_X, test_x, test_estCov, -1, test_m, maxLag = 2))
})

test_that("splines_est() fails for noninteger p", {
  expect_error(splines_est(test_X, test_x, test_estCov, 1.1, test_m, maxLag = 2))
})

test_that("splines_est() fails for nonnumeric m", {
  expect_error(splines_est(test_X, test_x, test_estCov, test_p, 'a', maxLag = 2))
  expect_error(splines_est(test_X, test_x, test_estCov, test_p, 1i, maxLag = 2))
})

test_that("splines_est() fails for m < 0", {
  expect_error(splines_est(test_X, test_x, test_estCov, test_p, -1, maxLag = 2))
})

test_that("splines_est() fails for noninteger m", {
  expect_error(splines_est(test_X, test_x, Test_estCov, test_p, 1.1, maxLag = 2))
})

test_that("splines_est() fails for length(initial_pars) !=  m + p", {
  expect_error(splines_est(test_X, test_x, test_estCov, test_p, test_m, initial_pars = c(1), maxLag = 2))
})

#' Plot Method for CovEsts Objects
#'
#' @description
#' This function plots a CovEsts object, over the lags used to estimate the autocovariance values.
#'
#' @param x A CovEst S3 object.
#' @param type Defaults to \code{'l'}, see [base::plot] for the possible values.
#' @param xlab Defaults to \code{'Lag (h)'}, see [graphics::title] for more information.
#' @param ylab Defaults to \code{'ACF'}, see [graphics::title] for more information.
#' @param ... Additional plotting arguments, refer to [graphics::par].
#'
#' @return A plot of a CovEsts S3 object.
#' @export
#'
#' @examples
#' plot(standard_est(c(1, 2, 3)))
plot.CovEsts <- function(x, type = 'l', xlab = 'Lag (h)', ylab = 'ACF', ...) {
  # add logic to see if it is autocorrelation, autocovariance, partial
  # change plot y-axis accordingly.
  plot(x = x$lags, y = x$acf, type = type, xlab = xlab, ylab = ylab, ...)
}

#' Lines Method for CovEsts Objects
#'
#' @description
#' This function plots a CovEsts object as a line onto another plot.
#'
#' @param x A CovEst S3 object.
#' @param type Defaults to \code{'l'}, see [base::plot] for the possible values.
#' @param ... Additional plotting arguments, refer to [graphics::par].
#'
#' @return A line of a CovEsts S3 object.
#' @export
#'
#' @examples
#' plot(standard_est(c(1, 2, 3)))
#' lines(standard_est(c(1, 2, 3), pd = FALSE))
lines.CovEsts <- function(x, type = "l", ...) {
  lines(x = x$lags, y = x$acf,  type = type, ...)
}

#' Print Method for CovEsts Objects
#'
#' @description This function plots a CovEsts object, printing the estimator
#' type, estimator used, lags and values.
#'
#' @param x A CovEst S3 object.
#' @param n_head The number of 'head' values to print, defaults to 5.
#' @param n_tail The number of 'tail' values to print, defaults to 5.
#' @param digits The number of digits to print. Defaults to \code{NULL}. Note, this value is passed into [base::format].
#' @param ... Addition printing parameters, see [base::print].
#'
#' @return Prints a CovEsts object.
#' @export
#'
#' @importFrom utils head
#' @importFrom utils tail
#'
#' @examples
#' print(standard_est(c(1, 2, 3)))
print.CovEsts <- function(x, n_head = 5, n_tail = 5, digits = NULL, ...) {
  cat("\n Estimated (Partial) Autocorrelation/Autocovariance\n\n")
  cat("Type:      ", x$est_type, "\n")
  cat("Estimator: ", x$est_used, "\n")

  full_df <- format(data.frame(Lag = x$lags, Estimate = x$acf), digits = digits)

  if(length(x$acf) <= (n_head + n_tail)) {
    print(full_df, row.names = FALSE, right = FALSE)
  }

  else {
    head_df <- utils::head(full_df, n_head)
    dots_df <- data.frame(Lag = "...", Estimate = "...")
    tail_df <- utils::tail(full_df, n_tail)

    print(rbind(head_df, dots_df, tail_df), row.names = FALSE, right = FALSE)
  }

  cat("\n")
  invisible(x)
}

#' Plot Method for VarioEsts Objects
#'
#' @description
#' This function plots a VarioEsts object, over the lags used to estimate the variogram values.
#'
#' @param x A VarioEsts S3 object.
#' @param type Defaults to \code{'l'}, see [base::plot] for the possible values.
#' @param xlab Defaults to \code{'Lag (h)'}, see [graphics::title] for more information.
#' @param ylab Defaults to \code{expression(gamma(h))}, see [graphics::title] for more information.
#' @param ... Additional plotting arguments, refer to [graphics::par].
#'
#' @return A plot of a VarioEsts S3 object.
#' @export
#'
#' @examples
#' plot(to_vario(standard_est(c(1, 2, 3))))
plot.VarioEsts <- function(x, type = 'l', xlab = 'Lag (h)', ylab = expression(gamma(h)), ...) {
  plot(x = x$lags, y = x$vario, type = type, xlab = xlab, ylab = ylab, ...)
}

#' Lines Method for VarioEsts Objects
#'
#' @description
#' This function plots a VarioEsts object as a line onto another plot.
#'
#' @param x A VarioEsts S3 object.
#' @param type Defaults to \code{'l'}, see [base::plot] for the possible values.
#' @param ... Additional plotting arguments, refer to [graphics::par].
#'
#' @return A line of a VarioEsts S3 object.
#' @export
#'
#' @examples
#' plot(to_vario(standard_est(c(1, 2, 3))))
#' lines(to_vario(standard_est(c(1, 2, 3), pd = FALSE)))
lines.VarioEsts <- function(x, type = "l", ...) {
  lines(x = x$lags, y = x$vario, type = type, ...)
}

#' Print Method for VarioEsts Objects
#'
#' @description This function plots a VarioEsts object, printing the lags and values.
#'
#' @param x A VarioEsts S3 object.
#' @param n_head The number of 'head' values to print, defaults to 5.
#' @param n_tail The number of 'tail' values to print, defaults to 5.
#' @param digits The number of digits to print. Defaults to \code{NULL}. Note, this value is passed into [base::format].
#' @param ... Addition printing parameters, see [base::print].
#'
#' @return Prints a VarioEsts object.
#' @export
#'
#' @importFrom utils head
#' @importFrom utils tail
#'
#' @examples
#' print(to_vario(standard_est(c(1, 2, 3))))
print.VarioEsts <- function(x, n_head = 5, n_tail = 5, digits = NULL, ...) {
  cat("\n Estimated Variogram\n\n")

  full_df <- format(data.frame(Lag = x$lags, Estimate = x$vario), digits = digits)

  if(length(x$vario) <= (n_head + n_tail)) {
    print(full_df, row.names = FALSE, right = FALSE)
  }

  else {
    head_df <- utils::head(full_df, n_head)
    dots_df <- data.frame(Lag = "...", Estimate = "...")
    tail_df <- utils::tail(full_df, n_tail)

    print(rbind(head_df, dots_df, tail_df), row.names = FALSE, right = FALSE)
  }

  cat("\n")
  invisible(x)
}

#' Plot Method for BootEsts Objects
#'
#' @description
#' This function plots a BootEsts object. It plots the confidence region, average bootstrap estimate, and the original nonbootstrapped estimate.
#'
#' @param x A BootEsts S3 object.
#' @param type Defaults to \code{'l'}, see [base::plot] for the possible values.
#' @param xlab Defaults to \code{'Lag (h)'}, see [graphics::title] for more information.
#' @param ylab Defaults to \code{'ACF'}, see [graphics::title] for more information.
#' @param ... Additional plotting arguments, refer to [graphics::par].
#'
#' @return A plot of a BootEsts S3 object.
#' @export
#' @importFrom grDevices rgb
#'
#' @examples
#' plot(block_bootstrap(c(1, 2, 3), 2))
plot.BootEsts <- function(x, type = 'l', xlab = 'Lag (h)', ylab = 'ACF', ...) {
  plot(x$lags, x$acf_orig, type = "n", xlab = 'Lag (h)', ylab = 'ACF', ...)
  polygon(c(rev(x$lags), x$lags), c(rev(x$conf_upper), x$conf_lower), col = rgb(0.9, 0.9, 0.9, 0.6))
  lines(x$lags, x$acf_orig, type = "l", lwd = 2)
  lines(x$lags, x$acf_avg, lty = 2, col = 2, lwd = 2)

  graphics::legend("topright", inset=0, title=paste0(toupper(substring(x$boot_type, 1, 1)), substring(x$boot_type, 2), " Block Bootstrap"),
                   c("Original Estimate", "Bootstrapped Average", paste0((1 - x$alpha) * 100, "% Confidence Region")),
                   col=c(1, 2, rgb(0.9, 0.9, 0.9, 0.6)), lwd=c(2, 2, NA), lty=c(1, 2, NA), density=c(0, 0, NA), fill=c(NA, NA, "#AAAAAAAA"), border=c(NA, NA, "#000000FF"),
                   y.intersp=1, cex=1)
}

#' Lines Method for BootEsts Objects
#'
#' @description
#' This function plots a BootEsts object as a line onto another plot.
#'
#' @param x A BootEsts S3 object.
#' @param type Defaults to \code{'l'}, see [base::plot] for the possible values.
#' @param ... Additional plotting arguments, refer to [graphics::par].
#'
#' @return A line of a BootEsts S3 object.
#' @export
#'
#' @examples
#' plot(block_bootstrap(c(1, 2, 3), 2))
#' lines(block_bootstrap(c(1, 2, 3), 2, pd = FALSE), lwd = 2, lty = 3, col = 'blue')
lines.BootEsts  <- function(x, type = "l", ...) {
  lines(x = x$lags, y = x$acf_avg, type = type, ...)
}

#' Print Method for BootEsts Objects
#'
#' @description This function plots a BootEsts object, printing estimtor type, bootstrap type, estimator used, the bootstrap average estimate, the original estimate,
#' and the pointwise confidence interval.
#'
#' @param x A BootEsts S3 object.
#' @param n_head The number of 'head' values to print, defaults to 5.
#' @param n_tail The number of 'tail' values to print, defaults to 5.
#' @param digits The number of digits to print. Defaults to \code{NULL}. Note, this value is passed into [base::format].
#' @param ... Addition printing parameters, see [base::print].
#'
#' @return Prints a BootEsts object.
#' @export
#'
#' @importFrom utils head
#' @importFrom utils tail
#'
#' @examples
#' print(block_bootstrap(c(1, 2, 3), 2))
print.BootEsts <- function(x, n_head = 5, n_tail = 5, digits = NULL, ...) {
  cat("\n Estimated Bootstrap Autocorrelation/Autocovariance\n\n")
  cat("Type:      ", x$est_type, "\n")
  cat("Boot Type: ", x$boot_type, "\n")
  cat("Estimator: ", x$est_used, "\n")

  avg_df <- format(data.frame(Lag = x$lags, Estimate = x$acf_avg), digits = digits)
  orig_df <- format(data.frame(Lag = x$lags, Estimate = x$acf_orig), digits = digits)
  conf_df <- format(data.frame(Lag = x$lags, Conf_Lower = x$conf_lower, Conf_Upper = x$conf_upper), digits = digits)

  if(length(x$acf_avg) <= (n_head + n_tail)) {
    cat("Average ACF boostrap estimate\n")
    print(avg_df, row.names = FALSE, right = FALSE)
    cat("\nOriginal ACF estimate\n")
    print(orig_df, row.names = FALSE, right = FALSE)
    cat("\nBootstrap pointwise 1 - ", x$alpha, "% confidence interval\n", sep = '')
    print(conf_df, row.names = FALSE, right = FALSE)
  }

  else {
    cat("Average ACF boostrap estimate\n")
    head_df <- utils::head(avg_df, n_head)
    dots_df <- data.frame(Lag = "...", Estimate = "...")
    tail_df <- utils::tail(avg_df, n_tail)
    print(rbind(head_df, dots_df, tail_df), row.names = FALSE, right = FALSE)

    cat("\nOriginal ACF estimate\n")
    head_df <- utils::head(orig_df, n_head)
    dots_df <- data.frame(Lag = "...", Estimate = "...")
    tail_df <- utils::tail(orig_df, n_tail)
    print(rbind(head_df, dots_df, tail_df), row.names = FALSE, right = FALSE)

    cat("\nBootstrap pointwise 1 - ", x$alpha, "% confidence interval\n", sep = "")
    head_df <- utils::head(conf_df, n_head)
    dots_df <- data.frame(Lag = "...", Conf_Lower = "...", Conf_Upper = "...")
    tail_df <- utils::tail(conf_df, n_tail)
    print(rbind(head_df, dots_df, tail_df), row.names = FALSE, right = FALSE)
  }
  cat("\n")
  invisible(x)
}


#' as.double Method for CovEsts Objects
#'
#' @description
#' This function provides a method for as.double for CovEsts objects, allowing for as.numeric to be called.
#' It extracts the acf values from the object.
#'
#' @param x A CovEsts S3 object.
#' @param ... Refer to [base::double].
#'
#' @return The acf values from the CovEsts object.
#' @export
#'
#' @examples
#' as.numeric(standard_est(c(1, 2, 3)))
as.double.CovEsts <- function(x, ...) {
  return(x$acf)
}

#' Compute Normalisation Factor
#'
#' This helper function is used in the computation of the normalisation factor the function [tapered_est],
#' \deqn{H_{2, n}(0) = \sum_{s=1}^{n} a((s - 1/2) / n; \rho)^{2}, }
#' where \eqn{a(\cdot; \cdot)} is a window function.
#'
#' @param n The sample size.
#' @param rho A scale parameter in \eqn{(0, 1].}
#' @param window_name The name of the [window_ec] function to be used. Possible values are:
#' tukey, triangular, power_sine, blackman_window, hann_poisson, welch. Alternatively, a custom window function can be provided, see the example for [taper].
#' @param window_params A vector of parameters of the window function.
#' @param custom_window If a custom window is to be used or not. Defaults to \code{FALSE}.
#'
#' @return A single value being \eqn{H_{2, n}(0)}.
#'
#' @examples
#' \dontrun{
#' H2n(3, 0.6, "tukey")
#' }
H2n <- function(n, rho, window_name = c("tukey", "triangular", "sine", "power_sine", "blackman", "hann_poisson", "welch"), window_params=c(1), custom_window = FALSE) {
  stopifnot(is.numeric(n), n >= 1, n %% 1 == 0, is.numeric(rho), rho > 0, rho <= 1, length(rho) == 1, is.logical(custom_window))
  window_name <- match.arg(window_name)
  sSeq <- 1:n
  hSeq <- taper(((sSeq - 1/2) / n), rho, window_name, window_params, custom_window)
  return(sum(hSeq))
}

#' Compute the Function \eqn{a(x; \rho).}
#'
#' This helper function computes the taper function for a given window function as
#' \deqn{a(x; \rho) = \left\{
#' \begin{array}{ll}
#'    w(2x/\rho)     & 0 \leq x < \frac{1}{2} \rho, \\
#'    1              & \frac{1}{2}\rho \leq x \leq \frac{1}{2} \\
#'    a(1 - x; \rho) & \frac{1}{2} < x \leq 1
#' \end{array} ,
#' \right. }
#' where \eqn{w(\cdot)} is a continuous increasing function with \eqn{w(0)=0, w(1)=1,}
#' \eqn{\rho \in (0, 1],} and \eqn{x \in [0, 1].} The possible window function choices are found in [window_ec].
#'
#' @param x A vector of numbers between 0 and 1 (inclusive).
#' @param rho A scale parameter in \eqn{(0, 1].}
#' @param window_name The name of the [window_ec] function to be used. Possible values are:
#' tukey, triangular, power_sine, blackman_window, hann_poisson, welch. Alternatively, a custom window function can be provided, see the example.
#' @param window_params A vector of parameters of the window function.
#' @param custom_window If a custom window is to be used or not. Defaults to \code{FALSE}.
#'
#' @return A vector of taper values.
#' @export
#'
#' @examples
#' X <- c(0.1, 0.2, 0.3)
#' taper(X, 0.5, "tukey")
#' curve(taper(x, 1, "tukey"), from = 0, to = 1)
#' curve(taper(x, 1, "power_sine", c(4)), from = 0, to = 1)
#' my_taper <- function(x, ...) {
#'   return(x)
#' }
#' taper(X, 0.5, my_taper, custom_window = TRUE)
taper <- function(x, rho, window_name = c("tukey", "triangular", "sine", "power_sine", "blackman", "hann_poisson", "welch"), window_params=c(1), custom_window = FALSE) {
  stopifnot(is.numeric(x), length(x) >= 1, !any(is.na(x)), is.numeric(rho),
            length(rho) == 1, rho > 0, rho <= 1, is.logical(custom_window))
  # after 1/2 , we have a(1-x), treat it as that.
  x_new <- pmin(x, 1 - x)

  # Check which are not in [rho/2, 1/2] : values in this are mapped to 1.
  x_new_idx <- x_new < (rho / 2)

  # allocate 1 as we return it for non in x_new_idx
  retVec <- rep(1, length(x))

  if(custom_window) {
    stopifnot(is.function(window_name))
    retVec[x_new_idx] <- window_name((2 * x_new[x_new_idx]) / rho, window_params)
  }

  else {
    window_name <- match.arg(window_name)
    retVec[x_new_idx] <- window_ec((2 * x_new[x_new_idx]) / rho, window_name, c(window_params[1]))
  }
  return(retVec)
}

#' Compute the Estimated Tapered Autocovariance Function over a Set of Lags.
#'
#' This function computes the tapered autocovariance over a set of lags.
#'
#' @details
#' This function computes the estimated tapered autocovariance over a set of lags,
#' \deqn{\widehat{C}_{N}^{a} (h) = (H_{2, n}(0))^{-1} \sum_{j=1}^{N-h}  (X(j) - \bar{X} ) ( X(j + h) - \bar{X} ) a((j - 1/2) / N; \rho) a((j + h - 1/2) / N; \rho) ,}
#' where \eqn{a(\cdot)} is a window function, \eqn{\rho \in (0, 1]}  is a scale parameter.
#' This estimator takes into account the edge effect during estimation, assigning a lower weight to values closer to the boundaries and higher weights for observations closer to the middle.
#' This estimator is positive-definite and asymptotically unbiased.
#'
#' @references
#' Dahlhaus R. & Künsch, H. (1987). Edge Effects and Efficient Parameter Estimation for Stationary Random Fields. Biometrika 74(4), 877-882. 10.1093/biomet/74.4.877
#'
#' @param X A vector representing observed values of the time series.
#' @param rho A scale parameter in \eqn{(0, 1].}
#' @param window_name The name of the [window_ec] function to be used. Possible values are:
#' tukey, triangular, sine, power_sine, blackman_window, hann_poisson, welch. Alternatively, a custom window_ec function can be provided, see the example in [taper].
#' @param window_params A vector of parameters of the window function.
#' @param custom_window If a custom window is to be used or not. Defaults to \code{FALSE}.
#' @param maxLag An optional parameter that determines the maximum lag to compute the estimated autocovariance function at. Defaults to \code{length(X) - 1}.
#' @param x A vector of lag indices. Defaults to the sequence \code{0:length(X)}. Must be at least as large as \code{maxLag + 1}.
#' @param type Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.
#' @param meanX The average value of \code{X}. Defaults to \code{mean(X)}.
#'
#' @return A \code{CovEsts} S3 object (list) with the following values
#' \describe{
#'  \item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
#'  \item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
#'  \item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the \code{type} parameter.}
#'  \item{\code{est_used}}{The estimator function used, in this case, 'tapered_est'.}
#' }
#' @export
#'
#' @examples
#' X <- c(1, 2, 3)
#' tapered_est(X, 0.5, "tukey", maxLag = 2)
tapered_est <- function(X, rho, window_name = c("tukey", "triangular", "sine", "power_sine", "blackman", "hann_poisson", "welch"), window_params = c(1), maxLag = length(X) - 1, x = 0:length(X), type = c("autocovariance", "autocorrelation"), meanX = mean(X), custom_window = FALSE) {
  stopifnot(is.numeric(X), length(X) >= 1, !any(is.na(X)), is.numeric(maxLag), length(maxLag) == 1,
            maxLag > 0, maxLag <= (length(X) - 1), maxLag %% 1 == 0, is.numeric(rho), length(rho) == 1, rho > 0, rho <= 1,
            is.logical(custom_window),  length(meanX) == 1, is.numeric(meanX), !is.na(meanX),
            is.numeric(x), all(!is.na(x)), length(x) > 0)

  type <- match.arg(type)

  h2n <- H2n(length(X), rho, window_name, window_params, custom_window)
  covVals <- rep(0, maxLag + 1)

  n <- length(X)
  taperVals_t <- taper(((1:n) - 1/2)/n, rho, window_name, window_params, custom_window)

  # auxiliary variable, compute the ACF of this new sequence instead.
  Y <- (X - mean(X)) * taperVals_t

  # compute the PD estimated ACF of Y, rescale to 1/h2n
  covVals <- standard_est(Y, maxLag = maxLag, type = type, meanX = 0)$acf * n / h2n

  if(type == 'autocorrelation') {
    covVals <- covVals / covVals[1]
  }

  res <- list(acf = covVals, lags = x[1:(maxLag + 1)], est_type = type, est_used = 'tapered_est')
  return(structure(res, class = "CovEsts"))
}

# H2n
test_that("H2n(, , \"tukey\") works", {
  expect_equal(H2n(3, 1, "tukey"), 1.5)
})

test_that("H2n(, , \"tukey\") fails for nonnumeric n", {
  expect_error(H2n('a', 1, "tukey"))
  expect_error(H2n(1i, 1, "tukey"))
})

test_that("H2n(, , \"tukey\") fails for n < 1", {
  expect_error(H2n(0, 1, "tukey"))
})

test_that("H2n(, , \"tukey\") fails for noninteger n", {
  expect_error(H2n(2.95, 1, "tukey"))
})

test_that("H2n() fails for nonnumeric rho", {
  expect_error(H2n(3, 'a', "tukey"))
  expect_error(H2n(3, 1i, "tukey"))
})

test_that("H2n() fails for negative rho", {
  expect_error(H2n(3, -0.01, "tukey"))
})

test_that("H2n() fails for rho > 1", {
  expect_error(H2n(3, 1.01, "tukey"))
})

test_that("H2n() fails for rho that is not a single number", {
  expect_error(H2n(0.5, c(1, 2), "tukey"))
})

test_that("H2n() fails for window_name that is not one of 'tukey', 'triangular', 'power_sine', 'blackman_window', 'hann_poisson', 'welch'", {
  expect_error(H2n(3, 1, "window"))
})

test_that("H2n() fails for nonboolean custom_window", {
  expect_error(H2n(3, 1, "window", custom_window = 1))
  expect_error(H2n(3, 1, "window", custom_window = 'TRUE'))
})

# taper
test_that("taper(, , \"tukey\") works", {
  expect_equal(taper(c(0.25, 0.5, 0.75), 1, "tukey"), c(0.5, 1, 0.5))
})

test_that("taper(, , \"tukey\") fails for nonnumeric x", {
  expect_error(taper(c(0.25, 'a', 0.75), 1, "tukey"))
  expect_error(taper(c(0.25, 1i, 0.75), 1, "tukey"))
})

test_that("taper(, , \"tukey\") fails for empty x", {
  expect_error(taper(c(), 1, "tukey"))
})

test_that("taper(, , \"tukey\") fails for any NAs in x", {
  expect_error(taper(c(0.25, NA, 0.75), 1, "tukey"))
})

test_that("taper() fails for nonnumeric rho", {
  expect_error(taper(3, 'a', "tukey"))
  expect_error(taper(3, 1i, "tukey"))
})

test_that("taper() fails for negative rho", {
  expect_error(taper(3, -0.01, "tukey"))
})

test_that("taper() fails for rho > 1", {
  expect_error(taper(3, 1.01, "tukey"))
})

test_that("taper() fails for rho that is not a single number", {
  expect_error(taper(0.5, c(1, 2), "tukey"))
})

test_that("taper() fails for nonboolean custom_window", {
  expect_error(taper(3, 1, "window", custom_window = 1))
  expect_error(taper(3, 1, "window", custom_window = 'TRUE'))
})

# tapered_est
test_that("tapered_est(, , , \"tukey\") works", {
  expect_equal(tapered_est(test_X, 1, "tukey", maxLag = 2)$acf, c(1/12, 0, -1/24))
})

test_that("tapered_est() fails for nonnumeric X", {
  expect_error(tapered_est(c(1, 'a', 3), 1, "tukey", maxLag = 2))
  expect_error(tapered_est(c(1, 1i, 3), 1, "tukey", maxLag = 2))
})

test_that("tapered_est() fails for empty X", {
  expect_error(tapered_est(c(), 1, "tukey", maxLag = 2))
})

test_that("tapered_est() fails for at least one NA in X", {
  expect_error(tapered_est(c(1, NA, 3), 1, "tukey", maxLag = 2))
})

test_that("tapered_est() fails for at nonnumeric maxLag", {
  expect_error(tapered_est(test_X, 1, "tukey", maxLag = 'a'))
  expect_error(tapered_est(test_X, 1, "tukey", maxLag = 1i))
})

test_that("tapered_est() fails for maxLag whose length is not 1", {
  expect_error(tapered_est(test_X, 1, "tukey", maxLag = c()))
  expect_error(tapered_est(test_X, 1, "tukey", maxLag = c(1, 2)))
})

test_that("tapered_est() fails for maxLag less than 1", {
  expect_error(tapered_est(test_X, 1, "tukey", , maxLag = 0))
})

test_that("tapered_est() fails for maxLag greater than or equal to length(X)", {
  expect_error(tapered_est(test_X, 1, "tukey", maxLag = 3))
})

test_that("tapered_est() fails for noninteger maxLag", {
  expect_error(tapered_est(test_X, 1, "tukey", maxLag = 1.2))
})

test_that("tapered_est() fails for nonnumeric rho", {
  expect_error(tapered_est(test_X, 'a', "tukey", maxLag = 2))
  expect_error(tapered_est(test_X, 1i, "tukey", maxLag = 2))
})

test_that("tapered_est() fails for rho not of length 1", {
  expect_error(tapered_est(test_X, c(), "tukey", maxLag = 2))
  expect_error(tapered_est(test_X, c(1, 2), "tukey", maxLag = 2))
})

test_that("tapered_est() fails for rho not in (0, 1]", {
  expect_error(tapered_est(test_X, -0.01, "tukey", maxLag = 2))
  expect_error(tapered_est(test_X, 1.01, "tukey", maxLag = 2))
})

test_that("tapered_est() fails for nonboolean custom_window", {
  expect_error(tapered_est(test_X, 1, "tukey", custom_window = 1, , maxLag = 2))
  expect_error(tapered_est(test_X, 1, "tukey", custom_window = 'TRUE', maxLag = 2))
})

test_that("tapered_est() fails for nonnumeric meanX", {
  expect_error(tapered_est(c(1, 2, 3), 1, "tukey", maxLag = 2, meanX = 'a'))
  expect_error(tapered_est(c(1, 2, 3), 1, "tukey", maxLag = 2, meanX = 1i))
})

test_that("tapered_est() fails for NA meanX", {
  expect_error(tapered_est(c(1, 2, 3), 1, "tukey", maxLag = 2, meanX = NA))
})

test_that("tapered_est() fails for nonnumeric x", {
  expect_error(tapered_est(c(1, 2, 3), 1, "tukey", maxLag = 2, x = c(1, 'a', 3)))
  expect_error(tapered_est(c(1, 2, 3), 1, "tukey", maxLag = 2, x = c(1, 1i, 3)))
})

test_that("tapered_est() fails for empty x", {
  expect_error(tapered_est(c(1, 2, 3), 1, "tukey", maxLag = 2, x = c()))
})

test_that("tapered_est() fails for at least one NA in x", {
  expect_error(tapered_est(c(1, 2, 3), 1, "tukey", maxLag = 2, x = c(1, NA, 3)))
})

#' Deprecated functions in package \pkg{CovEsts}.
#'
#' These functions are provided for compatibility with older verisons of the \pkg{CovEsts} package.
#' These functions may be removed entirely.
#' @rdname CovEsts-deprecated
#' @name CovEsts-deprecated
#' @keywords internal
#'
#' @param x,name,params Parameters passed to the updated functions.
#'
#' @section \code{kernel}:
#' For \code{kernel()}, use \code{\link{kernel_ec}} instead.
#' @section \code{kernel_symm}:
#' For \code{kernel_symm()}, use \code{\link{kernel_symm_ec}} instead.
#' @section \code{window}:
#' For \code{window()}, use \code{\link{window_ec}} instead.
#' @section \code{window_symm}:
#' For \code{window_symm()}, use \code{\link{window_symm_ec}} instead.
#'
NULL

#' @export
#' @rdname CovEsts-deprecated
kernel <- function(x, name, params=c(1)) {
  .Deprecated("kernel_ec", "CovEsts")
  kernel_ec(x = x, name = name, params = params)
}

#' @export
#' @rdname CovEsts-deprecated
kernel_symm <- function(x, name, params=c(1)) {
  .Deprecated("kernel_symm_ec", "CovEsts")
  kernel_symm_ec(x = x, name = name, params = params)
}

#' @export
#' @rdname CovEsts-deprecated
window <- function(x, name, params=c(1)) {
  .Deprecated("window_ec", "CovEsts")
  window_ec(x = x, name = name, params = params)
}

#' @export
#' @rdname CovEsts-deprecated
window_symm <- function(x, name, params=c(1)) {
  .Deprecated("window_symm_ec", "CovEsts")
  window_symm_ec(x = x, name = name, params = params)
}

#' Compute \eqn{X_{ij}} Matrix
#'
#' This helper function computes the matrix of pairwise values, \eqn{X_{ij},} for the kernel regression estimator,
#' \deqn{X_{ij} = (X_{i} - \bar{X}) (X_{j} - \bar{X}) .}
#'
#' @references
#' Hall, P. & Patil, P. (1994). Properties of nonparametric estimators of autocovariance for stationary random fields. Probability Theory and Related Fields 99(3), 399-424. https://doi.org/10.1007/bf01199899
#'
#' Hall, P., Fisher, N. I., & Hoffmann, B. (1994). On the nonparametric estimation of covariance functions. The Annals of Statistics 22(4), 2115-2134. https://doi.org/10.1214/aos/1176325774
#'
#' @param X A vector of values.
#' @param meanX The average value of \code{X}. Defaults to \code{mean(X)}.
#'
#' @return A matrix of size \eqn{N \times N}, where \eqn{N} is the length of the vector X.
#'
#' @examples
#' \dontrun{
#' X <- c(1, 2, 3, 4)
#' Xij_mat(X, mean(X))
#' }
Xij_mat <- function(X, meanX = mean(X)) {
  stopifnot(is.numeric(X), length(X) >= 1, !any(is.na(X)), length(meanX) == 1, is.numeric(meanX), !is.na(meanX))
  return((X - meanX) %o% (X - meanX))
}

#' Compute \eqn{\rho(T_{1})} used in the Truncated Kernel Regression Estimator.
#'
#' This helper function computes \eqn{\rho(T_{1})} used in the truncated kernel regression estimator, [truncated_est].
#'
#' @details
#' This function computes the following value,
#' \deqn{
#' \hat{\rho}(T_{1}) = \left( \sum_{i=1}^{N} \sum_{j=1}^{N} \check{X}_{ij} K((T_{1} - (t_{i} - t_{j})) / b) \right) \left( \sum_{i=1}^{N} \sum_{j=1}^{N} K((T_{1} - (t_{i} - t_{j}))) / b) \right)^{-1},
#' }
#' where \eqn{\check{X}_{ij} = (X(t_{i}) - \bar{X}) (X(t_{j}) - \bar{X}),}
#' which is then used in [truncated_est],
#' \deqn{\hat{\rho}_{1}(t) = \left\{ \begin{array}{ll}
#' \hat{\rho}(t) & 0 \leq t \leq T_{1} \\
#' \hat{\rho}(T_{1}) (T_{2} - t)(T_{2} - T_{1})^{-1} & T_{1} < t \leq T_{2} \\
#' 0 & t > T_{2}
#' \end{array} \right. .}
#'
#' @references
#' Hall, P. & Patil, P. (1994). Properties of nonparametric estimators of autocovariance for stationary random fields. Probability Theory and Related Fields 99(3), 399-424. https://doi.org/10.1007/bf01199899
#'
#' Hall, P., Fisher, N. I., & Hoffmann, B. (1994). On the nonparametric estimation of covariance functions. The Annals of Statistics 22(4), 2115-2134. https://doi.org/10.1214/aos/1176325774
#'
#' @param x A vector of lags.
#' @param meanX The average value of \code{X}.
#' @param T1 The first trunctation point.
#' @param b Bandwidth parameter, greater than 0.
#' @param xij_mat The matrix of pairwise covariance values.
#' @param kernel_name The name of the symmetric kernel (see [kernel_symm_ec]) function to be used. Possible values are:
#' gaussian, wave, rational_quadratic, and bessel_j. Alternatively, a custom kernel function can be provided, see the examples.
#' @param kernel_params A vector of parameters of the kernel function. See [kernel_symm_ec] for parameters.
#' @param custom_kernel If a custom kernel is to be used or not. Defaults to \code{FALSE}.
#'
#' @return The estimated autocovariance function at \eqn{T_{1}}.
#'
#' @examples
#' \dontrun{
#' X <- c(1, 2, 3, 4)
#' rho_T1(1:4, mean(X), 1, 0.1, Xij_mat(X, mean(X)), "gaussian", c(), FALSE)
#' my_kernel <- function(x, params) {
#'   stopifnot(params[1] > 0, length(x) >= 1)
#'   return(exp(-((abs(x) / params[1])^params[2])) * (2 * params[1]  * gamma(1 + 1/params[2])))
#' }
#' rho_T1(1:4, mean(X), 1, 0.1, Xij_mat(X, mean(X)), my_kernel, c(0.25), TRUE)
#' }
rho_T1 <- function(x, meanX, T1, b, xij_mat, kernel_name=c("gaussian", "wave", "rational_quadratic", "bessel_j"), kernel_params=c(), custom_kernel = FALSE) {
  stopifnot(is.numeric(x), length(x) >= 1, !any(is.na(x)), length(meanX) == 1, is.numeric(meanX), !is.na(meanX),
            length(T1) == 1, is.numeric(T1), !is.na(T1), T1 > 0, length(b) == 1, is.numeric(b), !is.na(b), b > 0,
            is.numeric(xij_mat), is.matrix(xij_mat), !any(is.na(xij_mat)), is.logical(custom_kernel))

  numerators <- numeric(length(x))
  denominators <- numeric(length(x))

  # Duplicating code as to only have 1 if statement, not length(x) if statements.
  if(custom_kernel) {
    stopifnot(is.function(kernel_name))
    for(i in 1:length(x)) {
      tij <- x[i] - x
      t_tij <- T1 - tij
      X_ij <- xij_mat[i, ]
      K_ij <- kernel_name(t_tij, c(b, kernel_params))
      numerator <- K_ij * X_ij
      denominator <- K_ij

      numerators[i] <- sum(numerator)
      denominators[i] <- sum(denominator)
    }
  }
  else {
    kernel_name <- match.arg(kernel_name)
    for(i in 1:length(x)) {
      tij <- x[i] - x
      t_tij <- T1 - tij
      X_ij <- xij_mat[i, ]
      K_ij <- kernel_symm_ec(t_tij, kernel_name, c(b, kernel_params[1], kernel_params[2]))
      numerator <- K_ij * X_ij
      denominator <- K_ij

      numerators[i] <- sum(numerator)
      denominators[i] <- sum(denominator)
    }
  }

  return( ( sum(numerators) / sum(denominators) ) )
}

#' Compute 1D Discrete Cosine Transform
#'
#' This function computes the Type-II discrete cosine transform.
#'
#' The Type-II discrete cosine transform is obtained using [stats::fft].
#' If \eqn{X} is of length N, construct a new signal \eqn{Y} of length \eqn{4N}, with values \eqn{Y_{2n} = 0, Y_{2n + 1} = x_{n}} for \eqn{0 \le n < N},
#' and \eqn{Y_{2N} = 0, Y_{4N - n} = y_{n}} for \eqn{0 < n < 2N.}
#' After this, the Type-II discrete cosine transform is computed by \code{0.5 * Re(stats::fft(Y))[1:(length(Y) / 4)]}.
#'
#' @references
#' Ochoa-Dominguez, H. & Rao, K.R. (2019). Discrete Cosine Transform, Second Edition. CRC Press. https://doi.org/10.1201/9780203729854
#'
#' Makhoul, J. (1980). A Fast Cosine Transform in One and Two Dimensions. IEEE Transactions on Acoustics, Speech, and Signal Processing 28(1), 27-34. https://doi.org/10.1109/TASSP.1980.1163351
#'
#' Stasiński, R. (2002). DCT Computation Using Real-Valued DFT Algorithms. Proceedings of the 11th European Signal Processing Conference.
#'
#' @param X A vector of values for which the discrete cosine transform is being computed.
#'
#' @return A vector of discrete cosine transform values.
#' @export
#'
#' @examples
#' X <- c(1, 2, 3)
#' dct_1d(X)
dct_1d <- function(X) {
  stopifnot(is.numeric(X), length(X) >= 1, !any(is.na(X)))
  newX <- c(X, rev(X))
  zerosX <- rep(0, length(newX))
  newX <- c(rbind(zerosX, newX))

  dct_seq <- 0.5*Re(stats::fft(newX))[1:(length(newX)/4)]
  return(dct_seq)
}

#' Compute 1D Inverse Discrete Cosine Transform
#'
#' This function computes the inverse of the Type-II discrete cosine transform.
#'
#' @details
#' The Type-II inverse discrete cosine transform is computed using [stats::fft].
#' For autocovariance function estimation, the spectrum is given in the input \code{X} and then an inverse FFT is applied.
#'
#' The original spectrum, \code{dct_full}, from \code{X} is obtained as follows.
#' First, the original sample Fourier spectrum is reconstructed using \code{dct_full <- c(X, 0, -X[-1], 0, rev(X[-1]))}. After this, an inverse FFT is applied,
#' \code{idct <- Re(stats::fft(dct_full, inverse = TRUE)) * (2 / length(dct_full))}, which gives the original function with additional zero-values at even indices.
#' The zeroes are dropped, which gives the untransformed X.
#'
#' @references
#' Ochoa-Dominguez, H. & Rao, K.R. (2019). Discrete Cosine Transform, Second Edition. CRC Press. https://doi.org/10.1201/9780203729854
#'
#' endolith (2013). Fast Cosine Transform via FFT. Signal Processing Stack Exchange.
#'     https://dsp.stackexchange.com/a/10606
#'
#' @param X A vector of values for which the discrete cosine transform is being computed.
#'
#' @return A vector with the inverse transform values.
#' @export
#'
#' @examples
#' X <- dct_1d(c(1, 2, 3))
#' idct_1d(X)
idct_1d <- function(X) {
  stopifnot(is.numeric(X), length(X) >= 1, !any(is.na(X)))
  # Reconstruct original frequency (i.e. the full dct_seq from dct_1d)
  dct_full <- c(X, 0, -rev(X), -X[-1], 0, rev(X[-1]))

  # Perform the inversion
  idct_vals <- Re(stats::fft(dct_full, inverse = TRUE)) * (2 / length(dct_full))

  # Recover original signal, drop 0s
  seqEven <- seq(2, 2*length(X), by=2)
  idct_vals_2 <- idct_vals[seqEven]

  return(idct_vals_2)
}

#' Compute the Truncated Kernel Regression Estimator.
#'
#' This function computes the truncated kernel regression estimator, based on the kernel regression estimator \eqn{\hat{\rho}(\cdot)}, see [adjusted_est].
#'
#' @details
#' This function computes the truncated kernel regression estimator,
#' \deqn{\hat{\rho}_{1}(t) = \left\{ \begin{array}{ll}
#' \hat{\rho}(t) & 0 \leq t \leq T_{1} \\
#' \hat{\rho}(T_{1}) (T_{2} - t)(T_{2} - T_{1})^{-1} & T_{1} < t \leq T_{2} \\
#' 0 & t > T_{2}
#' \end{array} \right. }
#' where \eqn{\hat{\rho}(\cdot)} is the kernel regression estimator, see [adjusted_est].
#'
#' Compared to [adjusted_est], this function brings down the estimate to zero linearly between \eqn{T_{1}} and \eqn{T_{2}}.
#' In the case of short-range dependence, this may be beneficial as it can remove estimation artefacts at large lags.
#'
#' To make this estimator positive-definite, the following procedure is used:
#' 1. Take the discrete Fourier cosine transform
#' \eqn{\widehat{\mathcal{F}}(\theta)}.
#' 2. Find the smallest frequency where its associated value in the spectral domain is negative
#' \deqn{\hat{\theta} = \inf\{ \theta > 0 :  \widehat{\mathcal{F}}(\theta)) < 0\}.}
#' 3. Set all values starting at the frequency to zero.
#' 4. Perform the Fourier inversion.
#'
#' If \eqn{\hat{\theta}} is a small frequency, most of the spectrum equals zero, resulting in an inaccurate estimate of the autocovariance function, see Bilchouris and Olenko (2025).
#'
#' @references
#' Hall, P. & Patil, P. (1994). Properties of nonparametric estimators of autocovariance for stationary random fields. Probability Theory and Related Fields 99(3), 399-424. https://doi.org/10.1007/bf01199899
#'
#' Hall, P., Fisher, N. I., & Hoffmann, B. (1994). On the nonparametric estimation of covariance functions. The Annals of Statistics 22(4), 2115-2134. https://doi.org/10.1214/aos/1176325774
#'
#' Bilchouris, A. & Olenko, A (2025). On Nonparametric Estimation of Covariogram. Austrian Statistical Society 54(1), 112-137. https://doi.org/10.17713/ajs.v54i1.1975
#'
#' @param X A vector representing observed values of the time series.
#' @param x A vector of lags.
#' @param t The arguments at which the autocovariance function is calculated at.
#' @param T1 The first truncation point, \eqn{T_{1} > 0.}
#' @param T2 The second truncation point, \eqn{T_{2} > T_{1} > 0.}
#' @param b Bandwidth parameter, greater than 0.
#' @param kernel_name The name of the symmetric kernel (see [kernel_symm_ec]) function to be used. Possible values are:
#' gaussian, wave, rational_quadratic, and bessel_j. Alternatively, a custom kernel function can be provided, see the examples.
#' @param kernel_params A vector of parameters of the kernel function. See [kernel_symm_ec] for parameters.
#' @param custom_kernel If a custom kernel is to be used or not. Defaults to \code{FALSE}.
#' @param pd Whether a positive-definite estimate should be used. Defaults to \code{TRUE}.
#' @param type Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.
#' @param meanX The average value of \code{X}. Defaults to \code{mean(X)}.
#' @param parallel Whether or not the computations should be done in parallel or not. Defaults to \code{FALSE}.
#' @param ncores The number of cores to be used in the parallel computations. Defaults to the number cores - 1 (threads if hyperthreading is available), calculated from \code{parallel::detectCores() - 1}.
#' @param cl_export A vector of any additional functions or variables to export for parallel computations. This may be required if \code{estimator} is not within the package. Defaults to \code{NULL}.
#' @param cl An optional cluster object created by \code{parallel::makeCluster}. Defaults to \code{NULL}, which creates a temporary PSOCK cluster.
#'
#' @return A \code{CovEsts} S3 object (list) with the following values
#' \describe{
#'  \item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
#'  \item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
#'  \item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the \code{type} parameter.}
#'  \item{\code{est_used}}{The estimator function used, in this case, 'truncated_est'.}
#' }
#' @export
#'
#' @importFrom utils tail
#'
#' @examples
#' X <- c(1, 2, 3, 4)
#' truncated_est(X, 1:4, 1:3, 1, 2, 0.1,
#'                   "gaussian")
#'
#' my_kernel <- function(x, params) {
#'   stopifnot(params[1] > 0, length(x) >= 1)
#'   return(exp(-((abs(x) / params[1])^params[2])) * (2 * params[1] * gamma(1 + 1/params[2])))
#' }
#'
#' truncated_est(X, 1:4, 1:3, 1, 2, 0.1, my_kernel, c(0.25), custom_kernel = TRUE)
#' \dontrun{
#' library(parallel)
#' X <- c(1, 2, 3, 4)
#' my_cl <- makePSOCKcluster(2)
#' truncated_est(X, 1:4, 1:3, 1, 2, 0.1, "gaussian", parallel = TRUE, cl = my_cl)
#' stopCluster(my_cl)
#' }
truncated_est <- function(X, x, t, T1, T2, b, kernel_name = c("gaussian", "wave", "rational_quadratic", "bessel_j"), kernel_params = c(), pd = TRUE, type = c("autocovariance", "autocorrelation"), meanX = mean(X), custom_kernel = FALSE, parallel = FALSE, ncores = parallel::detectCores() - 1, cl_export = NULL, cl = NULL) {
  stopifnot(is.numeric(X), length(X) >= 1, !any(is.na(X)), length(x) >= 1, is.numeric(x), !any(is.na(x)),
            length(meanX) == 1, is.numeric(meanX), !is.na(meanX), !any(is.na(t)), is.numeric(t), length(t) >= 1,
            length(T1) == 1, is.numeric(T1), !is.na(T1), T1 > 0, length(T2) == 1, is.numeric(T2),
            !is.na(T2), T2 > T1, length(b) == 1, is.numeric(b), b > 0, is.logical(custom_kernel),
            is.logical(pd), is.logical(parallel))

  type <- match.arg(type)

  if(custom_kernel) {
    stopifnot(is.function(kernel_name))
  }

  # outer_x_x <- outer(x, x, '-')
  vals_truncated_1 <- numeric(length(t))

  t_vals <- t[t <= T1]
  vals_truncated_1[1:length(t_vals)] <- adjusted_est(X = X, x = x , t = t_vals, b = b, kernel_name = kernel_name,
                                                                kernel_params = kernel_params, pd = FALSE, type = "autocovariance",
                                                                meanX = meanX, custom_kernel = custom_kernel, parallel = parallel,
                                                                ncores = ncores, cl_export = cl_export, cl = cl)$acf
  t_T1_T2 <- which(t > T1 & t <= T2)

  if(length(t_T1_T2) > 0) {
    xij_mat <- Xij_mat(X, mean(X))
    rhoT1 <- rho_T1(x, meanX, T1, b, xij_mat, kernel_name, kernel_params, custom_kernel)

    vals_truncated_1[t_T1_T2] <- rhoT1 * (T2 - t[t > T1 & t <= T2]) * (T2 - T1)^(-1)
    tail_t_T1_T2 <- utils::tail(t_T1_T2, 1)
    if(tail_t_T1_T2 > length(t)) {
      vals_truncated_1[(tail_t_T1_T2 + 1):length(vals_truncated_1)] <- 0
    }
  }

  if(pd) {
    vals_truncated_1 <- make_pd(vals_truncated_1, method.1 = FALSE)
  }

  if(type == 'autocorrelation') {
    vals_truncated_1 <- vals_truncated_1 / vals_truncated_1[1]
  }

  res <- list(acf = vals_truncated_1, lags = t, est_type = type, est_used = 'truncated_est')
  return(structure(res, class = "CovEsts"))
}

#' Compute the Kernel Regression Estimator.
#'
#' This function computes the kernel regression estimator of the autocovariance function.
#'
#' @details
#' The kernel regression estimator of an autocovariance function is defined as
#' \deqn{
#' \hat{\rho}(t) = \left( \sum_{i=1}^{N} \sum_{j=1}^{N} \check{X}_{ij} K((t -  (t_{i} - t_{j})) / b) \right) \left( \sum_{i=1}^{N} \sum_{j=1}^{N} K((t -  (t_{i} - t_{j})) / b) \right)^{-1},
#' }
#' where \eqn{\check{X}_{ij} = (X(t_{i}) - \bar{X}) (X(t_{j}) - \bar{X}).}
#'
#' If \code{pd} is \code{TRUE}, the estimator will be made positive-definite through the following procedure
#' 1. Take the discrete Fourier cosine transform,
#' \eqn{\widehat{\mathcal{F}}(\theta)}, of the estimated autocovariance function
#' 2. Compute a modified spectrum \eqn{\widetilde{\mathcal{F}}(\theta) = \max(\widehat{\mathcal{F}}(\theta), 0)} for all sample frequencies.
#' 3. Perform the Fourier inversion to obtain a new estimator.
#'
#' @references
#' Hall, P. & Patil, P. (1994). Properties of nonparametric estimators of autocovariance for stationary random fields. Probability Theory and Related Fields 99(3), 399-424. https://doi.org/10.1007/bf01199899
#'
#' Hall, P., Fisher, N. I., & Hoffmann, B. (1994). On the nonparametric estimation of covariance functions. The Annals of Statistics 22(4), 2115-2134. https://doi.org/10.1214/aos/1176325774
#'
#' @param X A vector representing observed values of the time series.
#' @param x A vector of lags.
#' @param t The arguments at which the autocovariance function is calculated at.
#' @param b Bandwidth parameter, greater than 0.
#' @param kernel_name The name of the symmetric kernel (see [kernel_symm_ec]) function to be used. Possible values are:
#' gaussian, wave, rational_quadratic, and bessel_j. Alternatively, a custom kernel function can be provided, see the examples.
#' @param kernel_params A vector of parameters of the kernel function. See [kernel_symm_ec] for parameters.
#' @param custom_kernel If a custom kernel is to be used or not. Defaults to \code{FALSE}.
#' @param pd Whether a positive-definite estimate should be used. Defaults to \code{TRUE}.
#' @param type Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.
#' @param meanX The average value of \code{X}. Defaults to \code{mean(X)}.
#' @param parallel Whether or not the computations should be done in parallel or not. Defaults to \code{FALSE}.
#' @param ncores The number of cores to be used in the parallel computations. Defaults to the number cores - 1 (threads if hyperthreading is available), calculated from \code{parallel::detectCores() - 1}.
#' @param cl_export A vector of any additional functions or variables to export for parallel computations. This may be required if \code{estimator} is not within the package. Defaults to \code{NULL}.
#' @param cl An optional cluster object created by \code{parallel::makeCluster}. Defaults to \code{NULL}, which creates a temporary PSOCK cluster.
#'
#' @return A \code{CovEsts} S3 object (list) with the following values
#' \describe{
#'  \item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
#'  \item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
#'  \item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the \code{type} parameter.}
#'  \item{\code{est_used}}{The estimator function used, in this case, 'adjusted_est'.}
#' }
#' @export
#'
#' @importFrom parallel makePSOCKcluster
#' @importFrom parallel stopCluster
#' @importFrom parallel clusterExport
#' @importFrom parallel parLapply
#'
#' @examples
#' X <- c(1, 2, 3, 4)
#' adjusted_est(X, 1:4, 1:3, 0.1, "gaussian")
#' my_kernel <- function(x, params) {
#'   stopifnot(params[1] > 0, length(x) >= 1)
#'   return(exp(-((abs(x) / params[1])^params[2])) * (2 * params[1] * gamma(1 + 1/params[2])))
#' }
#' adjusted_est(X, 1:4, 1:3, 0.1, my_kernel, c(0.25), custom_kernel = TRUE)
#' \dontrun{
#' library(parallel)
#' X <- c(1, 2, 3, 4)
#' my_cl <- makePSOCKcluster(2)
#' adjusted_est(X, 1:4, 1:3, 0.1, "gaussian", parallel = TRUE, cl = my_cl)
#' stopCluster(my_cl)
#' }
adjusted_est <- function(X, x, t, b, kernel_name = c("gaussian", "wave", "rational_quadratic", "bessel_j"), kernel_params=c(), pd = TRUE, type = c("autocovariance", "autocorrelation"), meanX = mean(X), custom_kernel = FALSE, parallel = FALSE, ncores = parallel::detectCores() - 1, cl_export = NULL, cl = NULL) {
  stopifnot(is.numeric(X), length(X) >= 1, !any(is.na(X)), is.numeric(x), length(x) >= 1, !any(is.na(x)),
            length(meanX) == 1, is.numeric(meanX), !is.na(meanX), !any(is.na(t)), is.numeric(t),
            length(t) >= 1, length(b) == 1, is.numeric(b), b > 0, is.logical(custom_kernel),
            is.logical(pd), is.logical(parallel))

  type <- match.arg(type)
  outer_x_x <- outer(x, x, '-')
  xij_mat <- Xij_mat(X, meanX)
  cov_vals <- numeric(length(t))

  kern_func <- NULL

  varList <- c()

  if(custom_kernel) {
    stopifnot(is.function(kernel_name))
    kern_func <- function(i) {
      K_ij <- kernel_name(t[i] - outer_x_x, c(b, kernel_params))
      return(sum(K_ij * xij_mat) / sum(K_ij))
    }
  }
  else {
    kernel_name <- match.arg(kernel_name)
    kern_func <- function(i) {
      K_ij <- kernel_symm_ec(t[i] - outer_x_x, kernel_name, c(b, kernel_params[1], kernel_params[2]))
      return(sum(K_ij * xij_mat) / sum(K_ij))
    }
    varList <- c(varList, "kernel_symm_ec")
  }

  if(parallel) {
    # See if the user objects actually exist before wasting time creating the PSOCK cluster.
    if(!is.null(cl_export)) {
      missing_exports <- cl_export[!sapply(cl_export, exists, envir = parent.frame())]
      if(length(missing_exports) > 0) {
        stop("The following objects in `cl_export` were not found in your environment: ", paste0(missing_exports, collapse = ", "))
      }
    }

    if(is.null(cl)) {
      cl <- parallel::makePSOCKcluster(ncores)
      on.exit(parallel::stopCluster(cl), add = TRUE)
    }

    varList <- c("t", "outer_x_x", "xij_mat", "b", "kernel_params", "kern_func", "kernel_symm_ec")
    parallel::clusterExport(cl, varlist = varList, envir = environment())

    # Export user supplied objects
    parallel::clusterExport(cl, varlist = cl_export, envir = parent.frame())

    cov_vals <- unlist(parallel::parLapply(cl, 1:length(t), kern_func))
  }

  else {
    for(i in 1:length(t)) {
      cov_vals[i] <- kern_func(i)
    }
  }

  if(pd) {
    # Perform DCT and set frequencies whose associated value is <0 to 0.
     cov_vals <- make_pd(cov_vals, method.1 = TRUE)
  }

  if(type == 'autocorrelation') {
    cov_vals <- cov_vals / cov_vals[1]
  }

  res <- list(acf = cov_vals, lags = t, est_type = type, est_used = 'adjusted_est')
  return(structure(res, class = "CovEsts"))
}

#' Make a Function Positive-Definite
#'
#' This function can make a function positive-definite using the methods proposed by P. Hall and his coauthors.
#'
#' @details
#' This function perform positive-definite adjustments proposed by P. Hall and his coauthors.
#'
#' Method 1 is as follows:
#' 1. Take the discrete Fourier cosine transform,
#' \eqn{\widehat{\mathcal{F}}(\theta)}.
#' 2. Compute a modified spectrum \eqn{\widetilde{\mathcal{F}}(\theta) = \max(\widehat{\mathcal{F}}(\theta), 0)} for all sample frequencies.
#' 3. Perform the Fourier inversion to obtain a new estimator.
#'
#'
#' Method 2 is as follows:
#' 1. Take the discrete Fourier cosine transform \eqn{\widehat{\mathcal{F}}(\theta)}.
#' 2. Find the smallest frequency where its associated value in the spectral domain is negative
#' \deqn{\hat{\theta} = \inf\{ \theta > 0 :  \widehat{\mathcal{F}}(\theta) < 0\}.}
#' 3. Compute a modified spectrum \eqn{\widetilde{\mathcal{F}} = \widehat{\mathcal{F}}(\theta)\textbf{1}(\theta < \hat{\theta}),}
#' where \eqn{\textbf{1}(A)} is the indicator function over the set \eqn{A.}
#' 4. Perform the Fourier inversion.
#'
#' @references
#' Hall, P. & Patil, P. (1994). Properties of nonparametric estimators of autocovariance for stationary random fields. Probability Theory and Related Fields 99(3), 399-424. https://doi.org/10.1007/bf01199899
#'
#' Hall, P., Fisher, N. I., & Hoffmann, B. (1994). On the nonparametric estimation of covariance functions. The Annals of Statistics 22(4), 2115-2134. https://doi.org/10.1214/aos/1176325774
#'
#' Bilchouris, A. & Olenko, A (2025). On Nonparametric Estimation of Covariogram. Austrian Statistical Society 54(1), 112-137. https://doi.org/10.17713/ajs.v54i1.1975
#'
#' @param x A vector of numeric values of an estimated autocovariance function.
#' @param method.1 Should method 1 be used (TRUE) or method 2 (FALSE).
#'
#' @return A vector that is the adjusted function or a \code{CovEsts} S3 object (list) with the following values
#' \describe{
#'  \item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
#'  \item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on, inherited from the argument \code{x}.}
#'  \item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', inherited from the argument \code{x}.}
#'  \item{\code{est_used}}{The estimator function used, inherited from the argument \code{x}.}
#'  \item{\code{correction_method}}{Either 'method.1' or 'method.2' depending on the argument \code{method.1}}
#' }
#' If a numeric vector is given for the argument \code{x}, then a numeric vector output is given, and if a \code{CovEsts} S3 object is given, a \code{CovEsts} object is given as output.
#' @export
#'
#' @examples
#' X <- c(1, 2, 3, 4)
#' make_pd(X)
#' check_pd(make_pd(X))
#' check_pd(make_pd(X, method.1 = FALSE))
make_pd <- function(x, method.1 = TRUE) {
  UseMethod("make_pd")
}

#' @describeIn make_pd Method for `CovEsts` objects.
#' @export
make_pd.CovEsts <- function(x, method.1 = TRUE) {
  idct_x <- make_pd.default(as.numeric(x), method.1)

  res <- list(acf = idct_x, lags = attr(x, 'lags'), est_type = attr(x, 'est_type'),
              est_used = attr(x, 'est_used'), correction_method = ifelse(method.1, 'method.1', 'method.2'))
  return(structure(res, class = "CovEsts"))
}

#' @describeIn make_pd Method for numeric vectors.
#' @export
make_pd.default <- function(x, method.1 = TRUE) {
  stopifnot(is.numeric(x), length(x) >= 1, !any(is.na(x)), is.logical(method.1))
  dct_x <- dct_1d(x)
  if(method.1) {
    dct_x[dct_x < 0] <- 0
  }
  else if(!method.1) {
    firstNeg <- which(dct_x[-1] < 0)[1] + 1

    if(length(firstNeg) == 0 || is.na(firstNeg)) {
      return(x)
    }

    # Set values after the first negative index onwards
    dct_x[firstNeg:length(dct_x)] <- 0
  }
  idct_x <- idct_1d(dct_x)
  return(idct_x)
}

# Xij_mat
test_that("Xij_mat works", {
  expect_equal(Xij_mat(c(1, 2), mean(c(1, 2))), matrix(c(0.25, -0.25, -0.25, 0.25), ncol=2, byrow=T))
})

test_that("Xij_mat fails for an empty process", {
  expect_error(Xij_mat(c(), mean(c(1, 2))))
})

test_that("Xij_mat fails for nonnumeric process", {
  expect_error(Xij_mat(c('a'), mean(c(1, 2))))
  expect_error(Xij_mat(c(1i), mean(c(1, 2))))
})

test_that("Xij_mat fails for a process with at least one NA", {
  expect_error(Xij_mat(c(1, NA, 2), mean(c(1, 2))))
})

test_that("Xij_mat fails for meanX not of length 1", {
  expect_error(Xij_mat(c(1, 2), c(1, 2)))
})

test_that("Xij_mat fails for nonnumeric meanX", {
  expect_error(Xij_mat(c(1, 2), 'a'))
  expect_error(Xij_mat(c(1, 2), 1i))
})

test_that("Xij_mat fails for NA meanX", {
  expect_error(Xij_mat(c(1, 2), NA))
})

# dct_1d
test_that("dct_1d works", {
  expect_equal(dct_1d(c(1, 2, 3)), c(6.000000, -1.732051, 0.000000), tolerance = 1e-6)
})

test_that("dct_1d fails for nonnumeric X", {
  expect_error(dct_1d(c('a')))
  expect_error(dct_1d(c(1i)))
})

test_that("dct_1d fails for empty X", {
  expect_error(dct_1d(c()))
})

test_that("dct_1d fails for at least one NA", {
  expect_error(dct_1d(c(1, NA, 3)))
})

# idct_1d
test_that("idct_1d works", {
  expect_equal(idct_1d(c(6.000000, -1.732051, 0.00000)), c(1, 2, 3), tolerance = 1e-6)
})

test_that("idct_1d fails for nonnumeric X", {
  expect_error(idct_1d(c('a')))
  expect_error(idct_1d(c(1i)))
})

test_that("idct_1d fails for empty X", {
  expect_error(idct_1d(c()))
})

test_that("idct_1d fails for at least one NA", {
  expect_error(idct_1d(c(1, NA, 3)))
})

# rho_T1
test_that("rho_T1 works", {
  expect_equal(rho_T1(test_x, test_meanX, 2, 0.01, test_xij), -1)
})

test_that("rho_T1 faisl for nonnumeric x", {
  expect_error(rho_T1(c('a'), test_meanX, 2, 0.01, test_xij))
  expect_error(rho_T1(c(1i), test_meanX, 2, 0.01, test_xij))
})

test_that("rho_T1 fails for empty x", {
  expect_error(rho_T1(c(), test_meanX, 2, 0.01, test_xij))
})

test_that("rho_T1 fails for at least one NA in x", {
  expect_error(rho_T1(c(1, NA, 2), test_meanX, 2, 0.01, test_xij))
})

test_that("rho_T1 fails for meanX of length not equal to 1", {
  expect_error(rho_T1(test_x, c(1, 2), 2, 0.01, test_xij))
  expect_error(rho_T1(test_x, c(), 2, 0.01, test_xij))
})

test_that("rho_T1 fails for nonnumeric meanX", {
  expect_error(rho_T1(test_x, 'a', 2, 0.01, test_xij))
  expect_error(rho_T1(test_x, 1i, 2, 0.01, test_xij))
})

test_that("rho_T1 fails for NA meanX", {
  expect_error(rho_T1(test_x, NA, 2, 0.01, test_xij))
})

test_that("rho_T1 fails for T1 of length not equal to 1", {
  expect_error(rho_T1(test_x, test_meanX, c(1, 2), 0.01, test_xij))
  expect_error(rho_T1(test_x, test_meanX, c(), 0.01, test_xij))
})

test_that("rho_T1 fails for nonnumeric T1", {
  expect_error(rho_T1(test_x, test_meanX, 'a', 0.01, test_xij))
  expect_error(rho_T1(test_x, test_meanX, 1i, 0.01, test_xij))
})

test_that("rho_T1 fails for NA T1", {
  expect_error(rho_T1(test_x, test_meanX, NA, 0.01, test_xij))
})

test_that("rho_T1 fails for nonpositive T1", {
  expect_error(rho_T1(test_x, test_meanX, 0, 0.01, test_xij))
  expect_error(rho_T1(test_x, test_meanX, -1, 0.01, test_xij))
})

test_that("rho_T1 fails for b of length not equal to 1", {
  expect_error(rho_T1(test_x, test_meanX, 2, c(1, 2), test_xij))
  expect_error(rho_T1(test_x, test_meanX, 2, c(), test_xij))
})

test_that("rho_T1 fails for nonnumeric b", {
  expect_error(rho_T1(test_x, test_meanX, 2, 'a', test_xij))
  expect_error(rho_T1(test_x, test_meanX, 2, 1i, test_xij))
})

test_that("rho_T1 fails for NA b", {
  expect_error(rho_T1(test_x, test_meanX, 2, NA, test_xij))
})

test_that("rho_T1 fails for nonpositive b", {
  expect_error(rho_T1(test_x, test_meanX, 2, 0, test_xij))
  expect_error(rho_T1(test_x, test_meanX, 2, -1, test_xij))
})

test_that("rho_T1 fails for nonnumeric Xij_mat", {
  expect_error(rho_T1(test_x, test_meanX, 2, 0.01, matrix(c('a', 1, 2, 1), ncol=2)))
  expect_error(rho_T1(test_x, test_meanX, 2, 0.01, matrix(c(1i, 1, 2, 1), ncol=2)))
})

test_that("rho_T1 fails for nonmatrix Xij_mat", {
  expect_error(rho_T1(test_x, test_meanX, 2, 0.01, 1))
  expect_error(rho_T1(test_x, test_meanX, 2, 0.01, c(1, 2)))
  expect_error(rho_T1(test_x, test_meanX, 2, 0.01, list(a=1, b=2)))
})

test_that("rho_T1 fails for any NA in Xij_mat", {
  expect_error(rho_T1(test_x, test_meanX, 2, 0, matrix(c(1, 1, NA, 1), ncol=2)))
})

test_that("rho_T1 fails for any nonboolean custom_kernel", {
  expect_error(rho_T1(test_x, test_meanX, 2, 0, test_xij, custom_kernel = 1))
  expect_error(rho_T1(test_x, test_meanX, 2, 0, test_xij, custom_kernel = 'TRUE'))
})

test_that("rho_T1 fails for any nonboolean custom_kernel", {
  expect_error(rho_T1(test_x, test_meanX, 2, 0, test_xij, custom_kernel = 1))
  expect_error(rho_T1(test_x, test_meanX, 2, 0, test_xij, custom_kernel = 'TRUE'))
})

test_that("rho_T1 fails for kernel_name not in 'gaussian' 'wave', 'rational_quadratic', 'bessel_j'", {
  expect_error(rho_T1(test_x, test_meanX, 2, 0, test_xij, kernel_name = "123", custom_kernel = FALSE))
})

# truncated_est
test_that("truncated_est works", {
  expect_equal(truncated_est(test_X, test_x, c(1, 2, 3), 2, 3, 0.01, "gaussian", meanX = test_meanX)$acf, c(0, -1, 0))
})

test_that("truncated_est fails for nonnumeric X", {
  expect_error(truncated_est(c(1, 'a', 3), test_x, 2, 2, 3, 0.01, "gaussian"))
  expect_error(truncated_est(c(1, 1i, 3), test_x, 2, 2, 3, 0.01, "gaussian"))
})

test_that("truncated_est fails for empty X", {
  expect_error(truncated_est(c(), test_x, 2, 2, 3, 0.01, "gaussian"))
})

test_that("truncated_est fails for at least one NA in X", {
  expect_error(truncated_est(c(1, NA, 3), test_x, 2, 2, 3, 0.01, "gaussian"))
})

test_that("truncated_est fails for empty x", {
  expect_error(truncated_est(test_X, c(), 2, 2, 3, 0.01, "gaussian"))
})

test_that("truncated_est fails for nonnumeric x", {
  expect_error(truncated_est(test_X, c(1, 'a', 3), 2, 2, 3, 0.01, "gaussian"))
  expect_error(truncated_est(test_X, c(1, 1i, 3), 2, 2, 3, 0.01, "gaussian"))
})

test_that("truncated_est fails for at least one NA in x", {
  expect_error(truncated_est(test_X, c(1, NA, 3), 2, 2, 3, 0.01, "gaussian"))
})

test_that("truncated_est fails for empty t", {
  expect_error(truncated_est(test_X, test_x, c(), 2, 3, 0.01, "gaussian"))
})

test_that("truncated_est fails for nonnumeric t", {
  expect_error(truncated_est(test_X, test_x, 'a', 2, 3, 0.01, "gaussian"))
  expect_error(truncated_est(test_X, test_x, 1i, 2, 3, 0.01, "gaussian"))
})

test_that("truncated_est fails for NA t", {
  expect_error(truncated_est(test_X, test_x, NA, 2, 3, 0.01, "gaussian"))
})

test_that("truncated_est fails for T1 of length not equal to 1", {
  expect_error(truncated_est(test_X, test_x, 2, c(1, 2), 3, 0.01, "gaussian"))
  expect_error(truncated_est(test_X, test_x, 2, c(), 3, 0.01, "gaussian"))
})

test_that("truncated_est fails for nonnumeric T1", {
  expect_error(truncated_est(test_X, test_x, 2, 'a', 3, 0.01, "gaussian"))
  expect_error(truncated_est(test_X, test_x, 2, 1i, 3, 0.01, "gaussian"))
})

test_that("truncated_est fails for NA T1", {
  expect_error(truncated_est(test_X, test_x, 2, NA, 3, 0.01, "gaussian"))
})

test_that("truncated_est fails for T1 <= 0", {
  expect_error(truncated_est(test_X, test_x, 2, 0, 3, 0.01, "gaussian"))
  expect_error(truncated_est(test_X, test_x, 2, -0.5, 3, 0.01, "gaussian"))
})

test_that("truncated_est fails for T2 of length not equal to 1", {
  expect_error(truncated_est(test_X, test_x, 2, 2, c(1, 2), 0.01, "gaussian"))
  expect_error(truncated_est(test_X, test_x, 2, 2, c(), 0.01, "gaussian"))
})

test_that("truncated_est fails for nonnumeric T2", {
  expect_error(truncated_est(test_X, test_x, 2, 2, 'a', 0.01, "gaussian"))
  expect_error(truncated_est(test_X, test_x, 2, 2, 1i, 0.01, "gaussian"))
})

test_that("truncated_est fails for NA T2", {
  expect_error(truncated_est(test_X, test_x, 2, 2, NA, 0.01, "gaussian"))
})

test_that("truncated_est fails for T2 <= T1", {
  expect_error(truncated_est(test_X, test_x, 2, 2, 1, 0.01, "gaussian"))
  expect_error(truncated_est(test_X, test_x, 2, 2, 2, 0.01, "gaussian"))
})

test_that("truncated_est fails for b of length not equal to 1", {
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, c(1, 2), "gaussian"))
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, c(), "gaussian"))
})

test_that("truncated_est fails for nonnumeric b", {
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 'a', "gaussian"))
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 1i, "gaussian"))
})

test_that("truncated_est fails for nonpositive b", {
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 0, "gaussian"))
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, -0.01, "gaussian"))
})

test_that("truncated_est fails for nonboolean custom_kernel", {
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 0.01, "gaussian", custom_kernel = 1))
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 0.01, "gaussian", custom_kernel = 'TRUE'))
})

test_that("truncated_est fails for nonboolean pd", {
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 0.01, "gaussian", pd = 1))
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 0.01, "gaussian", pd = 'TRUE'))
})

test_that("truncated_est fails for nonnumeric meanX", {
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 0.01, "gaussian", meanX = 'a'))
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 0.01, "gaussian", meanX = 1i))
})

test_that("truncated_est fails for NA meanX", {
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 0.01, "gaussian", meanX = NA))
})

test_that("truncated_est fails for meanX not of length 1", {
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 0.01, "gaussian", meanX = c()))
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 0.01, "gaussian", meanX = c(1, 2)))
})

test_that("truncated_est fails for type not 'autocovariance' or 'autocorrelation'", {
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 0.01, "gaussian", type = "covariance"))
})

test_that("truncated_est fails for nonboolean parallel", {
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 0.01, "gaussian", parallel = 1))
  expect_error(truncated_est(test_X, test_x, 2, 2, 3, 0.01, "gaussian", parallel = 'TRUE'))
})

# Assume the user knows what they're doing if they are working with parallelism.

# adjusted_est_est
test_that("adjusted_est works", {
  expect_equal(adjusted_est(test_X, test_x, c(1, 2, 3), 0.01, "gaussian", meanX = test_meanX)$acf, c(2/3, -1/3, -1/3))
})

test_that("adjusted_est fails for nonnumeric X", {
  expect_error(adjusted_est(c(1, 'a', 3), test_x, c(1, 2, 3), 0.01, "gaussian"))
  expect_error(adjusted_est(c(1, 1i, 3), test_x, c(1, 2, 3), 0.01, "gaussian"))
})

test_that("adjusted_est fails for empty X", {
  expect_error(adjusted_est(c(), test_x, c(1, 2, 3), 0.01, "gaussian"))
})

test_that("adjusted_est fails for at least one NA in X", {
  expect_error(adjusted_est(c(1, NA, 3), test_x, c(1, 2, 3), 0.01, "gaussian"))
})

test_that("adjusted_est fails for nonnumeric x", {
  expect_error(adjusted_est(test_X, c(1, 'a', 3), c(1, 2, 3), 0.01, "gaussian"))
  expect_error(adjusted_est(test_X, c(1, 1i, 3), c(1, 2, 3), 0.01, "gaussian"))
})

test_that("adjusted_est fails for empty x", {
  expect_error(adjusted_est(test_X, c(), c(1, 2, 3), 0.01, "gaussian"))
})

test_that("adjusted_est fails for at least one NA in x", {
  expect_error(adjusted_est(test_X, c(1, NA, 3), c(1, 2, 3), 0.01, "gaussian"))
})

test_that("adjusted_est fails for meanX of length not equal to 1", {
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 0.01, "gaussian", meanX = c(1, 2)))
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 0.01, "gaussian", meanX = c()))
})

test_that("adjusted_est fails for nonnumeric meanX", {
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 0.01, "gaussian", meanX = 'a'))
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 0.01, "gaussian", meanX = 1i))
})

test_that("adjusted_est fails for NA meanX", {
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 0.01, "gaussian", NA))
})

test_that("adjusted_est fails for at least one NA in t", {
  expect_error(adjusted_est(test_X, test_x, c(1, NA, 3), 0.01, "gaussian"))
})

test_that("adjusted_est fails for nonnumeric t", {
  expect_error(adjusted_est(test_X, test_x, c(1, 'a', 3), 0.01, "gaussian"))
  expect_error(adjusted_est(test_X, test_x, c(1, 1i, 3), 0.01, "gaussian"))
})

test_that("adjusted_est fails for empty t", {
  expect_error(adjusted_est(test_X, test_x, c(), 0.01, "gaussian"))
})

test_that("adjusted_est fails for b of length not equal to 1", {
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), c(1, 2), "gaussian"))
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), c(), "gaussian"))
})

test_that("adjusted_est fails for nonnumeric b", {
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 'a', "gaussian"))
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 1i, "gaussian"))
})

test_that("adjusted_est fails for b <= 0", {
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 0, "gaussian"))
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), -0.01, "gaussian"))
})

test_that("adjusted_est fails for nonboolean custom_kernel", {
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 0.01, "gaussian", custom_kernel = 1))
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 0.01, "gaussian", custom_kernel = 'TRUE'))
})

test_that("adjusted_est fails for nonboolean pd", {
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 0.01, "gaussian", pd = 1))
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 0.01, "gaussian", pd = 'TRUE'))
})

test_that("adjusted_est fails for kernel_name not in 'gaussian' 'wave', 'rational_quadratic', 'bessel_j'", {
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 0.01, kernel_name = "123"))
})

test_that("adjusted_est fails for nonboolean parallel", {
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 0.01, "gaussian", parallel = 1))
  expect_error(adjusted_est(test_X, test_x, c(1, 2, 3), 0.01, "gaussian", parallel = 'TRUE'))
})

# Assume the user knows what they're doing if they are working with parallelism.


# make_pd
test_that("make_pd works", {
  expect_equal(make_pd(c(1, 0, 3), TRUE), c(2, 0, 2))
  expect_equal(make_pd(c(1, 0, 3), FALSE), c(4/3, 4/3, 4/3))
})


test_that("make_pd fails for nonnumeric x", {
  expect_error(make_pd(c(1, 'a', 3), TRUE))
  expect_error(make_pd(c(1, 1i, 3), TRUE))
})

test_that("make_pd fails for empty x", {
  expect_error(make_pd(c(), TRUE))
})

test_that("make_pd fails for at least one NA in x", {
  expect_error(make_pd(c(1, NA, 3), TRUE))
})

test_that("make_pd fails for nonboolean method.1", {
  expect_error(make_pd(c(1, 2, 3), 1))
  expect_error(make_pd(c(1, 2, 3), 'TRUE'))
})

#' Kernel Correction of the Standard Estimator.
#'
#' @description
#' This function computes the standard autocovariance estimator and applies kernel correction to it,
#' \deqn{\widehat{C}_{T}^{(a)}(h) = \widehat{C}(h) a_{T}(h), } where \eqn{a_{T}(h) := a(h / N_{T}).}
#' It uses a kernel \eqn{a(\cdot)} which decays or vanishes to zero (depending on the type of kernel) where \eqn{a(0) = 1.}
#' The rate or value at which the kernel vanishes is \eqn{N_{T}}, which is recommended to be of order \eqn{0.1 N}, where \eqn{N} is the length of the observation window, however, one may need to play with this value.
#'
#' @details
#' The aim of this estimator is gradually bring the estimated values to zero through the use of a kernel multiplier. This can be useful when estimating an
#' autocovariance function that is short-range dependent as estimators can have large fluctuations as the lag increases, or to deal with the wave artefacts for large lags, see Bilchouris and Olenko (2025).
#' This estimator can be positive-definite depending on whether the choice of \eqn{\widehat{C}(\cdot)} and \eqn{a} are chosen to be positive-definite or not.
#'
#' @references
#' Yaglom, AM (1987). Correlation Theory of Stationary and Related Random Functions. Volume I: Basic Results. Springer New York. https://doi.org/10.1007/978-1-4612-4628-2
#'
#' Bilchouris, A. & Olenko, A (2025). On Nonparametric Estimation of Covariogram. Austrian Statistical Society (Vol. 54, Issue 1). https://doi.org/10.17713/ajs.v54i1.1975
#'
#' @param X A vector representing observed values of the time series.
#' @param kernel_name The name of the [kernel_ec] function to be used. Possible values are:
#' gaussian, exponential, wave, rational_quadratic, spherical, circular, bessel_j, matern, cauchy.
#' @param kernel_params A vector of parameters of the kernel function. See [kernel_ec] for parameters.
#' In the case of gaussian, wave, rational_quadratic, spherical and circular, \code{N_T} takes the place of \eqn{\theta}.
#' For kernels that require parameters other than \eqn{\theta}, such as the Matern kernel, those parameters are passed.
#' @param N_T The range at which the kernel function vanishes at. Recommended to be \eqn{0.1 N} when considering all lags. This parameter may be large for a lag small estimation lag.
#' @param meanX The average value of \code{X}. Defaults to \code{mean(X)}.
#' @param maxLag An optional parameter that determines the maximum lag to compute the estimated autocovariance function at. Defaults to \code{length(X) - 1}.
#' @param x A vector of lag indices. Defaults to the sequence \code{0:length(X)}. Must be at least as large as \code{maxLag + 1}.
#' @param pd Whether a positive-definite estimate should be used. Defaults to \code{TRUE}.
#' @param custom_kernel If a custom kernel is to be used or not. Defaults to \code{FALSE}. See examples.
#' @param type Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.
#'
#' @return A \code{CovEsts} S3 object (list) with the following values
#' \describe{
#'  \item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
#'  \item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
#'  \item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the \code{type} parameter.}
#'  \item{\code{est_used}}{The estimator function used, in this case, 'corrected_est'.}
#' }
#' @export
#'
#' @examples
#' X <- c(1, 2, 3)
#' corrected_est(X, "gaussian")
#'
#' X <- rnorm(1000)
#' Y <- c(X[1], X[2])
#' for(i in 3:length(X)) { Y[i] <- X[i] - 0.3*X[i - 1] - 0.6*X[i - 2] }
#' plot(Y)
#' plot(corrected_est(Y, "bessel_j",
#'      kernel_params=c(0, 1), N_T=0.2*length(Y)))
#'
#' # Custom kernel
#' my_kernel <- function(x, theta, params) {
#'   stopifnot(theta > 0, length(x) >= 1, all(x >= 0))
#'   return(sapply(x, function(t) ifelse(t == 0, 1,
#'          ifelse(t == Inf, 0,
#'          (sin((t^params[1]) / theta) / ((t^params[1]) / theta)) * cos((t^params[2]) / theta)))))
#' }
#' plot(corrected_est(Y,
#'      my_kernel, kernel_params=c(2, 0.25), custom_kernel = TRUE))
corrected_est <- function(X, kernel_name = c("gaussian", "exponential", "wave", "rational_quadratic", "spherical", "circular", "bessel_j", "matern", "cauchy"), kernel_params = c(), N_T = 0.1 * length(X), pd = TRUE, maxLag = length(X) - 1, x = 0:length(X), type = c("autocovariance", "autocorrelation"), meanX = mean(X), custom_kernel = FALSE) {
  stopifnot(is.logical(custom_kernel), length(X) > 0, all(!is.na(X)), is.numeric(X), is.numeric(N_T), !is.na(N_T),
            N_T > 0, is.numeric(meanX), !is.na(meanX), is.logical(pd), is.numeric(maxLag), maxLag >= 0, !is.na(maxLag),
            maxLag <= (length(X) - 1), maxLag %% 1 == 0, length(meanX) == 1, is.numeric(meanX), !is.na(meanX),
            is.numeric(x), all(!is.na(x)), length(x) > 0)

  type <- match.arg(type)

  retVec <- standard_est(X, pd = pd, maxLag = maxLag, type='autocovariance', meanX = meanX)$acf
  if(!custom_kernel) {
    kernel_name <- match.arg(kernel_name)
    retVec <- retVec * kernel_ec(0:maxLag, kernel_name, c(N_T, kernel_params))

    if(type == 'autocorrelation') {
      retVec <- retVec / retVec[1]
    }

    res <- list(acf = retVec, lags = x[1:(maxLag + 1)], est_type = type, est_used = 'corrected_est')
    return(structure(res, class = "CovEsts"))
  }

  if(custom_kernel) {
    stopifnot(is.function(kernel_name))
    # Keep this as an apply statement as the user may not implement a kernel that allows for a vector paramter.
    retVec <- retVec * sapply(seq(0, maxLag, by=1), function(t) kernel_name(t, N_T, kernel_params))

    if(type == 'autocorrelation') {
      retVec <- retVec / retVec[1]
    }

    res <- list(acf = retVec, lags = x[1:(maxLag + 1)], est_type = type, est_used = 'corrected_est')
    return(structure(res, class = "CovEsts"))
  }
}

#' Kernel Correction for an Estimated Autocovariance Function.
#'
#' @description
#' This function applies kernel correction to an estimated autocovariance function,
#' \deqn{\widehat{C}_{T}^{(a)}(h) = \widehat{C}(h) a_{T}(h),} where \eqn{a_{T}(h) := a(h / N_{T}).}
#' It uses a kernel \eqn{a(\cdot)} which decays or vanishes to zero (depending on the type of kernel) where \eqn{a(0) = 1.}
#' The rate or value at which the kernel vanishes is \eqn{N_{T}}, which is recommended to be of order \eqn{0.1 N}, where \eqn{N} is the length of the observation window, however, one may need to play with this value.
#'
#' @param estCov A vector whose values are an estimate autocovariance function.
#' @param kernel_name The name of the [kernel_ec] function to be used. Possible values are:
#' gaussian, exponential, wave, rational_quadratic, spherical, circular, bessel_j, matern, cauchy.
#' @param kernel_params A vector of parameters of the kernel function. See [kernel_ec] for parameters.
#' In the case of gaussian, wave, rational_quadratic, spherical and circular, \code{N_T} takes the place of \eqn{\theta}.
#' For kernels that require parameters other than \eqn{\theta}, such as the Matern kernel, those parameters are passed.
#' @param N_T The range at which the kernel function vanishes at. Recommended to be \eqn{0.1 N} when considering all lags. This parameter may be large for a lag small estimation lag.
#' @param maxLag An optional parameter that determines the maximum lag to compute the estimated autocovariance function at. Defaults to \code{length(estCov) - 1}.
#' @param x A vector of lag indices. Defaults to the sequence \code{0:length(X)}. Must be at least as large as \code{maxLag + 1}.
#' @param custom_kernel If a custom kernel is to be used or not. Defaults to \code{FALSE}. See the examples of [corrected_est] for usage.
#' @param type Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.
#'
#' @return A vector whose values are the kernel corrected autocovariance estimates or \code{CovEsts} S3 object (list) with the following values
#' \describe{
#'  \item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
#'  \item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on, inherited from the argument \code{estCov}.}
#'  \item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the argument \code{type}.}
#'  \item{\code{est_used}}{The estimator function used, in this case, 'kernel_est'.}
#' }
#' If a numeric vector is given for the argument \code{estCov}, then a numeric vector output is given, and if a \code{CovEsts} S3 object is given, a \code{CovEsts} object is given as output.
#' @export
#'
#' @examples
#' X <- rnorm(1000)
#' Y <- c(X[1], X[2])
#' for(i in 3:length(X)) { Y[i] <- X[i] - 0.3*X[i - 1] - 0.6*X[i - 2] }
#' cov_est <- standard_est(Y)
#' plot(cov_est)
#' plot(kernel_est(cov_est,
#'      "bessel_j", kernel_params=c(0, 1), N_T=0.2*length(Y)))
kernel_est <- function(estCov, kernel_name = c("gaussian", "exponential", "wave", "rational_quadratic", "spherical", "circular", "bessel_j", "matern", "cauchy"), kernel_params = c(), N_T = 0.1 * length(estCov), maxLag = length(estCov) - 1, x = 0:length(estCov), type = c("autocovariance", "autocorrelation"), custom_kernel = FALSE) {
  UseMethod("kernel_est")
}

#' @describeIn kernel_est Method for `CovEsts` objects.
#' @export
kernel_est.CovEsts <- function(estCov, kernel_name = c("gaussian", "exponential", "wave", "rational_quadratic", "spherical", "circular", "bessel_j", "matern", "cauchy"), kernel_params = c(), N_T = 0.1 * length(estCov$acf), maxLag = length(estCov$acf) - 1, x = estCov$lags, type = c("autocovariance", "autocorrelation"), custom_kernel = FALSE) {
  est <- kernel_est.default(estCov$acf, kernel_name, kernel_params = kernel_params, N_T = N_T, maxLag = length(estCov$lags) - 1, type = type, custom_kernel = custom_kernel)

  res <- list(acf = est, lags = estCov$lags, est_type = type, est_used = 'kernel_est')
  return(structure(res, class = "CovEsts"))
}

#' @describeIn kernel_est Method for numeric vectors.
#' @export
kernel_est.default <- function(estCov, kernel_name = c("gaussian", "exponential", "wave", "rational_quadratic", "spherical", "circular", "bessel_j", "matern", "cauchy"), kernel_params = c(), N_T = 0.1 * length(estCov), maxLag = length(estCov) - 1, x = 0:length(estCov), type = c("autocovariance", "autocorrelation"), custom_kernel = FALSE) {
  stopifnot(is.logical(custom_kernel), length(estCov) > 0, is.numeric(estCov), !any(is.na(estCov)), is.numeric(N_T), N_T > 0,
            !is.na(N_T), is.numeric(maxLag), maxLag >= 0, maxLag <= (length(estCov) - 1), maxLag %% 1 == 0, !is.na(maxLag))

  type <- match.arg(type)

  if(!custom_kernel) {
    kernel_name <- match.arg(kernel_name)

    estCov <- estCov[1:(maxLag+1)] * kernel_ec(0:maxLag, kernel_name, c(N_T, kernel_params))

    if(type == 'autocorrelation') {
      estCov <- estCov / estCov[1]
    }

    return(estCov)
  }

  if(custom_kernel) {
    stopifnot(is.function(kernel_name))
    # Keep this as an apply statement as the user may not implement a kernel that allows for a vector paramter.
    estCov <- estCov[1:(maxLag+1)] * sapply(seq(0, maxLag, by=1), function(t) kernel_name(t, N_T, kernel_params))

    if(type == 'autocorrelation') {
      estCov <- estCov / estCov[1]
    }

    return(estCov)
  }

  return("Something went wrong in `kernel_est`.")
}

#' Solve Linear Shrinking
#'
#' This is an objective function used to select \eqn{\lambda \in [0, 1]} in linear shrinking, see [shrinking].
#'
#' @param par The initial parameter  used in the maximisation process.
#' @param corr_mat The autocorrelation matrix of the considered time series.
#' @param target A shrinkage target matrix used in the shrinking process. This should only be used if you wish to use a specific matrix as the target.
#'
#' @references
#' Devlin, S. J., Gnanadesikan R. & Kettenring, J. R. (1975). Robust Estimation and Outlier Detection with Correlation Coefficients. Biometrika, 62(3), 531-545. 10.1093/biomet/62.3.531
#'
#' Rousseeuw, P. J. & Molenberghs, G. (1993). Transformation of Non Positive Semidefinite Correlation Matrices. Communications in Statistics - Theory and Methods, 22(4), 965-984. 10.1080/03610928308831068
#'
#' @return A numeric value that is either equal to \eqn{-}\code{par} or 1.
#'
#' @examples
#' \dontrun{
#' estCorr <- c(1, 0.5, 0)
#' corr_mat <- cyclic_matrix(estCorr)
#' solve_shrinking(0.5, corr_mat, diag(length(estCorr)))
#' }
solve_shrinking <- function(par, corr_mat, target) {
  stopifnot(is.numeric(par), !is.na(par), length(par) == 1, is.numeric(corr_mat), !any(is.na(corr_mat)), is.numeric(target), !any(is.na(target)),
            all(dim(corr_mat) == dim(target)))
  adj_corr_mat <- (par[1] * corr_mat) + (1 - par[1]) * target
  isPd <- check_pd(adj_corr_mat)
  if(isPd) {
    return(-par[1])
  }
  return(1)
}

#' Linear Shrinking
#'
#' This function corrects an autocovariance/autocorrelation function estimate via linear shrinking of the autocorrelation matrix.
#'
#' @details
#' This function corrects an autocovariance/autocorrelation function estimate via linear shrinking of the autocorrelation matrix.
#' The shrunken autocorrelation matrix is computed as follows
#' \deqn{\widetilde{R} = \lambda R + (1 - \lambda)I_{p},}
#' where \eqn{\widetilde{R}} is the shrunken autocorrelation matrix, \eqn{R} is the original autocorrelation matrix, \eqn{\lambda \in [0, 1],} and \eqn{I_{p}} is the \eqn{p\times p} identity matrix.
#' \eqn{\lambda} is chosen in such a away that largest value which still results in a positive-definite matrix.
#' The shrunken matrix will be positive-definite.
#'
#' @references
#' Devlin, S. J., Gnanadesikan R. & Kettenring, J. R. (1975). Robust Estimation and Outlier Detection with Correlation Coefficients. Biometrika, 62(3), 531-545. 10.1093/biomet/62.3.531
#'
#' Rousseeuw, P. J. & Molenberghs, G. (1993). Transformation of Non Positive Semidefinite Correlation Matrices. Communications in Statistics - Theory and Methods, 22(4), 965-984. 10.1080/03610928308831068
#'
#' @param estCov A vector whose values are an estimate autocovariance/autocorrelation function.
#' @param return_matrix A boolean determining whether the shrunken matrix or the corresponding vector is returned. If \code{FALSE}, it returns a vector whose values are the shrunken autocorrelation function. Defaults to \code{FALSE}.
#' @param target A shrinkage target matrix used in the shrinking process. This should only be used if you wish to use a specific matrix as the target.
#'
#' @return A vector with values of the shrunken autocorrelation function, the corresponding matrix (depending on \code{return_matrix}), or \code{CovEsts} S3 object (list) with the following values
#' \describe{
#'  \item{\code{acf}}{A numeric vector containing the shrunken autocovariance/autocorrelation estimates.}
#'  \item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on, inherited from the argument \code{estCov}.}
#'  \item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the argument \code{type}.}
#'  \item{\code{est_used}}{The estimator function used, in this case, inherited from the argument \code{estCov}.}
#'  \item{\code{correction_method}}{This value is 'shrinking'.}
#'  \item{\code{lambda}}{The \eqn{\lambda} value obtained during the shrinking process.}
#' }
#' If a numeric vector is given for the argument \code{estCov}, then a numeric vector output is given, and if a \code{CovEsts} S3 object is given, a \code{CovEsts} object is given as output.
#'
#' @export
#'
#' @examples
#' estCorr <- c(1, 0.8, 0.5, -1.2)
#' shrinking(estCorr)
#' target <- diag(length(estCorr))
#' shrinking(estCorr, TRUE, target)
shrinking <- function(estCov, return_matrix = FALSE, target = NULL) {
  UseMethod("shrinking")
}

#' @describeIn shrinking Method for `CovEsts` objects.
#' @export
shrinking.CovEsts <- function(estCov, return_matrix = FALSE, target = NULL) {
  shrunk <- shrinking.default(as.numeric(estCov), return_matrix = return_matrix, target = target)

  res <- list(acf=shrunk[['shrunken']], lags = estCov$lags, est_type = estCov$est_type,
              est_used = estCov$est_used, correction_method = 'shrinking',
              lambda = shrunk[['lambda']])
  return(structure(res, class = "CovEsts"))
}

#' @describeIn shrinking Method for numeric vectors
#' @export
shrinking.default <- function(estCov, return_matrix = FALSE, target = NULL) {
  stopifnot(length(estCov) > 0, is.numeric(estCov), !any(is.na(estCov)), is.logical(return_matrix))
  estCorr <- estCov / estCov[1]
  corr_mat <- cyclic_matrix(estCorr)

  if(is.null(target)) {
    target <- diag(nrow(corr_mat))
  }
  else {
    stopifnot(is.matrix(target), is.numeric(target),  !any(is.na(target)))
  }

  optimRes <- optim(0.5, solve_shrinking, corr_mat = corr_mat, target = target, method = 'Brent', lower = 0, upper = 1)
  if(optimRes$convergence != 0) {
    stop(paste0("Failed to converge. Convergence code: ", optimRes$convergence))
  }

  adj_corr_mat <- (optimRes$par * corr_mat) + (1 - optimRes$par) * diag(nrow(corr_mat))

  if(return_matrix) {
    return(list('shrunken'=adj_corr_mat, 'lambda'=optimRes$par))
  }

  return(list('shrunken'=adj_corr_mat[1, ], 'lambda'=optimRes$par))
}

# Most basic call works.
test_that("corrected_est() works", {
  expect_equal(corrected_est(c(1, 2, 3), "gaussian")$acf, c(2/3, 0, -5.398656*10^(-7)))
})

test_that("corrected_est() fails for empty X", {
  expect_error(corrected_est(c(), "gaussian")$acf)
})

test_that("corrected_est() fails for nonnumeric X", {
  expect_error(corrected_est(c(1, 'a', 3), "gaussian")$acf)
  expect_error(corrected_est(c(1, 1i, 3), "gaussian")$acf)
})

test_that("corrected_est() fails for at least one NA in X", {
  expect_error(corrected_est(c(1, NA, 3), "gaussian")$acf)
})

test_that("corrected_est() fails if nonnumeric N_T", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", N_T = 'a')$acf)
  expect_error(corrected_est(c(1, 2, 3), "gaussian", N_T = 1i)$acf)
})

test_that("corrected_est() fails if N_T is NA", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", N_T = NA)$acf)
})

test_that("corrected_est() fails if N_T <= 0", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", N_T = 0)$acf)
  expect_error(corrected_est(c(1, 2, 3), "gaussian", N_T = -0.1)$acf)
})

test_that("corrected_est() fails for nonnumeric meanX", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", meanX = 'a')$acf)
  expect_error(corrected_est(c(1, 2, 3), "gaussian", meanX = 1i)$acf)
})

test_that("corrected_est() fails for NA meanX", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", meanX = NA)$acf)
})

test_that("corrected_est() fails if pd is nonboolean", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", pd = 'TRUE')$acf)
  expect_error(corrected_est(c(1, 2, 3), "gaussian", pd = 1)$acf)
})

test_that("corrected_est() fails if maxLag is nonnumeric", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", maxLag = 'a')$acf)
  expect_error(corrected_est(c(1, 2, 3), "gaussian", maxLag = 1i)$acf)
})

test_that("corrected_est() fails if maxLag < 0", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", maxLag = -1)$acf)
})

test_that("corrected_est() fails if maxLag >= length(X)", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", maxLag = 3)$acf)
})

test_that("corrected_est() fails for noninteger maxLag", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", maxLag = 1.5)$acf)
})

test_that("corrected_est() fails if maxLag is NA", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", maxLag = NA)$acf)
})

test_that("corrected_est() fails if meanX's length is not 1", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", meanX = c(1, 2))$acf)
})

test_that("corrected_est() fails for nonnumeric meanX", {
    expect_error(corrected_est(c(1, 2, 3), "gaussian", meanX = 'a')$acf)
    expect_error(corrected_est(c(1, 2, 3), "gaussian", meanX = 1i)$acf)
})

test_that("corrected_est() fails if meanX is NA", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", meanX = NA)$acf)
})

test_that("corrected_est() fails if type is neither 'autocovariance' or 'autocorrelation'", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", type = 'covariance')$acf)
})

test_that("corrected_est() fails if x has at least one NA", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", x = c(1, NA, 3))$acf)
})

test_that("corrected_est() fails for empty x", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", x = c())$acf)
})

test_that("corrected_est() fails for nonnumeric x", {
  expect_error(corrected_est(c(1, 2, 3), "gaussian", x = c(1, 'a', 3))$acf)
  expect_error(corrected_est(c(1, 2, 3), "gaussian", x = c(1, 1i, 3))$acf)
})

test_that("corrected_est() fails for nonboolean custom_kernel", {
  expect_error(corrected_est(c(1, 2, 3), "my_kernel", customKernel = 1)$acf)
  expect_error(corrected_est(c(1, 2, 3), "my_kernel", customKernel = 'TRUE')$acf)
})

# kernel_est
test_that("kernel_est() works", {
  expect_equal(kernel_est(c(1, 2, 3), "gaussian"), c(1, 0.071347986694504816896, 0.000004858790376937838049672957), tolerance = sqrt(.Machine$double.eps))
})

test_that("kernel_est() fails for nonboolean custom_kernel", {
  expect_error(kernel_est(c(1, 2, 3), "my_kernel", customKernel = 1))
  expect_error(kernel_est(c(1, 2, 3), "my_kernel", customKernel = 'TRUE'))
})

test_that("kernel_est() fails for empty estCov", {
expect_error(kernel_est(c(), "gaussian"))
})

test_that("kernel_est() fails for at least one NA in estCov", {
  expect_error(kernel_est(c(1, NA, 3), "gaussian"))
})

test_that("kernel_est() fails for nonnumeric estCov", {
  expect_error(kernel_est(c(1, 'a', 3), "gaussian"))
  expect_error(kernel_est(c(1, 1i, 3), "gaussian"))
})

test_that("kernel_est() fails for nonnumeric N_T", {
  expect_error(kernel_est(c(1, 2, 3), "gaussian", N_T = 'a'))
  expect_error(kernel_est(c(1, 2, 3), "gaussian", N_T = 1i))
})

test_that("kernel_est() fails for N_T <= 0", {
  expect_error(kernel_est(c(1, 2, 3), "gaussian", N_T = 0))
  expect_error(kernel_est(c(1, 2, 3), "gaussian", N_T = -0.1))
})

test_that("kernel_est() fails for NA N_T", {
  expect_error(kernel_est(c(1, 2, 3), "gaussian", N_T = NA))
})

test_that("kernel_est() fails for nonnumeric maxLag", {
  expect_error(kernel_est(c(1, 2, 3), "gaussian", maxLag='a'))
  expect_error(kernel_est(c(1, 2, 3), "gaussian", maxLag=1i))
})

test_that("kernel_est() fails for maxLag < 0", {
  expect_error(kernel_est(c(1, 2, 3), "gaussian", maxLag=-1))
})

test_that("kernel_est() fails for maxLag > length(X) - 1", {
  expect_error(kernel_est(c(1, 2, 3), "gaussian", maxLag=3))
})

test_that("kernel_est() fails for noninteger maxLag", {
  expect_error(kernel_est(c(1, 2, 3), "gaussian", maxLag=1.5))
})

test_that("kernel_est() fails for NA maxLag", {
  expect_error(kernel_est(c(1, 2, 3), "gaussian", maxLag = NA))
})

# solve_shrinking
test_that("solve_shrinking() fails for nonnumeric par", {
  expect_error(solve_shrinking('a', diag(2), diag(2)))
  expect_error(solve_shrinking(1i, diag(2), diag(2)))
})

test_that("solve_shrinking() fails for NA in par", {
  expect_error(solve_shrinking(NA, diag(2), diag(2)))
})

test_that("solve_shrinking() fails for par not of length 1", {
  expect_error(solve_shrinking(c(0.5, 0.5), diag(2), diag(2)))
})

test_that("solve_shrinking() fails for nonnumeric corr_mat", {
  expect_error(solve_shrinking(0.5, matrix(c(1, 'a', 3, 4), 2), diag(2)))
  expect_error(solve_shrinking(0.5, matrix(c(1, 2i, 3, 4), 2), diag(2)))
})

test_that("solve_shrinking() fails for at least one NA in corr_mat", {
  expect_error(solve_shrinking(0.5, matrix(c(1, NA, 3, 4), 2), diag(2)))
})

test_that("solve_shrinking() fails for nonnumeric target", {
  expect_error(solve_shrinking(0.5, diag(2), matrix(c(1, 'a', 3, 4), 2)))
  expect_error(solve_shrinking(0.5, diag(2), matrix(c(1, 2i, 3, 4), 2)))
})

test_that("solve_shrinking() fails for at least one NA in target", {
  expect_error(solve_shrinking(0.5, diag(2), matrix(c(1, NA, 3, 4), 2)))
})

test_that("solve_shrinking() fails for corr_mat and target having different sizes", {
  expect_error(solve_shrinking(0.5, diag(2), diag(3)))
})

# shrinking
test_that("shrinking() fails for estCov having length less than 1", {
  expect_error(shrinking(c()))
})

test_that("shrinking() fails for nonnumeric estCov", {
  expect_error(shrinking(c(1, 'a', 3)))
  expect_error(shrinking(c(1, 2i, 3)))
})

test_that("shrinking() fails for at least one NA in estCov", {
  expect_error(shrinking(c(1, NA, 3)))
})

test_that("shrinking() fails for nonlogical return_matrix", {
  expect_error(shrinking(c(1, 2, 3), return_matrix = 1))
  expect_error(shrinking(c(1, 2, 3), return_matrix = 'TRUE'))
})

test_that("shrinking() fails for nonmatrix target", {
  expect_error(shrinking(c(1, 2, 3), target = 1))
})

test_that("shrinking() fails for nonnumeric target", {
  expect_error(shrinking(c(1, 2, 3), target = matrix(c(1, 'a', 3, 4), 2)))
  expect_error(shrinking(c(1, 2, 3), target = matrix(c(1, 2i, 3, 4), 2)))
})

test_that("shrinking() fails for at least one NA in target", {
  expect_error(shrinking(c(1, 2, 3), target = matrix(c(1, NA, 3, 4), 2)))
})

# The function in this file are intended to be used with the tapered estimator.
# Despite being called windows, they are 'half' of the window.
# At 0, the window has a value of 0, at 1 it has a value of 1.

.tukey_window <- function(x) {
  return((1/2) * (1 - cos(pi * x)))
}

.triangular_window <- function(x) {
  return(x)
}

.sine_window <- function(x) {
  return(sin(pi * x / 2))
}

.power_sine_window <- function(x, params = 1) {
  return((sin(pi * x  / 2))^params[1])
}

.blackman_window <- function(x, params = 0.16) {
  return(((1 - params[1]) / 2)  - (1 / 2) * cos(pi * x) + (params[1] / 2) * cos(2 * pi * x))
}

.hann_poisson_window <- function(x, params = 1) {
  (1 / 2) * (1 - cos(pi * x)) * exp(- (params[1] * abs(1 - x)))
}

.welch_window <- function(x) {
  return(1 - (x - 1)^2)
}

.validate_window_params <- function(name = c("tukey", "triangular", "sine", "power_sine", "blackman", "hann_poisson", "welch"), params) {
  name <- match.arg(name)

  switch(name,
         tukey = TRUE,
         triangular = TRUE,
         sine = TRUE,
         power_sine = ifelse(params[1] > 0, TRUE, FALSE),
         blackman = ifelse(abs(params[1]) <= 0.25, TRUE, FALSE),
         hann_poisson = ifelse(params[1] > 0, TRUE, FALSE),
         welch = TRUE
  )
}

#' 1D Window Functions.
#'
#' A window function in this context is a continuous nondecreasing function such that at 0 it is 0, and at 1, it is 1.
#' Note that `window()` is **deprecated**, please use `window_ec()` instead.
#' This computes one of the window functions listed below.
#'
#' \strong{Tukey Window}.
#' The Tukey window is defined as
#' \deqn{w(x) = \frac{1}{2} - \frac{1}{2} \cos(\pi x) , x \in [0, 1].}
#' The \code{params} argument is empty.
#'
#' \strong{Triangular Window}.
#' The triangular window is given by
#' \deqn{w(x) = x, x \in [0, 1].}
#' The \code{params} argument is empty.
#'
#' \strong{Sine Window}.
#' The sine window is given by
#' \deqn{w(x) = \sin\left(\pi x / 2 \right), x \in [0, 1].}
#' The \code{params} argument is empty.
#'
#' \strong{Power Sine Window}.
#' The power sine window is given by
#' \deqn{w(x; a) = \sin^{a}(\pi x / 2), x \in [0, 1], a > 0.}
#' The \code{params} argument is of the form \code{c(}\eqn{a}\code{)}.
#'
#' \strong{Blackman Window}.
#' The Blackman window is defined as
#' \deqn{w(x; a) = ( (1 - a) / 2) - \frac{1}{2} \cos(\pi x) + \frac{a}{2} \cos(2 \pi x), x \in [0, 1], a \in [-0.25, 0.25] .}
#' The \code{params} argument is of the form \code{c(}\eqn{a}\code{)}.
#' It is recommended that \eqn{a \in [-0.25, 0.25]} to ensure that the window is nondecreasi#'ng on \eqn{[0, 1].}
#'
#' \strong{Hann-Poisson Window}.
#' The Hann-Poisson window is defined as
#' \deqn{w(x; a) = \frac{1}{2} (1 - \cos(\pi x)) \exp( - (a \left|1 - x \right|) ) , x \in [0, 1], a > 0.}
#' The \code{params} argument is of the form \code{c(}\eqn{a}\code{)}.
#'
#' \strong{Welch Window}.
#' The Welch window is given by
#' \deqn{w(x) = 1 - (x - 1)^2 , x \in [0, 1] . }
#' The \code{params} argument is empty.
#'
#' See the function call examples below.
#'
#' @param x A vector or matrix of arguments of at least length 1. Each value must be between 0 and 1, inclusive.
#' @param name The name of the window. Options are: tukey, triangular, sine, power_sine, blackman, hann_poisson, welch.
#' @param params A vector of parameters for the windows. See the documentation below for the position of the parameters.
#'
#' @return A vector or matrix of values.
#' @export
#'
#' @rdname window_ec
#'
#' @examples
#' x <- c(0.2, 0.4, 0.6)
#' window_ec(x, "tukey")
#' window_ec(x, "triangular")
#' window_ec(x, "sine")
#' window_ec(x, "power_sine", c(0.7))
#' window_ec(x, "blackman", c(0.16))
#' window_ec(x, "hann_poisson", c(0.7))
#' window_ec(x, "welch")
#' curve(window_ec(x, "tukey"), from = 0, to = 1)
#' curve(window_ec(x, "triangular"), from = 0, to = 1)
#' curve(window_ec(x, "sine"), from = 0, to = 1)
#' curve(window_ec(x, "power_sine", c(0.7)), from = 0, to = 1)
#' curve(window_ec(x, "blackman", c(0.16)), from = 0, to = 1)
#' curve(window_ec(x, "hann_poisson", c(0.7)), from = 0, to = 1)
#' curve(window_ec(x, "welch"), from = 0, to = 1)
window_ec <- function(x, name = c("tukey", "triangular", "sine", "power_sine", "blackman", "hann_poisson", "welch"), params=c(1)) {
  stopifnot(length(x) >= 1, is.numeric(x), all(x >= 0), all(x <= 1), length(params) > 0, is.numeric(params[1]))

  name <- match.arg(name)

  if(!.validate_window_params(name, params)) {
    stop("Parameter choice `(", paste0(params, collapse = ", "), ")` is not appropriate for `", name, "` window.")
  }

  returnVal <- switch(name,
         tukey = .tukey_window(x),
         triangular = .triangular_window(x),
         sine = .sine_window(x),
         power_sine = .power_sine_window(x, params),
         blackman = .blackman_window(x, params),
         hann_poisson = .hann_poisson_window(x, params),
         welch = .welch_window(x)
  )

  return(returnVal)
}

#' 1D Symmetric Window Functions.
#'
#' A symmetric window function in this context are traditional window functions, unlike the [window] functions.
#' Note that `window_symm()` is **deprecated**, please use `window_symm_ec()` instead.
#' This computes one of the symmetric window functions listed below, all of which are defined for \eqn{x \in [-1, 1],} and are 0 otherwise.
#'
#' \strong{Tukey Window}.
#' The Tukey window is defined as
#' \deqn{w(x) = \frac{1}{2} + \frac{1}{2} \cos(\pi |x|) , x \in [-1, 1].}
#' The \code{params} argument is empty, see the example.
#'
#' \strong{Triangular Window}.
#' The triangular window is given by
#' \deqn{w(x) = 1 - |x|, x \in [-1, 1].}
#' The \code{params} argument is empty, see the example.
#'
#' \strong{Sine Window}.
#' The sine window is given by
#' \deqn{w(x) = 1 - \sin\left(\pi |x| / 2 \right), x \in [-1, 1].}
#' The \code{params} argument is empty, see the example.
#'
#' \strong{Power Sine Window}.
#' The power sine window is given by
#' \deqn{w(x; a) = 1 - \sin^{a}(\pi |x| / 2), x \in [-1, 1], a > 0.}
#' The \code{params} argument is of the form \code{c(}\eqn{a}\code{)}
#'
#' \strong{Blackman Window}.
#' The Blackman window is defined as
#' \deqn{w(x; a) = 1 +  ( (a - 1) / 2) + \frac{1}{2} \cos(\pi |x|) - \frac{a}{2} \cos(2 \pi |x|), x \in [-1, 1], a \in [-0.25, 0.25] .}
#' The \code{params} argument is of the form \code{c(}\eqn{a}\code{)}.
#' The standard value of \eqn{a} for this window is \eqn{0.16.}
#'
#' \strong{Hann-Poisson Window}.
#' The Hann-Poisson window is defined as
#' \deqn{w(x; a) = 1 - \frac{1}{2} (1 - \cos(\pi |x|)) \exp( - (a \left|1 - |x| \right|) ) , x \in [-1, 1], a > 0 .}
#' The \code{params} argument is of the form \code{c(}\eqn{a}\code{)}
#'
#' \strong{Welch Window}.
#' The Welch window is given by
#' \deqn{w(x) = (|x| - 1)^2 , x \in [-1, 1] . }
#' The \code{params} argument is empty, see the example.
#'
#' @param x A vector or matrix of arguments of at least length 1. Each value must be between 0 and 1, inclusive.
#' @param name The name of the window. Options are: tukey, triangular, sine, power_sine, blackman, hann_poisson, welch.
#' @param params A vector of parameters for the windows. See the documentation below for the position of the parameters.
#'
#' @return A vector or matrix of values.
#' @export
#'
#' @rdname window_symm_ec
#'
#' @examples
#' x <- c(-0.6, -0.4, -0.2, 0, 0.2, 0.4, 0.6)
#' window_symm_ec(x, "tukey")
#' window_symm_ec(x, "triangular")
#' window_symm_ec(x, "sine")
#' window_symm_ec(x, "power_sine", c(0.7))
#' window_symm_ec(x, "blackman", c(0.16))
#' window_symm_ec(x, "hann_poisson", c(0.7))
#' window_symm_ec(x, "welch")
#' curve(window_symm_ec(x, "tukey"), from = -1, to = 1)
#' curve(window_symm_ec(x, "triangular"), from = -1, to = 1)
#' curve(window_symm_ec(x, "sine"), from = -1, to = 1)
#' curve(window_symm_ec(x, "power_sine", c(0.7)), from = -1, to = 1)
#' curve(window_symm_ec(x, "blackman", c(0.16)), from = -1, to = 1)
#' curve(window_symm_ec(x, "hann_poisson", c(0.7)), from = -1, to = 1)
#' curve(window_symm_ec(x, "welch"), from = -1, to = 1)
window_symm_ec <- function(x, name = c("tukey", "triangular", "sine", "power_sine", "blackman", "hann_poisson", "welch"), params=c(1)) {
  stopifnot(length(x) >= 1, is.numeric(x), !any(is.na(x)), length(params) > 0, is.numeric(params[1]))

  name <- match.arg(name)

  if(!.validate_window_params(name, params)) {
    stop("Parameter choice `(", paste0(params, collapse = ", "), ")` is not appropriate for `", name, "` symmetric window.")
  }

  returnVal <- switch(name,
         tukey = 1 - .tukey_window(abs(x)),
         triangular = 1 - .triangular_window(abs(x)),
         sine = 1 - .sine_window(abs(x)),
         power_sine = 1 - .power_sine_window(abs(x), params),
         blackman = 1 - .blackman_window(abs(x), params),
         hann_poisson = 1 - .hann_poisson_window(abs(x), params),
         welch = 1 - .welch_window(abs(x))
  )
  returnVal[abs(x) > 1] <- 0

  return(returnVal)
}

# Tukey window
test_that("window_ec(, \"tukey\") works", {
  expect_equal(window_ec(c(1, 2, 3) / 3, "tukey"), c(0.25, 0.75, 1))
})

test_that("window_ec(, \"tukey\") fails for empty x", {
  expect_error(window_ec(c(), "tukey"))
})

test_that("window_ec(, \"tukey\") fails for a single negative x", {
  expect_error(window_ec(c(0.1 , 0.3, -0.02), "tukey"))
})

test_that("window_ec(, \"tukey\") fails for a single x above 1", {
  expect_error(window_ec(c(0.1 , 0.3, 1.02), "tukey"))
})

# Triangular window
test_that("window_ec(, \"trianguar\") works", {
  expect_equal(window_ec(c(1, 2, 3) / 4, "triangular"), c(0.25, 0.5, 0.75))
})

test_that("window_ec(, \"trianguar\") fails for empty x", {
  expect_error(window_ec(c(), "triangular"))
})

test_that("window_ec(, \"trianguar\") fails for a single negative x", {
  expect_error(window_ec(c(0.1 , 0.3, -0.02), "triangular"))
})

test_that("window_ec(, \"trianguar\") fails for a single x above 1", {
  expect_error(window_ec(c(0.1 , 0.3, 1.02), "triangular"))
})

# Sine window
test_that("window_ec(, \"sine\") works", {
  expect_equal(window_ec(c(0, 1/3, 1/2, 2/3), "sine"), c(0, sin(pi/6), sin(pi/4), sin(pi/3)))
})

test_that("window_ec(, \"sine\") fails for empty x", {
  expect_error(window_ec(c(), "sine"))
})

test_that("window_ec(, \"sine\") fails for a single negative x", {
  expect_error(window_ec(c(0.1 , 0.3, -0.02), "sine"))
})

test_that("window_ec(, \"sine\") fails for a single x above 1", {
  expect_error(window_ec(c(0.1 , 0.3, 1.02), "sine"))
})

# Power sine window
test_that("window_ec(, \"power_sine\", ) works", {
  expect_equal(window_ec(c(0, 1/3, 1/2, 2/3), "power_sine", c(2)), c(0, 0.25, 0.5, 0.75))
})

test_that("window_ec(, \"power_sine\", ) fails for empty x", {
  expect_error(window_ec(c(), "power_sine", c(2)))
})

test_that("window_ec(, \"power_sine\", ) fails for a single negative x", {
  expect_error(window_ec(c(0.1 , 0.3, -0.02), "power_sine", c(2)))
})

test_that("window_ec(, \"power_sine\", ) fails for a single x above 1", {
  expect_error(window_ec(c(0.1 , 0.3, 1.02), "power_sine", c(2)))
})

test_that("window_ec(, \"power_sine\", ) fails for a <= 0", {
  expect_error(window_ec(c(0.1 , 0.3, 1.02), "power_sine", c(0)))
  expect_error(window_ec(c(0.1 , 0.3, 1.02), "power_sine", c(-1)))
})

# Blackman window
test_that("window_ec(, \"blackman\", ) works", {
  expect_equal(window_ec(c(0, 1/3, 1/2, 2/3), "blackman", c(0.16)), c(-1.387779e-17, 1.300000e-01, 3.400000e-01, 6.300000e-01))
})

test_that("window_ec(, \"blackman\", ) fails for empty x", {
  expect_error(window_ec(c(), "blackman", c(2)))
})

test_that("window_ec(, \"blackman\", ) fails for a single negative x", {
  expect_error(window_ec(c(0.1 , 0.3, -0.02), "blackman", c(2)))
})

test_that("window_ec(, \"blackman\", ) fails for a single x above 1", {
  expect_error(window_ec(c(0.1 , 0.3, 1.02), "blackman", c(2)))
})

test_that("window_ec(, \"blackman\", ) fails for a nonreal alpha.", {
  expect_error(window_ec(c(0.1 , 0.3, 1.02), "blackman", c(1i)))
})

# Hann-Poisson window
test_that("window_ec(, \"hann_poisson\", ) works", {
  expect_equal(window_ec(c(0, 1/3, 1/2, 2/3), "hann_poisson", c(0.5)), c(0, (1 / 2) * (1 - cos(pi * (1/3))) * exp(- (0.5 * abs(1 - (1/3)))),
                                                                (1 / 2) * (1 - cos(pi * (1/2))) * exp(- (0.5 * abs(1 - (1/2)))),
                                                                ((1 / 2) * (1 - cos(pi * (2/3))) * exp(- (0.5 * abs(1 - (2/3)))))))
})

test_that("window_ec(, \"hann_poisson\", ) fails for empty x", {
  expect_error(window_ec(c(), "hann_poisson", c(2)))
})

test_that("window_ec(, \"hann_poisson\", ) fails for a single negative x", {
  expect_error(window_ec(c(0.1 , 0.3, -0.02), "hann_poisson", c(2)))
})

test_that("window_ec(, \"hann_poisson\", ) fails for a single x above 1", {
  expect_error(window_ec(c(0.1 , 0.3, 1.02), "hann_poisson", c(2)))
})

test_that("window_ec(, \"blackman\", ) fails for a nonreal alpha.", {
  expect_error(window_ec(c(0.1 , 0.3, 1.02), "hann_poisson", c(1i)))
})

# Welch window
test_that("window_ec(, \"welch\") works", {
  expect_equal(window_ec(c(0, 1/3, 1/2, 2/3), "welch"), c(0, 1 - ((1/3)-1)^2, 0.75, 1 - ((2/3)-1)^2))
})

test_that("window_ec(, \"welch\") fails for empty x", {
  expect_error(window_ec(c(), "welch"))
})

test_that("window_ec(, \"welch\") fails for a single negative x", {
  expect_error(window_ec(c(0.1 , 0.3, -0.02), "welch"))
})

test_that("window_ec(, \"welch\") fails for a single x above 1", {
  expect_error(window_ec(c(0.1 , 0.3, 1.02), "welch"))
})

test_that("window_symm_ec(, \"tukey\") works", {
  expect_equal(window_symm_ec(c(1, 2, 3) / 3, "tukey"), c(0.75, 0.25, 0))
})

test_that("window_symm_ec(, \"tukey\") fails for at least one NA in x", {
  expect_error(window_symm_ec(c(1, NA, 3) / 3, "tukey"))
})

test_that("window_symm_ec(, \"tukey\") fails for nonnumeric x", {
  expect_error(window_symm_ec(c(1, 'a', 3), "tukey"))
  expect_error(window_symm_ec(c(1, 1i, 3) / 3, "tukey"))
})

test_that("window_symm_ec(, \"tukey\") fails empty x", {
  expect_error(window_symm_ec(c(), "tukey"))
})

test_that('.validate_window_params() works', {
  expect_equal(.validate_window_params("tukey", c(1)), TRUE)
})

test_that('.validate_window_params() fails for name not in "tukey", "triangular", "sine", "power_sine", "blackman", "hann_poisson", "welch"', {
  expect_error(.validate_window_params("test", c(1)))
})

test_X <- c(1, 2, 3)
test_meanX <- mean(test_X)
test_h2n <- H2n(length(test_X), 1, "tukey")
test_n <- 3
test_taperVals_t <- taper(((1:test_n) - 1/2)/test_n, 1, "tukey")
test_taperVals_h <- taper((((1:(test_n-1)) - 1/2) +1) / test_n, 1, "tukey")

#' Computes the Standard Estimator of the Autocovariance Function.
#'
#' This function computes the following two estimates of the autocovariance function depending on
#' the parameter \code{pd}, see the Details section.
#'
#' For \code{pd = TRUE}:
#' \deqn{
#' \widehat{C}(h) = \frac{1}{N} \sum_{j=1}^{N-h} ( X(j) - \bar{X} ) ( X(j + h) - \bar{X} ) .
#' }
#'
#'For \code{pd = FALSE}:
#'\deqn{
#' \widehat{C}(h) = \frac{1}{N - h} \sum_{j=1}^{N-h} ( X(j) - \bar{X} ) ( X(j + h) - \bar{X} ) .
#' }
#'
#' This function will generate autocovariance values for lags \eqn{h} from the set \eqn{\{0, \dots, \mbox{maxLag}\}.}
#'
#' The positive-definite estimator must be used cautiously when estimating over all lags as the sum of all values of the autocorrelation function equals to \eqn{-1/2}.
#' For the nonpositive-definite estimator a similar constant summation property holds.
#'
#' @references
#' Bilchouris, A. & Olenko, A (2025). On Nonparametric Estimation of Covariogram. Austrian Statistical Society 54(1), 112-137. https://doi.org/10.17713/ajs.v54i1.1975
#'
#' @param X A vector representing observed values of the time series.
#' @param pd Whether a positive-definite estimate should be used. Defaults to \code{TRUE}.
#' @param meanX The average value of \code{X}. Defaults to \code{mean(X)}.
#' @param x A vector of lag indices. Defaults to the sequence \code{0:length(X)}. Must be at least as large as \code{maxLag + 1}.
#' @param maxLag An optional parameter that determines the maximum lag to compute the estimated autocovariance function at. Defaults to \code{length(X) - 1}.
#' @param type Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.
#'
#' @return A \code{CovEsts} S3 object (list) with the following values
#' \describe{
#'  \item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
#'  \item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
#'  \item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the \code{type} parameter.}
#'  \item{\code{est_used}}{The estimator function used, in this case, 'standard_est'.}
#' }
#' @export
#'
#' @importFrom stats acf
#'
#' @examples
#' X <- c(1, 2, 3)
#' standard_est(X, pd = FALSE, maxLag = 2, meanX = mean(X))
standard_est <- function(X, pd = TRUE, maxLag = length(X) - 1, x = 0:length(X), type = c("autocovariance", "autocorrelation"), meanX = mean(X)) {
  type <- match.arg(type)
  stopifnot(length(X) > 0, is.numeric(X), is.logical(pd), maxLag >= 0, maxLag <= (length(X) - 1),
            maxLag %% 1 == 0, length(meanX) == 1, is.numeric(meanX), !is.na(meanX),
            is.numeric(x), all(!is.na(x)), length(x) > 0)
  retVec <- as.vector(stats::acf(X - meanX, lag.max = maxLag, type = "covariance", plot = FALSE, demean = FALSE)$acf)
  if(!pd) {
    retVec <- retVec * (length(X) / (length(X) - 0:maxLag))
  }
  if(type == 'autocorrelation') {
    retVec <- retVec / retVec[1]
  }

  res <- list(acf=retVec, lags=x[1:(maxLag + 1)], est_type=type, est_used='standard_est')
  return(structure(res, class = "CovEsts"))
}

#' Autocovariance to Semivariogram
#'
#' This function computes an estimated semivariogram using an estimated autocovariance function.
#'
#' @details
#' The semivariogram, \eqn{\gamma(h)} and autocovariance function, \eqn{C(h)}, under the assumption of weak stationarity are related as follows:
#' \deqn{
#' \gamma(h) = C(0) - C(h) .
#' }
#'
#' When an empirical autocovariance function is considered instead, this relation does not necessarily hold, however,
#' it can be used to obtain a function that is close to a semivariogram, see Bilchouris and Olenko (2025).
#'
#' @references Bilchouris, A. & Olenko, A (2025). On Nonparametric Estimation of Covariogram. Austrian Statistical Society 54(1), 112-137. 10.17713/ajs.v54i1.1975
#'
#' @param estCov A vector whose values are an estimate autocovariance function or a \code{CovEsts} S3 object.
#'
#' @return A numeric vector whose values are an estimate of the semivariogram or an \code{VarioEsts} S3 object (list) with the following values
#' \describe{
#'  \item{\code{vario}}{A numeric vector whose values are an estimate of the semivariogram.}
#'  \item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on, inherited from the argument \code{estCov}.}
#'  \item{\code{est_used}}{The estimator function used, in this case, 'to_vario'.}
#' }
#' If a numeric vector is given for the argument \code{estCov}, then a numeric vector output is given, and if a \code{CovEsts} S3 object is given, a \code{VarioEsts} object is given as output.
#' @export
#'
#' @examples
#' X <- c(1, 2, 3)
#' estCov <- standard_est(X, meanX=mean(X), maxLag = 2, pd=FALSE)
#' to_vario(estCov)
to_vario <- function(estCov) {
  UseMethod("to_vario")
}

#' @describeIn to_vario Method for numeric vectors.
#' @export
to_vario.default <- function(estCov) {
  stopifnot(length(estCov) > 0, is.numeric(estCov), !any(is.na(estCov)))
  return(estCov[1] - estCov)
}

#' @describeIn to_vario Method for `CovEsts` objects.
#' @export
to_vario.CovEsts <- function(estCov) {
  ret_vec <- to_vario.default(estCov$acf)

  res <- list(vario=ret_vec, lags=estCov$lags, est_used='to_vario')
  return(structure(res, class = "VarioEsts"))
}

#' Computes the Standard Estimator of the Autocovariance Function.
#'
#' This function computes the partial autocorrelation function from an estimated autocovariance or autocorrelation function.
#'
#' @details
#' This function is a translation of the 'uni_pacf' function in src/library/stats/src/pacf.c of the R source code which is an implementation of the Durbin-Levinson algorithm.
#'
#' @param estCov A numeric vector representing an estimated autocovariance or autocorrelation function.
#'
#' @return A vector whose values are an estimate partial autocorrelation function or a \code{CovEsts} S3 object (list) with the following values
#' \describe{
#'  \item{\code{acf}}{A numeric vector containing the partial autocorrelation estimates.}
#'  \item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
#'  \item{\code{est_type}}{The type of estimate, namely 'partial'.}
#'  \item{\code{est_used}}{The estimator function used, in this case, 'to_pacf'.}
#' }
#' If a numeric vector is given for the argument \code{estCov}, then a numeric vector output is given, and if a \code{CovEsts} S3 object is given, a \code{CovEsts} object is given as output.
#' @export
#'
#' @examples
#' X <- c(1, 2, 3)
#' to_pacf(standard_est(X, pd = FALSE, maxLag = 2, meanX = mean(X)))
to_pacf <- function(estCov) {
  UseMethod("to_pacf")
}

#' @describeIn to_pacf Method for `CovEsts` objects.
#' @export
to_pacf.CovEsts <- function(estCov) {
  pacf_vec <- to_pacf.default(estCov$acf)

  res <- list(acf=pacf_vec, lags=1:length(pacf_vec), est_type='partial', est_used='to_pacf')
  return(structure(res, class = "CovEsts"))
}

#' @describeIn to_pacf Method for numeric vectors.
#' @export
to_pacf.default <- function(estCov) {
  stopifnot(length(estCov) > 0, is.numeric(estCov), !any(is.na(estCov)))

  # Force to autocorrelation
  estCov <- estCov / estCov[1]

  # This function is a translation of the 'uni_pacf' function in src/library/stats/src/pacf.c of the R source code.
  n <- length(estCov)
  if(n == 2) {
    return(c(estCov[2]))
  }

  phi <- c(estCov[2], rep(NA, n - 2))
  temp <- rep(NA, n - 1)
  pacf_vec <- numeric(n - 2)
  pacf_vec[1] <- estCov[2]

  for(i in 2:(n - 1)) {
    temp <- phi[1:(i - 1)]
    numer <- estCov[i + 1]
    denom <- estCov[1]

    for(k in 1:(i - 1)) {
      numer <- numer - (phi[k] * estCov[i - k + 1])
      denom <- denom - (phi[k] * estCov[k + 1])
    }

    pacf_val <- numer / denom
    pacf_vec[i] <- pacf_val
    phi[i] <- pacf_val

    if (i == (n - 1)) {
      break
    }

    # update the phi for next run
    # phi_{n, k} <- phi_{n - 1, k} - (phi_{n, n} * phi_{n - 1, n - k})
    for(k in 1:(i - 1)) {
      phi[k] <- temp[k] - (pacf_val * temp[i - k])
    }
  }

  return(pacf_vec)
}

#' Normalise a CovEsts Object
#'
#' @description
#' This function normalises a CovEsts S3 object, that is, turning an autocovariance function into an autocorrelation function.
#' This procedure is a one-way transformation.
#' This function does not support any other type of argument.
#'
#' @param estCov A CovEst S3 object.
#'
#' @return A CovEsts S3 object with a normalised \code{acf} attribute.
#' @export
#'
#' @examples
#' normalise_acf(standard_est(c(1, 2, 3)))
normalise_acf <- function(estCov) {
  UseMethod("normalise_acf")
}

#' @describeIn normalise_acf Method for other objects.
#' @export
normalise_acf.default <- function(estCov) {
  stop("Input must be of class 'CovEsts'. Objects of class ", class(estCov), " are not supported.")
}

#' @describeIn normalise_acf Method for 'CovEsts' objects.
#' @export
normalise_acf.CovEsts <- function(estCov) {
  estCov$acf <- estCov$acf / estCov$acf[1]

  estCov$est_type <- "autocorrelation"

  return(estCov)
}

# standard_est works
test_that("standard_est() works", {
  expect_equal(standard_est(c(1, 2, 3))$acf, c(2/3, 0, -1/3))
})

# standard_est fails for empty X
test_that("standard_est() fails for empty X", {
  expect_error(standard_est(c())$acf)
})

# standard_est fails for nonnumeric X
test_that("standard_est() fails for empty X", {
  expect_error(standard_est(c(1, 'a', 3))$acf)
  expect_error(standard_est(c(1, 1i, 3))$acf)
})

# standard_est fails for nonboolean pd
test_that("standard_est() for nonboolean pd", {
  expect_error(standard_est(c(1, 2, 3), pd='TRUE')$acf)
  expect_error(standard_est(c(1, 2, 3), pd=1)$acf)
})

# standard_est fails for maxLag < 0
test_that("standard_est() for maxLag < 0", {
  expect_error(standard_est(c(1, 2, 3), maxLag = -1)$acf)
})

# standard_est fails for maxLag > length(X) - 1
test_that("standard_est() for maxLag > length(X) - 1", {
  expect_error(standard_est(c(1, 2, 3), maxLag = 3)$acf)
})

# standard_est fails for noninteger maxLag
test_that("standard_est() for noninteger maxLag", {
  expect_error(standard_est(c(1, 2, 3), maxLag = 1.5)$acf)
})

# standard_est fails for meanX not of length 1
test_that("standard_est() for meanX not of length 1", {
  expect_error(standard_est(c(1, 2, 3), meanX = c(1, 2))$acf)
})

# standard_est fails for nonnumeric meanX
test_that("standard_est() for nonnumeric meanX", {
  expect_error(standard_est(c(1, 2, 3), meanX = 'a')$acf)
  expect_error(standard_est(c(1, 2, 3), meanX = 1i)$acf)
})

# standard_est fails for NA meanX
test_that("standard_est() for NA meanX", {
  expect_error(standard_est(c(1, 2, 3), meanX = NA)$acf)
})

# standard_est fails if type is neither autocovariance or autocorrelation
test_that("standard_est() for NA meanX", {
  expect_error(standard_est(c(1, 2, 3), type = 'covariance')$acf)
})

# standard_est fails for empty x
test_that("standard_est() fails for empty x", {
  expect_error(standard_est(c(1, 2, 3), x = c())$acf)
})

# standard_est fails for nonnumeric x
test_that("standard_est() fails nonnumeric x", {
  expect_error(standard_est(c(1, 2, 3), x = c(1, 'a', 3))$acf)
  expect_error(standard_est(c(1, 2, 3), x = c(1, 1i, 3))$acf)
})

# standard_est fails for at least one NA in x
test_that("standard_est() fails for at least one NA in", {
  expect_error(standard_est(c(1, 2, 3), x = c(1, NA, 3))$acf)
})

# to_vario works
test_that("to_vario() works", {
  expect_equal(to_vario(c(1, 0.5, 0)), c(0, 0.5, 1))
})

# to_vario fails if estCov is empty
test_that("to_vario() fails if estCov is empty", {
  expect_error(to_vario(c()))
})

# to_vario fails if at least one value in estCov is nonnumeric
test_that("to_vario() fails if at least one value in estCov is nonnumeric", {
  expect_error(to_vario(c(1, 'a', 3)))
  expect_error(to_vario(c(1, 1i, 3)))
})

# to_vario fails if estCov has at least one NA
test_that("to_vario() fails f estCov has at least one NA", {
  expect_error(to_vario(c(1, NA, 3)))
})

# to_pacf works
test_that("to_pacf() works", {
  expect_equal(to_pacf(c(1, 0.5, 0)), c(0.5, -1/3))
})

# to_pacf fails if estCov is empty
test_that("to_pacf() fails if estCov is empty", {
  expect_error(to_pacf(c()))
})

# to_pacf fails if at least one value in estCov is nonnumeric
test_that("to_pacf() fails if at least one value in estCov is nonnumeric", {
  expect_error(to_pacf(c(1, 'a', 3))$acf)
  expect_error(to_pacf(c(1, 1i, 3))$acf)
})

# to_pacf fails if estCov has at least one NA
test_that("to_pacf() fails if estCov has at least one NA", {
  expect_error(to_pacf(c(1, NA, 3))$vario)
})

test_p <- 2
test_m <- 3
test_kVec <- generate_knots(test_m)
test_taus <- get_taus(test_p, test_m)
test_x <- c(1, 2, 3)
test_X <- c(1, 2, 3)

test_estCov <- standard_est(test_X, maxLag=length(test_X) - 1)

#' Random Block Locations
#'
#' This function performs random sampling to obtain random starting locations for block bootstrap.
#'
#' @details
#' This function performs random sampling to obtain random starting locations for block bootstrap.
#' If \code{type = 'moving'}, the set \eqn{\{1, \dots, N - \ell  + 1\}} is randomly sampled, with replacement, \eqn{k} times to obtain random block locations for moving block bootstrap.
#' If \code{type = 'circular'}, the set \eqn{\{1, \dots, N\}} is randomly sampled, with replacement, \eqn{k} times to obtain random block locations for moving block bootstrap.
#'
#' @param N The length of the observation window.
#' @param l The block length considered for bootstrap.
#' @param k The number of blocks considered for bootstrap.
#' @param boot_type What type of block bootstrap should be used, either 'moving' for moving block bootstrap or 'circular' for circular block bootstrap.
#'
#' @references
#' Chapters 2.5 and 2.7 in Lahiri, S. N. (2003). Resampling Methods for Dependent Data. Springer. https://doi.org/10.1007/978-1-4757-3803-2
#'
#' Künsch, H. R. (1989). The Jackknife and the Bootstrap for General Stationary Observations. The Annals of Statistics 17(3), 1217-1241. https://doi.org/10.1214/aos/1176347265
#'
#' Politis, D. N. & Romano, J. P. (1991). A Circular Block-Resampling Procedure for Stationary Data. In R. LePage & L. Billard, eds, Exploring the Limits of Bootstrap, Wiley, 263-270.
#'
#' @return A vector of length \code{k} whose values are random block locations.
#' @export
#'
#' @examples
#' starting_locs(4, 2, 2)
starting_locs <- function(N, l, k, boot_type = c("moving", "circular")) {
  stopifnot(is.numeric(N), N >= 1, !is.na(N), is.numeric(l), l >= 1, !is.na(l),
            is.numeric(k), k >= 1, !is.na(k))

  boot_type <- match.arg(boot_type)

  sampleVec <- switch(boot_type,
                      moving = 1:(N - l + 1),
                      circular = 1:N)

  return(sample(sampleVec, k, replace = TRUE))
}

#' Block Bootstrap Sample
#'
#' This function generates block bootstrap samples for either moving block bootstrap or circular bootstrap.
#'
#' @details
#' This function generates a block bootstrap sample for a time series \eqn{X.}
#' For the moving block bootstrap and circular bootstrap procedures see [block_bootstrap] and the included references.
#'
#' @references
#' Chapters 2.5 and 2.7 in Lahiri, S. N. (2003). Resampling Methods for Dependent Data. Springer. https://doi.org/10.1007/978-1-4757-3803-2
#'
#' Künsch, H. R. (1989). The Jackknife and the Bootstrap for General Stationary Observations. The Annals of Statistics 17(3), 1217-1241. https://doi.org/10.1214/aos/1176347265
#'
#' Politis, D. N. & Romano, J. P. (1991). A Circular Block-Resampling Procedure for Stationary Data. In R. LePage & L. Billard, eds, Exploring the Limits of Bootstrap, Wiley, 263-270.
#'
#' @param X A vector representing observed values of the time series.
#' @param l The block length considered for bootstrap.
#' @param k The number of blocks considered for bootstrap.
#' @param boot_type What type of block bootstrap should be used, either 'moving' for moving block bootstrap or 'circular' for circular block bootstrap.
#'
#' @return A vector of length \code{length(X)} whose values are a bootstrapped time series.
#' @export
#'
#' @examples
#' X <- c(1, 2, 3, 3, 2, 1)
#' bootstrap_sample(X, 2, 3)
bootstrap_sample <- function(X, l, k, boot_type = c("moving", "circular")) {
  stopifnot(is.numeric(X), length(X) >= 1, !is.na(X), is.numeric(l), l >= 1, !is.na(l),
            is.numeric(k), k >= 1, !is.na(k))
  boot_type <- match.arg(boot_type)

  N <- length(X)
  S <- starting_locs(N, l, k, boot_type)
  block_offsets <- 0:(l-1)

  if(boot_type == 'moving') {
    indices <- outer(S, block_offsets, "+")
  }

  else if(boot_type == 'circular') {
    indices <- outer(S, block_offsets, "+")
    indices <- ((indices - 1) %% N) + 1
  }

  Y <- X[t(indices)]
  Y <- Y[1:N]
  return(Y)
}

#' Block Bootstrap
#'
#' This function performs block bootstrap (moving or circular) to obtain a \eqn{(1-\alpha)\%} confidence-interval for the autocovariance function. It will also compute average autocovariance function across all bootstrapped estimates.
#'
#' @details
#' This function performs block bootstrap to obtain a \eqn{(1-\alpha)\%} confidence-interval for the autocovariance function. It will also compute average autocovariance function across all bootstrapped estimates.
#'
#' Moving block bootstrap can be described as follows.
#' For some times series \eqn{X(1), X(2), \dots, X(n),} construct \eqn{k \in N} overlapping blocks of the form
#' \eqn{B_{i} = (X(i), \dots, X(i + \ell - 1)),} where \eqn{\ell \in \{1, \dots , n\}} is the block length.
#' Randomly sample, with replacement, from the discrete uniform distribution with on \eqn{\{1, \dots, n - \ell + 1\}} to obtain a set of random starting locations \eqn{I_{1}, \dots, I_{k}.}
#' Construct a bootstrapped time series \eqn{B_{1}^{\ast}, B_{2}^{\ast}, \dots, B_{k}^{\ast},} where \eqn{B_{i}^{\ast} = B_{I_{i}}.}
#' The bootstrapped time series is truncated to have length \eqn{n,} and will be of the form \eqn{X^{\ast}(1), \dots , X^{\ast}(n).}
#' The autocovariance function is then computed for the bootstrapped time series.
#'
#' An alternative to moving block bootstrap is circular block bootstrap.
#' Circular block bootstrap uses the time series like a circle, that is, the observation at \eqn{n + i} is the same as the observation at location \eqn{i.}
#' For example, for the block \eqn{B_{n - \ell + 2}}, we obtain \eqn{(X(n - \ell + 2) , \dots , X(n), X(n + 1))} is the same as \eqn{(X(n - \ell + 2) , \dots , X(n), X(1)).}
#' When performing random sampling to obtain starting locations, the set \eqn{\{1, \dots, n\}} is now considered.
#' The procedure for constructing the bootstrap time series after constructing the blocks and selecting the starting indices is the same as moving block bootstrap.
#'
#' This process is repeated \code{n_bootstrap} times to obtain \code{n_boostrap} estimates of the autocovariance function using the bootstrapped time series, for which the average autocovariance function
#' and the \eqn{(1 - \alpha)\%} confidence intervals are constructed pointwise for each lag.
#'
#' The choice of the block length, \eqn{\ell ,} depends on the degree of dependence present in the time series. If the time series has a high degree of dependence, a larger block size should be chosen to ensure the dependency structure is maintained within the block.
#'
#' Any estimator of the autocovariance function can be used in this function, including a custom estimator not in the package, see the examples.
#'
#' When \code{parallel = TRUE}, L'Ecuyer-CMRG is used to ensure independent random numbers across workers. The serial version uses the Mersenne-Twister algorithm.
#'
#' @references
#' Chapters 2.5 and 2.7 in Lahiri, S. N. (2003). Resampling Methods for Dependent Data. Springer. https://doi.org/10.1007/978-1-4757-3803-2
#'
#' Künsch, H. R. (1989). The Jackknife and the Bootstrap for General Stationary Observations. The Annals of Statistics 17(3), 1217-1241. https://doi.org/10.1214/aos/1176347265
#'
#' Politis, D. N. & Romano, J. P. (1991). A Circular Block-Resampling Procedure for Stationary Data. In R. LePage & L. Billard, eds, Exploring the Limits of Bootstrap, Wiley, 263-270.
#'
#' @param X A vector representing observed values of the time series.
#' @param maxLag The maximum lag to compute the estimated autocovariance function at.
#' @param x A vector of lag indices. Defaults to the sequence \code{0:length(X)}. Must be at least as large as \code{maxLag + 1}.
#' @param n_bootstrap The number of times to run moving block bootstrap. Defaults to 100.
#' @param l The block length considered for bootstrap. Defaults to \eqn{\lceil N \rceil^{1/3}}, where \eqn{N} is the length of the observation window.
#' @param estimator The function name of the estimator to use. Defaults to \code{standard\_est}.
#' @param type Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.
#' @param alpha The quantile used to compute the \eqn{(1 - \alpha)\%} confidence interval. Defaults to \eqn{0.05.}
#' @param boot_type What type of block bootstrap should be used, either 'moving' for moving block bootstrap or 'circular' for circular block bootstrap.
#' @param parallel Whether or not the bootstrap computations should be done in parallel or not. Defaults to \code{FALSE}.
#' @param ncores The number of cores to be used in the parallel bootstrap computations. Defaults to the number cores - 1 (threads if hyperthreading is available), calculated from \code{parallel::detectCores() - 1}.
#' @param cl_export A vector of any additional functions or variables to export for parallel computations. This may be required if \code{estimator} is not within the package. Defaults to \code{NULL}.
#' @param boot_seed An integer seed for reproducibility. This is used for
#' @param cl An optional cluster object created by \code{parallel::makeCluster}. Defaults to \code{NULL}, which creates a temporary PSOCK cluster.
#' @param ... Optional named arguments to the chosen estimator. See the examples.
#'
#' @return A \code{BootEsts} S3 object (list) with the following values
#' \describe{
#'  \item{\code{acf_avg}}{A numeric vector containing the average autocovariance/autocorrelation bootstrap estimate.}
#'  \item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
#'  \item{\code{acf_orig}}{A numeric vector containing the nonbootstrapped autocovariance/autocorrelation estimate.}
#'  \item{\code{acf_mat}}{A numeric matrix of the a matrix of all of the bootstrap estimates.}
#'  \item{\code{conf_lower}}{A numeric vector containing the lower bounds for the estimated pointwise confidence interval.}
#'  \item{\code{conf_upper}}{A numeric vector containing the upper bounds for the estimated pointwise confidence interval.}
#'  \item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the argument \code{type}.}
#'  \item{\code{est_used}}{The estimator function used, this depends on the argument \code{estimator}.}
#'  \item{\code{boot_type}}{The type of estimate, namely 'moving' or 'circular', this depends on the argument \code{boot_type}.}
#'  \item{\code{alpha}}{A numeric value, which is the \eqn{\alpha} value used to compute the confidence intervals, this depends on the argument \code{alpha}.}
#' }
#' @export
#'
#' @importFrom stats quantile
#' @importFrom parallel makePSOCKcluster
#' @importFrom parallel stopCluster
#' @importFrom parallel clusterExport
#' @importFrom parallel clusterSetRNGStream
#' @importFrom parallel parLapply
#'
#' @examples
#' X <- c(1, 2, 3, 3, 2, 1)
#' block_bootstrap(X, 4, n_bootstrap = 3, l = 2, type = 'autocorrelation')
#' block_bootstrap(X, 4, n_bootstrap = 3, l = 2, type = 'autocovariance')
#' block_bootstrap(X, 4, n_bootstrap = 3, l = 2, estimator=tapered_est,
#'     rho = 0.5, window_name = 'blackman', window_params = c(0.16),
#'     type='autocorrelation')
#
#' my_cov_est <- function(X, maxLag) {
#'   n <- length(X)
#'   covVals <- rep(0, maxLag + 1)
#'   for(h in 0:maxLag) {
#'     covVals_t <- (X[1:(n-h)] - mean(X)) * (X[(1+h):n] - mean(X))
#'     covVals[h] <- sum(covVals_t) / (n - h)
#'   }
#'   return(covVals)
#' }
#' block_bootstrap(X, 4, n_bootstrap = 3, l = 2, estimator=my_cov_est)
#'
#' plot(LakeHuron, main="Lake Huron Levels", ylab="Feet")
#' X <- as.vector(LakeHuron)
#' block_bootstrap(X, 20, n_bootstrap = 100, l = 40, type = 'autocorrelation')
#' block_bootstrap(X, 20, n_bootstrap = 100, l = 40, type = 'autocorrelation')
#' block_bootstrap(X, 20, n_bootstrap = 100, l = 40, estimator=tapered_est,
#'     rho = 0.5, window_name = 'blackman', window_params = c(0.16),
#'     type='autocorrelation')
#'
#' my_cov_est <- function(X, maxLag) {
#'   n <- length(X)
#'   covVals <- rep(0, maxLag + 1)
#'   for(h in 0:maxLag) {
#'     covVals_t <- (X[1:(n-h)] - mean(X)) * (X[(1+h):n] - mean(X))
#'     covVals[h] <- sum(covVals_t) / (n - h)
#'   }
#'   return(covVals)
#' }
#' block_bootstrap(X, 20, n_bootstrap = 100, l = 40, estimator = my_cov_est,
#'     type = 'autocorrelation')
#' \dontrun{
#' library(parallel)
#' X <- c(1, 2, 3, 3, 2, 1)
#' my_cl <- makePSOCKcluster(2)
#' block_bootstrap(X, 4, n_bootstrap = 1000, l = 3, parallel = TRUE, cl = my_cl)
#' stopCluster(my_cl)
#' }
block_bootstrap <- function(X, maxLag, x = 0:length(X), n_bootstrap = 100, l = ceiling(length(X)^(1/3)), estimator = standard_est, type = c("autocovariance", "autocorrelation"), alpha = 0.05, boot_type = c("moving", "circular"), parallel = FALSE, ncores = parallel::detectCores() - 1, cl_export = NULL, boot_seed = NULL, cl = NULL, ...) {
  stopifnot(is.numeric(X), length(X) >= 1, !any(is.na(X)), is.numeric(maxLag), length(maxLag) == 1,
            maxLag > 0, maxLag <= (length(X) - 1), maxLag %% 1 == 0, is.numeric(x), length(x) >= maxLag,
            is.numeric(n_bootstrap), n_bootstrap > 0, n_bootstrap %% 1 ==0, is.numeric(l), length(l) == 1,
            l > 0, l <= length(X), l %% 1 == 0, is.function(estimator), is.numeric(alpha), alpha <= 1, alpha >= 0,
            is.logical(parallel))

  type <- match.arg(type)
  boot_type <- match.arg(boot_type)

  ellipsis <- list(...)

  N <- length(X)
  k <- ceiling(N / l)
  acf_mat <- matrix(NA, ncol = maxLag + 1, nrow = n_bootstrap)

  boot_results <- NULL
  boot_seed <- if(is.null(boot_seed)) sample.int(.Machine$integer.max, 1) else boot_seed

  boot_iter <- function(i) {
    Y <- bootstrap_sample(X, l, k, boot_type)
    attempts <- 1
    max_attempts <- 100

    while(length(unique(Y)) == 1 && attempts < max_attempts) {
      Y <- bootstrap_sample(X, l, k, boot_type)
      attempts <- attempts + 1
    }

    if(attempts == max_attempts) {
      stop("Failed to generate a nonconstant bootstrap sample after 100 attempts. Exiting.")
    }

    est_args <- c(list(Y, maxLag = maxLag), ellipsis)
    return(as.numeric(do.call(estimator, est_args)))
  }

  if(parallel) {
    # See if the user objects actually exist before wasting time creating the PSOCK cluster.
    if(!is.null(cl_export)) {
      missing_exports <- cl_export[!sapply(cl_export, exists, envir = parent.frame())]
      if(length(missing_exports) > 0) {
        stop("The following objects in `cl_export` were not found in your environment: ", paste0(missing_exports, collapse = ", "))
      }
    }

    if(is.null(cl)) {
      cl <- parallel::makePSOCKcluster(ncores)
      on.exit(parallel::stopCluster(cl), add = TRUE)
    }

    varList <- c("X", "l", "k", "boot_type", "maxLag", "estimator", "ellipsis", "boot_iter")
    parallel::clusterExport(cl, varlist = varList, envir = environment())

    package_funcs <- c("bootstrap_sample", "starting_locs")
    parallel::clusterExport(cl, varlist = package_funcs, envir = asNamespace("CovEsts"))

    parallel::clusterSetRNGStream(cl, boot_seed)

    # Export user supplied objects
    parallel::clusterExport(cl, varlist = cl_export, envir = parent.frame())

    boot_results <- parallel::parLapply(cl, 1:n_bootstrap, boot_iter)
  }

  else {
    prev_rng <- RNGkind()
    on.exit(RNGkind(kind = prev_rng[1], normal.kind = prev_rng[2]), add = TRUE)
    set.seed(boot_seed)

    boot_results <- lapply(1:n_bootstrap, boot_iter)
  }

  acf_mat <- do.call(rbind, boot_results)

  original_acf <- as.numeric(estimator(X, ..., maxLag = maxLag))
  if(type == 'autocorrelation') {
    original_acf <- original_acf / original_acf[1]

    acf_mat <- acf_mat / acf_mat[, 1]
  }

  quantiles <- apply(acf_mat, 2, quantile, probs = c(alpha / 2, 1 - (alpha / 2)))
  lower_vals <- quantiles[1, ]
  upper_vals <- quantiles[2, ]
  average_acf <- colMeans(acf_mat)

  res <- list(acf_avg = average_acf, lags=x[1:(maxLag + 1)], acf_orig = original_acf,
              acf_mat = acf_mat, conf_lower = lower_vals, conf_upper = upper_vals,
              est_type = type, est_used = paste0(substitute(estimator)),
              boot_type = boot_type, alpha = alpha)

  return(structure(res, class = "BootEsts"))
}

# starting_locs
test_that("starting_locs fails for nonnumeric N", {
  expect_error(starting_locs('a', 2, 2))
  expect_error(starting_locs(1i, 2, 2))
})

test_that("starting_locs fails for N < 1", {
  expect_error(starting_locs(0, 2, 2))
})

test_that("starting_locs fails for NA N", {
  expect_error(starting_locs(NA, 2, 2))
})

test_that("starting_locs fails for nonnumeric l", {
  expect_error(starting_locs(4, 'a', 2))
  expect_error(starting_locs(4, 1i, 2))
})

test_that("starting_locs fails for l < 1", {
  expect_error(starting_locs(4, 0, 2))
})

test_that("starting_locs fails for NA l", {
  expect_error(starting_locs(4, NA, 2))
})

test_that("starting_locs fails for nonnumeric k", {
  expect_error(starting_locs(4, 2, 'a'))
  expect_error(starting_locs(4, 2, 1i))
})

test_that("starting_locs fails for k < 1", {
  expect_error(starting_locs(4, 2, 0))
})

test_that("starting_locs fails for NA k", {
  expect_error(starting_locs(4, 2, NA))
})

test_that("starting_locs fails for boot_type being neither 'moving' or 'circular'", {
  expect_error(starting_locs(4, 2, 2, 'stationary'))
})

#  bootstrap_samples
test_that("bootstrap_samples fails for nonnumeric X", {
  expect_error(bootstrap_samples(c(1,'a', 3, 4), 2, 2))
  expect_error(bootstrap_samples(c(1, 2i, 3, 4), 2, 2))
})

test_that("bootstrap_samples fails for X less than length 1", {
  expect_error(bootstrap_samples(c(), 2, 2))
})

test_that("bootstrap_samples fails for at least one NA in X", {
  expect_error(bootstrap_samples(c(1, NA, 3, 4), 2, 2))
})

test_that("bootstrap_samples fails for nonnumeric l", {
  expect_error(bootstrap_samples(c(1, 2, 3, 4), 'a', 2))
  expect_error(bootstrap_samples(c(1, 2, 3, 4), 1i, 2))
})

test_that("bootstrap_samples fails for l < 1", {
  expect_error(bootstrap_samples(c(1, 2, 3, 4), 0, 2))
})

test_that("bootstrap_samples fails for NA l", {
  expect_error(bootstrap_samples(c(1, 2, 3, 4), NA, 2))
})

test_that("bootstrap_samples fails for nonnumeric k", {
  expect_error(bootstrap_samples(c(1, 2, 3, 4), 2, 'a'))
  expect_error(bootstrap_samples(c(1, 2, 3, 4), 2, 1i))
})

test_that("bootstrap_samples fails for k < 1", {
  expect_error(bootstrap_samples(c(1, 2, 3, 4), 2, 0))
})

test_that("bootstrap_samples fails for NA k", {
  expect_error(bootstrap_samples(c(1, 2, 3, 4), 2, NA))
})

test_that("bootstrap_samples fails for boot_type being neither 'moving' or 'circular'", {
  expect_error(bootstrap_samples(c(1, 2, 3, 4), 2, 2, 'stationary'))
})

# block_bootstrap
test_that("block_bootstrap fails for nonnumeric X", {
  expect_error(block_bootstrap(c(1,'a', 3, 4), 2))
  expect_error(block_bootstrap(c(1, 2i, 3, 4), 2))
})

test_that("block_bootstrap fails for X less than length 1", {
  expect_error(block_bootstrap(c(), 2))
})

test_that("block_bootstrap fails for at least one NA in X", {
  expect_error(block_bootstrap(c(1, NA, 3, 4), 2))
})

test_that("block_bootstrap fails for nonnumeric maxLag", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 'a'))
  expect_error(block_bootstrap(c(1, 2, 3, 4), 1i))
})

test_that("block_bootstrap fails for maxLag not being length 1", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), c(1, 2)))
})

test_that("block_bootstrap fails for maxLag not being greater than 0", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 0))
})

test_that("block_bootstrap fails for maxLag not being less than length(X)-1", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 4))
})

test_that("block_bootstrap fails for maxLag not being an integer", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2.5))
})

test_that("block_bootstrap fails for nonnumeric x", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, x=c(1, 'a', 3, 4)))
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, x=c(1, 2i, 3, 4)))
})

test_that("block_bootstrap fails for length(x) != length(X)", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, x=1))
})

test_that("block_bootstrap fails for nonnumeric n_bootstrap", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, n_bootstrap='a'))
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, n_bootstrap=1i))
})

test_that("block_bootstrap fails n_bootstrap being less than or equal to zero", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, n_bootstrap=0))
})

test_that("block_bootstrap fails noninteger n_bootstrap", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, n_bootstrap=50.5))
})

test_that("block_bootstrap fails for nonnumeric l", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, l='a'))
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, l=1i))
})

test_that("block_bootstrap fails for lenght of l not being 1", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, l=c(1, 2)))
})

test_that("block_bootstrap fails for l being less than or equal to 0", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, l=0))
})

test_that("block_bootstrap fails for l being greater than length(X)", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, l=5))
})

test_that("block_bootstrap fails noninteger l", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, l=2.5))
})

test_that("block_bootstrap fails for nonnumeric alpha", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, alpha='a'))
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, alpha=1i))
})

test_that("block_bootstrap fails for alpha being greater than 1", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, alpha=1.01))
})

test_that("block_bootstrap fails for alpha being less than 0", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, alpha=-0.01))
})

test_that("block_bootstrap fails for at least type being neither autocovariance or autocorrelation", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, type = "covariance"))
})

test_that("block_bootstrap fails for boot_type being neither 'moving' or 'circular'", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, boot_type = 'stationary'))
})

test_that("block_bootstrap fails for boot_type nonboolean parallel", {
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, parallel = 1))
  expect_error(block_bootstrap(c(1, 2, 3, 4), 2, paralel = 'TRUE'))
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{as.double.CovEsts}
\alias{as.double.CovEsts}
\title{as.double Method for CovEsts Objects}
\usage{
\method{as.double}{CovEsts}(x, ...)
}
\arguments{
\item{x}{A CovEsts S3 object.}

\item{...}{Refer to \link[base:double]{base::double}.}
}
\value{
The acf values from the CovEsts object.
}
\description{
This function provides a method for as.double for CovEsts objects, allowing for as.numeric to be called.
It extracts the acf values from the object.
}
\examples{
as.numeric(standard_est(c(1, 2, 3)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernel_regression_estimator.R
\name{adjusted_est}
\alias{adjusted_est}
\title{Compute the Kernel Regression Estimator.}
\usage{
adjusted_est(
  X,
  x,
  t,
  b,
  kernel_name = c("gaussian", "wave", "rational_quadratic", "bessel_j"),
  kernel_params = c(),
  pd = TRUE,
  type = c("autocovariance", "autocorrelation"),
  meanX = mean(X),
  custom_kernel = FALSE,
  parallel = FALSE,
  ncores = parallel::detectCores() - 1,
  cl_export = NULL,
  cl = NULL
)
}
\arguments{
\item{X}{A vector representing observed values of the time series.}

\item{x}{A vector of lags.}

\item{t}{The arguments at which the autocovariance function is calculated at.}

\item{b}{Bandwidth parameter, greater than 0.}

\item{kernel_name}{The name of the symmetric kernel (see \link{kernel_symm_ec}) function to be used. Possible values are:
gaussian, wave, rational_quadratic, and bessel_j. Alternatively, a custom kernel function can be provided, see the examples.}

\item{kernel_params}{A vector of parameters of the kernel function. See \link{kernel_symm_ec} for parameters.}

\item{pd}{Whether a positive-definite estimate should be used. Defaults to \code{TRUE}.}

\item{type}{Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.}

\item{meanX}{The average value of \code{X}. Defaults to \code{mean(X)}.}

\item{custom_kernel}{If a custom kernel is to be used or not. Defaults to \code{FALSE}.}

\item{parallel}{Whether or not the computations should be done in parallel or not. Defaults to \code{FALSE}.}

\item{ncores}{The number of cores to be used in the parallel computations. Defaults to the number cores - 1 (threads if hyperthreading is available), calculated from \code{parallel::detectCores() - 1}.}

\item{cl_export}{A vector of any additional functions or variables to export for parallel computations. This may be required if \code{estimator} is not within the package. Defaults to \code{NULL}.}

\item{cl}{An optional cluster object created by \code{parallel::makeCluster}. Defaults to \code{NULL}, which creates a temporary PSOCK cluster.}
}
\value{
A \code{CovEsts} S3 object (list) with the following values
\describe{
\item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
\item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
\item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the \code{type} parameter.}
\item{\code{est_used}}{The estimator function used, in this case, 'adjusted_est'.}
}
}
\description{
This function computes the kernel regression estimator of the autocovariance function.
}
\details{
The kernel regression estimator of an autocovariance function is defined as
\deqn{
\hat{\rho}(t) = \left( \sum_{i=1}^{N} \sum_{j=1}^{N} \check{X}_{ij} K((t -  (t_{i} - t_{j})) / b) \right) \left( \sum_{i=1}^{N} \sum_{j=1}^{N} K((t -  (t_{i} - t_{j})) / b) \right)^{-1},
}
where \eqn{\check{X}_{ij} = (X(t_{i}) - \bar{X}) (X(t_{j}) - \bar{X}).}

If \code{pd} is \code{TRUE}, the estimator will be made positive-definite through the following procedure
\enumerate{
\item Take the discrete Fourier cosine transform,
\eqn{\widehat{\mathcal{F}}(\theta)}, of the estimated autocovariance function
\item Compute a modified spectrum \eqn{\widetilde{\mathcal{F}}(\theta) = \max(\widehat{\mathcal{F}}(\theta), 0)} for all sample frequencies.
\item Perform the Fourier inversion to obtain a new estimator.
}
}
\examples{
X <- c(1, 2, 3, 4)
adjusted_est(X, 1:4, 1:3, 0.1, "gaussian")
my_kernel <- function(x, params) {
  stopifnot(params[1] > 0, length(x) >= 1)
  return(exp(-((abs(x) / params[1])^params[2])) * (2 * params[1] * gamma(1 + 1/params[2])))
}
adjusted_est(X, 1:4, 1:3, 0.1, my_kernel, c(0.25), custom_kernel = TRUE)
\dontrun{
library(parallel)
X <- c(1, 2, 3, 4)
my_cl <- makePSOCKcluster(2)
adjusted_est(X, 1:4, 1:3, 0.1, "gaussian", parallel = TRUE, cl = my_cl)
stopCluster(my_cl)
}
}
\references{
Hall, P. & Patil, P. (1994). Properties of nonparametric estimators of autocovariance for stationary random fields. Probability Theory and Related Fields 99(3), 399-424. https://doi.org/10.1007/bf01199899

Hall, P., Fisher, N. I., & Hoffmann, B. (1994). On the nonparametric estimation of covariance functions. The Annals of Statistics 22(4), 2115-2134. https://doi.org/10.1214/aos/1176325774
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{lines.BootEsts}
\alias{lines.BootEsts}
\title{Lines Method for BootEsts Objects}
\usage{
\method{lines}{BootEsts}(x, type = "l", ...)
}
\arguments{
\item{x}{A BootEsts S3 object.}

\item{type}{Defaults to \code{'l'}, see \link[base:plot]{base::plot} for the possible values.}

\item{...}{Additional plotting arguments, refer to \link[graphics:par]{graphics::par}.}
}
\value{
A line of a BootEsts S3 object.
}
\description{
This function plots a BootEsts object as a line onto another plot.
}
\examples{
plot(block_bootstrap(c(1, 2, 3), 2))
lines(block_bootstrap(c(1, 2, 3), 2, pd = FALSE), lwd = 2, lty = 3, col = 'blue')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tapered_estimator.R
\name{tapered_est}
\alias{tapered_est}
\title{Compute the Estimated Tapered Autocovariance Function over a Set of Lags.}
\usage{
tapered_est(
  X,
  rho,
  window_name = c("tukey", "triangular", "sine", "power_sine", "blackman",
    "hann_poisson", "welch"),
  window_params = c(1),
  maxLag = length(X) - 1,
  x = 0:length(X),
  type = c("autocovariance", "autocorrelation"),
  meanX = mean(X),
  custom_window = FALSE
)
}
\arguments{
\item{X}{A vector representing observed values of the time series.}

\item{rho}{A scale parameter in \eqn{(0, 1].}}

\item{window_name}{The name of the \link{window_ec} function to be used. Possible values are:
tukey, triangular, sine, power_sine, blackman_window, hann_poisson, welch. Alternatively, a custom window_ec function can be provided, see the example in \link{taper}.}

\item{window_params}{A vector of parameters of the window function.}

\item{maxLag}{An optional parameter that determines the maximum lag to compute the estimated autocovariance function at. Defaults to \code{length(X) - 1}.}

\item{x}{A vector of lag indices. Defaults to the sequence \code{0:length(X)}. Must be at least as large as \code{maxLag + 1}.}

\item{type}{Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.}

\item{meanX}{The average value of \code{X}. Defaults to \code{mean(X)}.}

\item{custom_window}{If a custom window is to be used or not. Defaults to \code{FALSE}.}
}
\value{
A \code{CovEsts} S3 object (list) with the following values
\describe{
\item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
\item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
\item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the \code{type} parameter.}
\item{\code{est_used}}{The estimator function used, in this case, 'tapered_est'.}
}
}
\description{
This function computes the tapered autocovariance over a set of lags.
}
\details{
This function computes the estimated tapered autocovariance over a set of lags,
\deqn{\widehat{C}_{N}^{a} (h) = (H_{2, n}(0))^{-1} \sum_{j=1}^{N-h}  (X(j) - \bar{X} ) ( X(j + h) - \bar{X} ) a((j - 1/2) / N; \rho) a((j + h - 1/2) / N; \rho) ,}
where \eqn{a(\cdot)} is a window function, \eqn{\rho \in (0, 1]}  is a scale parameter.
This estimator takes into account the edge effect during estimation, assigning a lower weight to values closer to the boundaries and higher weights for observations closer to the middle.
This estimator is positive-definite and asymptotically unbiased.
}
\examples{
X <- c(1, 2, 3)
tapered_est(X, 0.5, "tukey", maxLag = 2)
}
\references{
Dahlhaus R. & Künsch, H. (1987). Edge Effects and Efficient Parameter Estimation for Stationary Random Fields. Biometrika 74(4), 877-882. 10.1093/biomet/74.4.877
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{check_pd}
\alias{check_pd}
\alias{check_pd.CovEsts}
\alias{check_pd.default}
\title{Check if an Autocovariance Function Estimate is Positive-Definite or Not.}
\usage{
check_pd(est)

\method{check_pd}{CovEsts}(est)

\method{check_pd}{default}(est)
}
\arguments{
\item{est}{A numeric vector, corresponding cyclic matrix representing an estimated autocovariance function, or a \code{CovEsts} S3 object.}
}
\value{
A boolean where \code{TRUE} denotes a positive-definite autocovariance function estimate and \code{FALSE} for an estimate that is not positive-definite.
}
\description{
This function checks if an autocovariance function estimate is positive-definite or not by determining if the eigenvalues of the corresponding matrix (see the Details section) are all positive.
}
\details{
For an autocovariance function estimate \eqn{\hat{C}(\cdot)} over a set of lags separated by a constant difference \eqn{\{h_{0}, h_{1} , h_{2} , \dots , h_{n} \},}
construct the symmetric matrix
\deqn{\left[ {\begin{array}{ccccc}
\hat{C}(h_{0})     & \hat{C}(h_{1})     & \cdots & \hat{C}(h_{n - 1}) & \hat{C}(h_{n})     \\
\hat{C}(h_{1})     & \hat{C}(h_{0})     & \cdots & \hat{C}(h_{n - 2}) & \hat{C}(h_{n - 1}) \\
\vdots             & \vdots             & \ddots & \vdots             & \vdots             \\
\hat{C}(h_{n - 1}) & \hat{C}(h_{n - 2}) & \cdots & \hat{C}(h_{0})     & \hat{C}(h_{1})     \\
\hat{C}(h_{n})     & \hat{C}(h_{n - 1}) & \cdots & \hat{C}(h_{1})     & \hat{C}(h_{0})     \\
\end{array}} \right] .
}

The eigendecomposition of this matrix is computed to determine if all eigenvalues are positive. If so, the estimated autocovariance function is assumed to be positive-definite.
}
\section{Methods (by class)}{
\itemize{
\item \code{check_pd(CovEsts)}: Method for \code{CovEsts} objects.

\item \code{check_pd(default)}: Method for numeric vectors.

}}
\examples{
x <- seq(0, 5, by=0.1)
estCov <- exp(-x^2)
check_pd(estCov)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tapered_estimator.R
\name{H2n}
\alias{H2n}
\title{Compute Normalisation Factor}
\usage{
H2n(
  n,
  rho,
  window_name = c("tukey", "triangular", "sine", "power_sine", "blackman",
    "hann_poisson", "welch"),
  window_params = c(1),
  custom_window = FALSE
)
}
\arguments{
\item{n}{The sample size.}

\item{rho}{A scale parameter in \eqn{(0, 1].}}

\item{window_name}{The name of the \link{window_ec} function to be used. Possible values are:
tukey, triangular, power_sine, blackman_window, hann_poisson, welch. Alternatively, a custom window function can be provided, see the example for \link{taper}.}

\item{window_params}{A vector of parameters of the window function.}

\item{custom_window}{If a custom window is to be used or not. Defaults to \code{FALSE}.}
}
\value{
A single value being \eqn{H_{2, n}(0)}.
}
\description{
This helper function is used in the computation of the normalisation factor the function \link{tapered_est},
\deqn{H_{2, n}(0) = \sum_{s=1}^{n} a((s - 1/2) / n; \rho)^{2}, }
where \eqn{a(\cdot; \cdot)} is a window function.
}
\examples{
\dontrun{
H2n(3, 0.6, "tukey")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{print.VarioEsts}
\alias{print.VarioEsts}
\title{Print Method for VarioEsts Objects}
\usage{
\method{print}{VarioEsts}(x, n_head = 5, n_tail = 5, digits = NULL, ...)
}
\arguments{
\item{x}{A VarioEsts S3 object.}

\item{n_head}{The number of 'head' values to print, defaults to 5.}

\item{n_tail}{The number of 'tail' values to print, defaults to 5.}

\item{digits}{The number of digits to print. Defaults to \code{NULL}. Note, this value is passed into \link[base:format]{base::format}.}

\item{...}{Addition printing parameters, see \link[base:print]{base::print}.}
}
\value{
Prints a VarioEsts object.
}
\description{
This function plots a VarioEsts object, printing the lags and values.
}
\examples{
print(to_vario(standard_est(c(1, 2, 3))))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/standard_est.R
\name{to_vario}
\alias{to_vario}
\alias{to_vario.default}
\alias{to_vario.CovEsts}
\title{Autocovariance to Semivariogram}
\usage{
to_vario(estCov)

\method{to_vario}{default}(estCov)

\method{to_vario}{CovEsts}(estCov)
}
\arguments{
\item{estCov}{A vector whose values are an estimate autocovariance function or a \code{CovEsts} S3 object.}
}
\value{
A numeric vector whose values are an estimate of the semivariogram or an \code{VarioEsts} S3 object (list) with the following values
\describe{
\item{\code{vario}}{A numeric vector whose values are an estimate of the semivariogram.}
\item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on, inherited from the argument \code{estCov}.}
\item{\code{est_used}}{The estimator function used, in this case, 'to_vario'.}
}
If a numeric vector is given for the argument \code{estCov}, then a numeric vector output is given, and if a \code{CovEsts} S3 object is given, a \code{VarioEsts} object is given as output.
}
\description{
This function computes an estimated semivariogram using an estimated autocovariance function.
}
\details{
The semivariogram, \eqn{\gamma(h)} and autocovariance function, \eqn{C(h)}, under the assumption of weak stationarity are related as follows:
\deqn{
\gamma(h) = C(0) - C(h) .
}

When an empirical autocovariance function is considered instead, this relation does not necessarily hold, however,
it can be used to obtain a function that is close to a semivariogram, see Bilchouris and Olenko (2025).
}
\section{Methods (by class)}{
\itemize{
\item \code{to_vario(default)}: Method for numeric vectors.

\item \code{to_vario(CovEsts)}: Method for \code{CovEsts} objects.

}}
\examples{
X <- c(1, 2, 3)
estCov <- standard_est(X, meanX=mean(X), maxLag = 2, pd=FALSE)
to_vario(estCov)
}
\references{
Bilchouris, A. & Olenko, A (2025). On Nonparametric Estimation of Covariogram. Austrian Statistical Society 54(1), 112-137. 10.17713/ajs.v54i1.1975
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/corrected_standard_estimator.R
\name{solve_shrinking}
\alias{solve_shrinking}
\title{Solve Linear Shrinking}
\usage{
solve_shrinking(par, corr_mat, target)
}
\arguments{
\item{par}{The initial parameter  used in the maximisation process.}

\item{corr_mat}{The autocorrelation matrix of the considered time series.}

\item{target}{A shrinkage target matrix used in the shrinking process. This should only be used if you wish to use a specific matrix as the target.}
}
\value{
A numeric value that is either equal to \eqn{-}\code{par} or 1.
}
\description{
This is an objective function used to select \eqn{\lambda \in [0, 1]} in linear shrinking, see \link{shrinking}.
}
\examples{
\dontrun{
estCorr <- c(1, 0.5, 0)
corr_mat <- cyclic_matrix(estCorr)
solve_shrinking(0.5, corr_mat, diag(length(estCorr)))
}
}
\references{
Devlin, S. J., Gnanadesikan R. & Kettenring, J. R. (1975). Robust Estimation and Outlier Detection with Correlation Coefficients. Biometrika, 62(3), 531-545. 10.1093/biomet/62.3.531

Rousseeuw, P. J. & Molenberghs, G. (1993). Transformation of Non Positive Semidefinite Correlation Matrices. Communications in Statistics - Theory and Methods, 22(4), 965-984. 10.1080/03610928308831068
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tapered_estimator.R
\name{taper}
\alias{taper}
\title{Compute the Function \eqn{a(x; \rho).}}
\usage{
taper(
  x,
  rho,
  window_name = c("tukey", "triangular", "sine", "power_sine", "blackman",
    "hann_poisson", "welch"),
  window_params = c(1),
  custom_window = FALSE
)
}
\arguments{
\item{x}{A vector of numbers between 0 and 1 (inclusive).}

\item{rho}{A scale parameter in \eqn{(0, 1].}}

\item{window_name}{The name of the \link{window_ec} function to be used. Possible values are:
tukey, triangular, power_sine, blackman_window, hann_poisson, welch. Alternatively, a custom window function can be provided, see the example.}

\item{window_params}{A vector of parameters of the window function.}

\item{custom_window}{If a custom window is to be used or not. Defaults to \code{FALSE}.}
}
\value{
A vector of taper values.
}
\description{
This helper function computes the taper function for a given window function as
\deqn{a(x; \rho) = \left\{
\begin{array}{ll}
   w(2x/\rho)     & 0 \leq x < \frac{1}{2} \rho, \\
   1              & \frac{1}{2}\rho \leq x \leq \frac{1}{2} \\
   a(1 - x; \rho) & \frac{1}{2} < x \leq 1
\end{array} ,
\right. }
where \eqn{w(\cdot)} is a continuous increasing function with \eqn{w(0)=0, w(1)=1,}
\eqn{\rho \in (0, 1],} and \eqn{x \in [0, 1].} The possible window function choices are found in \link{window_ec}.
}
\examples{
X <- c(0.1, 0.2, 0.3)
taper(X, 0.5, "tukey")
curve(taper(x, 1, "tukey"), from = 0, to = 1)
curve(taper(x, 1, "power_sine", c(4)), from = 0, to = 1)
my_taper <- function(x, ...) {
  return(x)
}
taper(X, 0.5, my_taper, custom_window = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/standard_est.R
\name{to_pacf}
\alias{to_pacf}
\alias{to_pacf.CovEsts}
\alias{to_pacf.default}
\title{Computes the Standard Estimator of the Autocovariance Function.}
\usage{
to_pacf(estCov)

\method{to_pacf}{CovEsts}(estCov)

\method{to_pacf}{default}(estCov)
}
\arguments{
\item{estCov}{A numeric vector representing an estimated autocovariance or autocorrelation function.}
}
\value{
A vector whose values are an estimate partial autocorrelation function or a \code{CovEsts} S3 object (list) with the following values
\describe{
\item{\code{acf}}{A numeric vector containing the partial autocorrelation estimates.}
\item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
\item{\code{est_type}}{The type of estimate, namely 'partial'.}
\item{\code{est_used}}{The estimator function used, in this case, 'to_pacf'.}
}
If a numeric vector is given for the argument \code{estCov}, then a numeric vector output is given, and if a \code{CovEsts} S3 object is given, a \code{CovEsts} object is given as output.
}
\description{
This function computes the partial autocorrelation function from an estimated autocovariance or autocorrelation function.
}
\details{
This function is a translation of the 'uni_pacf' function in src/library/stats/src/pacf.c of the R source code which is an implementation of the Durbin-Levinson algorithm.
}
\section{Methods (by class)}{
\itemize{
\item \code{to_pacf(CovEsts)}: Method for \code{CovEsts} objects.

\item \code{to_pacf(default)}: Method for numeric vectors.

}}
\examples{
X <- c(1, 2, 3)
to_pacf(standard_est(X, pd = FALSE, maxLag = 2, meanX = mean(X)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block_bootstrap.R
\name{starting_locs}
\alias{starting_locs}
\title{Random Block Locations}
\usage{
starting_locs(N, l, k, boot_type = c("moving", "circular"))
}
\arguments{
\item{N}{The length of the observation window.}

\item{l}{The block length considered for bootstrap.}

\item{k}{The number of blocks considered for bootstrap.}

\item{boot_type}{What type of block bootstrap should be used, either 'moving' for moving block bootstrap or 'circular' for circular block bootstrap.}
}
\value{
A vector of length \code{k} whose values are random block locations.
}
\description{
This function performs random sampling to obtain random starting locations for block bootstrap.
}
\details{
This function performs random sampling to obtain random starting locations for block bootstrap.
If \code{type = 'moving'}, the set \eqn{\{1, \dots, N - \ell  + 1\}} is randomly sampled, with replacement, \eqn{k} times to obtain random block locations for moving block bootstrap.
If \code{type = 'circular'}, the set \eqn{\{1, \dots, N\}} is randomly sampled, with replacement, \eqn{k} times to obtain random block locations for moving block bootstrap.
}
\examples{
starting_locs(4, 2, 2)
}
\references{
Chapters 2.5 and 2.7 in Lahiri, S. N. (2003). Resampling Methods for Dependent Data. Springer. https://doi.org/10.1007/978-1-4757-3803-2

Künsch, H. R. (1989). The Jackknife and the Bootstrap for General Stationary Observations. The Annals of Statistics 17(3), 1217-1241. https://doi.org/10.1214/aos/1176347265

Politis, D. N. & Romano, J. P. (1991). A Circular Block-Resampling Procedure for Stationary Data. In R. LePage & L. Billard, eds, Exploring the Limits of Bootstrap, Wiley, 263-270.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/corrected_standard_estimator.R
\name{shrinking}
\alias{shrinking}
\alias{shrinking.CovEsts}
\alias{shrinking.default}
\title{Linear Shrinking}
\usage{
shrinking(estCov, return_matrix = FALSE, target = NULL)

\method{shrinking}{CovEsts}(estCov, return_matrix = FALSE, target = NULL)

\method{shrinking}{default}(estCov, return_matrix = FALSE, target = NULL)
}
\arguments{
\item{estCov}{A vector whose values are an estimate autocovariance/autocorrelation function.}

\item{return_matrix}{A boolean determining whether the shrunken matrix or the corresponding vector is returned. If \code{FALSE}, it returns a vector whose values are the shrunken autocorrelation function. Defaults to \code{FALSE}.}

\item{target}{A shrinkage target matrix used in the shrinking process. This should only be used if you wish to use a specific matrix as the target.}
}
\value{
A vector with values of the shrunken autocorrelation function, the corresponding matrix (depending on \code{return_matrix}), or \code{CovEsts} S3 object (list) with the following values
\describe{
\item{\code{acf}}{A numeric vector containing the shrunken autocovariance/autocorrelation estimates.}
\item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on, inherited from the argument \code{estCov}.}
\item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the argument \code{type}.}
\item{\code{est_used}}{The estimator function used, in this case, inherited from the argument \code{estCov}.}
\item{\code{correction_method}}{This value is 'shrinking'.}
\item{\code{lambda}}{The \eqn{\lambda} value obtained during the shrinking process.}
}
If a numeric vector is given for the argument \code{estCov}, then a numeric vector output is given, and if a \code{CovEsts} S3 object is given, a \code{CovEsts} object is given as output.
}
\description{
This function corrects an autocovariance/autocorrelation function estimate via linear shrinking of the autocorrelation matrix.
}
\details{
This function corrects an autocovariance/autocorrelation function estimate via linear shrinking of the autocorrelation matrix.
The shrunken autocorrelation matrix is computed as follows
\deqn{\widetilde{R} = \lambda R + (1 - \lambda)I_{p},}
where \eqn{\widetilde{R}} is the shrunken autocorrelation matrix, \eqn{R} is the original autocorrelation matrix, \eqn{\lambda \in [0, 1],} and \eqn{I_{p}} is the \eqn{p\times p} identity matrix.
\eqn{\lambda} is chosen in such a away that largest value which still results in a positive-definite matrix.
The shrunken matrix will be positive-definite.
}
\section{Methods (by class)}{
\itemize{
\item \code{shrinking(CovEsts)}: Method for \code{CovEsts} objects.

\item \code{shrinking(default)}: Method for numeric vectors

}}
\examples{
estCorr <- c(1, 0.8, 0.5, -1.2)
shrinking(estCorr)
target <- diag(length(estCorr))
shrinking(estCorr, TRUE, target)
}
\references{
Devlin, S. J., Gnanadesikan R. & Kettenring, J. R. (1975). Robust Estimation and Outlier Detection with Correlation Coefficients. Biometrika, 62(3), 531-545. 10.1093/biomet/62.3.531

Rousseeuw, P. J. & Molenberghs, G. (1993). Transformation of Non Positive Semidefinite Correlation Matrices. Communications in Statistics - Theory and Methods, 22(4), 965-984. 10.1080/03610928308831068
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{lines.CovEsts}
\alias{lines.CovEsts}
\title{Lines Method for CovEsts Objects}
\usage{
\method{lines}{CovEsts}(x, type = "l", ...)
}
\arguments{
\item{x}{A CovEst S3 object.}

\item{type}{Defaults to \code{'l'}, see \link[base:plot]{base::plot} for the possible values.}

\item{...}{Additional plotting arguments, refer to \link[graphics:par]{graphics::par}.}
}
\value{
A line of a CovEsts S3 object.
}
\description{
This function plots a CovEsts object as a line onto another plot.
}
\examples{
plot(standard_est(c(1, 2, 3)))
lines(standard_est(c(1, 2, 3), pd = FALSE))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{plot.BootEsts}
\alias{plot.BootEsts}
\title{Plot Method for BootEsts Objects}
\usage{
\method{plot}{BootEsts}(x, type = "l", xlab = "Lag (h)", ylab = "ACF", ...)
}
\arguments{
\item{x}{A BootEsts S3 object.}

\item{type}{Defaults to \code{'l'}, see \link[base:plot]{base::plot} for the possible values.}

\item{xlab}{Defaults to \code{'Lag (h)'}, see \link[graphics:title]{graphics::title} for more information.}

\item{ylab}{Defaults to \code{'ACF'}, see \link[graphics:title]{graphics::title} for more information.}

\item{...}{Additional plotting arguments, refer to \link[graphics:par]{graphics::par}.}
}
\value{
A plot of a BootEsts S3 object.
}
\description{
This function plots a BootEsts object. It plots the confidence region, average bootstrap estimate, and the original nonbootstrapped estimate.
}
\examples{
plot(block_bootstrap(c(1, 2, 3), 2))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/corrected_standard_estimator.R
\name{corrected_est}
\alias{corrected_est}
\title{Kernel Correction of the Standard Estimator.}
\usage{
corrected_est(
  X,
  kernel_name = c("gaussian", "exponential", "wave", "rational_quadratic", "spherical",
    "circular", "bessel_j", "matern", "cauchy"),
  kernel_params = c(),
  N_T = 0.1 * length(X),
  pd = TRUE,
  maxLag = length(X) - 1,
  x = 0:length(X),
  type = c("autocovariance", "autocorrelation"),
  meanX = mean(X),
  custom_kernel = FALSE
)
}
\arguments{
\item{X}{A vector representing observed values of the time series.}

\item{kernel_name}{The name of the \link{kernel_ec} function to be used. Possible values are:
gaussian, exponential, wave, rational_quadratic, spherical, circular, bessel_j, matern, cauchy.}

\item{kernel_params}{A vector of parameters of the kernel function. See \link{kernel_ec} for parameters.
In the case of gaussian, wave, rational_quadratic, spherical and circular, \code{N_T} takes the place of \eqn{\theta}.
For kernels that require parameters other than \eqn{\theta}, such as the Matern kernel, those parameters are passed.}

\item{N_T}{The range at which the kernel function vanishes at. Recommended to be \eqn{0.1 N} when considering all lags. This parameter may be large for a lag small estimation lag.}

\item{pd}{Whether a positive-definite estimate should be used. Defaults to \code{TRUE}.}

\item{maxLag}{An optional parameter that determines the maximum lag to compute the estimated autocovariance function at. Defaults to \code{length(X) - 1}.}

\item{x}{A vector of lag indices. Defaults to the sequence \code{0:length(X)}. Must be at least as large as \code{maxLag + 1}.}

\item{type}{Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.}

\item{meanX}{The average value of \code{X}. Defaults to \code{mean(X)}.}

\item{custom_kernel}{If a custom kernel is to be used or not. Defaults to \code{FALSE}. See examples.}
}
\value{
A \code{CovEsts} S3 object (list) with the following values
\describe{
\item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
\item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
\item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the \code{type} parameter.}
\item{\code{est_used}}{The estimator function used, in this case, 'corrected_est'.}
}
}
\description{
This function computes the standard autocovariance estimator and applies kernel correction to it,
\deqn{\widehat{C}_{T}^{(a)}(h) = \widehat{C}(h) a_{T}(h), } where \eqn{a_{T}(h) := a(h / N_{T}).}
It uses a kernel \eqn{a(\cdot)} which decays or vanishes to zero (depending on the type of kernel) where \eqn{a(0) = 1.}
The rate or value at which the kernel vanishes is \eqn{N_{T}}, which is recommended to be of order \eqn{0.1 N}, where \eqn{N} is the length of the observation window, however, one may need to play with this value.
}
\details{
The aim of this estimator is gradually bring the estimated values to zero through the use of a kernel multiplier. This can be useful when estimating an
autocovariance function that is short-range dependent as estimators can have large fluctuations as the lag increases, or to deal with the wave artefacts for large lags, see Bilchouris and Olenko (2025).
This estimator can be positive-definite depending on whether the choice of \eqn{\widehat{C}(\cdot)} and \eqn{a} are chosen to be positive-definite or not.
}
\examples{
X <- c(1, 2, 3)
corrected_est(X, "gaussian")

X <- rnorm(1000)
Y <- c(X[1], X[2])
for(i in 3:length(X)) { Y[i] <- X[i] - 0.3*X[i - 1] - 0.6*X[i - 2] }
plot(Y)
plot(corrected_est(Y, "bessel_j",
     kernel_params=c(0, 1), N_T=0.2*length(Y)))

# Custom kernel
my_kernel <- function(x, theta, params) {
  stopifnot(theta > 0, length(x) >= 1, all(x >= 0))
  return(sapply(x, function(t) ifelse(t == 0, 1,
         ifelse(t == Inf, 0,
         (sin((t^params[1]) / theta) / ((t^params[1]) / theta)) * cos((t^params[2]) / theta)))))
}
plot(corrected_est(Y,
     my_kernel, kernel_params=c(2, 0.25), custom_kernel = TRUE))
}
\references{
Yaglom, AM (1987). Correlation Theory of Stationary and Related Random Functions. Volume I: Basic Results. Springer New York. https://doi.org/10.1007/978-1-4612-4628-2

Bilchouris, A. & Olenko, A (2025). On Nonparametric Estimation of Covariogram. Austrian Statistical Society (Vol. 54, Issue 1). https://doi.org/10.17713/ajs.v54i1.1975
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/B_splines_est.R
\name{splines_df}
\alias{splines_df}
\title{Construct Data Frame of Basis Functions.}
\usage{
splines_df(x, p, m, taus)
}
\arguments{
\item{x}{A vector of lags.}

\item{p}{The order of the splines.}

\item{m}{The number of nonboundary knots.}

\item{taus}{Vector of \eqn{\tau}s, see \link{get_taus}.}
}
\value{
A data frame of the above structure.
}
\description{
This helper function constructs a data frame with the following structure:
\itemize{
\item One column for the x-values
\item m + p columns values of squared basis functions evaluated at the corresponding x.
}
}
\examples{
\dontrun{
taus <- get_taus(3, 2)
splines_df(seq(0, 2, by=0.25), 3, 2, taus)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernels.R
\name{kernel_ec}
\alias{kernel_ec}
\title{1D Isotropic Kernels.}
\usage{
kernel_ec(
  x,
  name = c("gaussian", "exponential", "wave", "rational_quadratic", "spherical",
    "circular", "bessel_j", "matern", "cauchy"),
  params = c(1)
)
}
\arguments{
\item{x}{A vector or matrix of arguments of at least length 1 for which the kernel is computed at.}

\item{name}{The name of the kernel. Options are: gaussian, exponential, wave, rational_quadratic, spherical, circular, bessel_j, matern, and cauchy.}

\item{params}{A vector of parameters for the kernel. See the documentation below for the position of the parameters. All kernels have a scale parameter as the first value in the vector.}
}
\value{
A vector or matrix of kernel values.
}
\description{
This function computes one of the isotropic kernels listed below. Note that \code{kernel()} is \strong{deprecated}, please use \code{kernel_ec()} instead.
Unlike \link{kernel_symm_ec}, these kernels are only defined when \eqn{x \geq 0.} They are used as
kernel multipliers in estimators \link{corrected_est} and \link{kernel_est}.
}
\details{
\strong{Gaussian Kernel}.
The isotropic Gaussian kernel, which is positive-definite for \eqn{{R}^{d}, d \in N,} is defined as
\deqn{a(x;\theta) = \exp(-x^{2} / \theta), \theta > 0 .}
The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}.

\strong{Exponential Kernel}.
The isotropic exponential kernel, which is positive-definite for \eqn{{R}^{d}, d \in N,} is defined as
\deqn{a(x;\theta) = \exp(-x / \theta) , \theta > 0 .}
The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}.

\strong{Isotropic Wave (Cardinal Sine) Kernel}.
The isotropic wave (cardinal sine) kernel, which is positive-definite for \eqn{{R}^{d}, d \leq 3,} is given by
\deqn{a(x;\theta) = \left\{ \begin{array}{ll}
\frac{\theta}{x} \sin\left( \frac{x}{\theta} \right), & x \neq 0 \\
1, & x = 0
\end{array} \right.
}
where \eqn{\theta > 0 .}
The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}.

\strong{Isotropic Rational Quadratic Kernel}.
The isotropic rational quadratic kernel, which is positive-definite for \eqn{{R}^{d}, d \in N,} is defined as
\deqn{a(x;\theta) = \frac{\theta}{x^{2} + \theta} , \theta >  0 .}
The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}.

\strong{Isotropic Spherical Kernel}.
The isotropic spherical kernel, which is positive-definite for \eqn{{R}^{3}, d \leq 3,} is given by
\deqn{a(x;\theta) = \left\{
\begin{array}{ll}
1 - \frac{3}{2}\frac{x}{\theta} + \frac{1}{2}\left( \frac{x}{\theta} \right)^{3}, & x < \theta \\
0, & \mbox{otherwise}
\end{array} \right.
}
where \eqn{\theta > 0 .}
The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}.

\strong{Isotropic Circular Kernel}.
The isotropic circular kernel, which is positive-definite for \eqn{{R}^{d}, d \leq 2,} is given by
\deqn{a(x;\theta) = \left\{
\begin{array}{ll}
\frac{2}{\pi}\arccos\left( \frac{x}{\theta} \right) - \frac{2}{\pi}\frac{x}{\theta} \sqrt{ 1 - \left( \frac{x}{\theta} \right)^{2} }, & x < \theta \\
0, & \mbox{otherwise}
\end{array} \right.
}
where \eqn{\theta > 0 .}
The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}.

\strong{Isotropic Matérn Kernel}.
The isotropic Matérn kernel, which is positive-definite for \eqn{{R}^{d}, d \in N,} and when \eqn{\nu > 0,} is defined as
\deqn{a(x; \theta, \nu) = \left(\sqrt{2\nu} \frac{x}{\theta} \right)^{\nu} \left(2^{\nu - 1} \Gamma(\nu) \right)^{-1} K_{\nu}\left( \sqrt{2\nu} \frac{x}{\theta} \right) ,}
where \eqn{\theta > 0} and \eqn{K_{\nu}(\cdot)} is the modified Bessel function of the second kind.
The \code{params} argument is of the form \code{c(}\eqn{\theta, \nu}\code{)}.

\strong{Isotropic Bessel Kernel}.
The isotropic Bessel kernel, which is positive-definite for \eqn{{R}^{d}, d \in N,} and when \eqn{\nu \geq \frac{d}{2} - 1,} is given by
\deqn{a(x; \theta, \nu) = 2^{\nu} \Gamma(\nu + 1) J_{\nu}(x / \theta) (x / \theta)^{-\nu} ,}
where \eqn{\theta > 0} and \eqn{J_{\nu}(\cdot)} is the Bessel function of the first kind.
The \code{params} argument is of the form \code{c(}\eqn{\theta, \nu, d} \code{)}.

\strong{Isotropic Cauchy Kernel}.
The isotropic Cauchy kernel, which is positive-definite for \eqn{{R}^{d}, d \in N,} and when \eqn{0 < \alpha \leq 2} and \eqn{\beta \geq 0,} is defined by
\deqn{a(x ; \theta, \alpha, \beta) = (1 + (x / \theta)^{\alpha})^{-(\beta / \alpha)} , \theta > 0 .}
The \code{params} argument is of the form \code{c(}\eqn{\theta, \alpha, \beta} \code{)}.
}
\examples{
x <- c(0.2, 0.4, 0.6)
theta <- 0.9
kernel_ec(x, "gaussian", c(theta))
kernel_ec(x, "exponential", c(theta))
kernel_ec(x, "wave", c(theta))
kernel_ec(x, "rational_quadratic", c(theta))
kernel_ec(x, "spherical", c(theta))
kernel_ec(x, "circular", c(theta))
nu <- 1
kernel_ec(x, "matern", c(theta, nu))
dim <- 1
kernel_ec(x, "bessel_j", c(theta, nu, dim))
alpha <- 1
beta <- 2
kernel_ec(x, "cauchy", c(theta, alpha, beta))
curve(kernel_ec(x, "gaussian", c(theta)), from = 0, to = 5)
curve(kernel_ec(x, "exponential", c(theta)), from = 0, to = 5)
curve(kernel_ec(x, "wave", c(theta)), from = 0, to = 5)
curve(kernel_ec(x, "rational_quadratic", c(theta)), from = 0, to = 5)
curve(kernel_ec(x, "spherical", c(theta)), from = 0, to = 5)
curve(kernel_ec(x, "circular", c(theta)), from = 0, to = 5)
curve(kernel_ec(x, "matern", c(theta, nu)), from = 0, to = 5)
curve(kernel_ec(x, "bessel_j", c(theta, nu, dim)), from = 0, to = 5)
curve(kernel_ec(x, "cauchy", c(theta, alpha, beta)), from = 0, to = 5)
}
\references{
Genton, M. (2001). Classes of Kernels for Machine Learning: A Statistics Perspective. Journal of Machine Learning Research. 2, 299-312. https://doi.org/10.1162/15324430260185646

Table 4.2 in Hristopulos, D. T. (2020). Random Fields for Spatial Data Modeling:  A Primer for Scientists and Engineers. Springer. https://doi.org/10.1007/978-94-024-1918-4
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernel_regression_estimator.R
\name{make_pd}
\alias{make_pd}
\alias{make_pd.CovEsts}
\alias{make_pd.default}
\title{Make a Function Positive-Definite}
\usage{
make_pd(x, method.1 = TRUE)

\method{make_pd}{CovEsts}(x, method.1 = TRUE)

\method{make_pd}{default}(x, method.1 = TRUE)
}
\arguments{
\item{x}{A vector of numeric values of an estimated autocovariance function.}

\item{method.1}{Should method 1 be used (TRUE) or method 2 (FALSE).}
}
\value{
A vector that is the adjusted function or a \code{CovEsts} S3 object (list) with the following values
\describe{
\item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
\item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on, inherited from the argument \code{x}.}
\item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', inherited from the argument \code{x}.}
\item{\code{est_used}}{The estimator function used, inherited from the argument \code{x}.}
\item{\code{correction_method}}{Either 'method.1' or 'method.2' depending on the argument \code{method.1}}
}
If a numeric vector is given for the argument \code{x}, then a numeric vector output is given, and if a \code{CovEsts} S3 object is given, a \code{CovEsts} object is given as output.
}
\description{
This function can make a function positive-definite using the methods proposed by P. Hall and his coauthors.
}
\details{
This function perform positive-definite adjustments proposed by P. Hall and his coauthors.

Method 1 is as follows:
\enumerate{
\item Take the discrete Fourier cosine transform,
\eqn{\widehat{\mathcal{F}}(\theta)}.
\item Compute a modified spectrum \eqn{\widetilde{\mathcal{F}}(\theta) = \max(\widehat{\mathcal{F}}(\theta), 0)} for all sample frequencies.
\item Perform the Fourier inversion to obtain a new estimator.
}

Method 2 is as follows:
\enumerate{
\item Take the discrete Fourier cosine transform \eqn{\widehat{\mathcal{F}}(\theta)}.
\item Find the smallest frequency where its associated value in the spectral domain is negative
\deqn{\hat{\theta} = \inf\{ \theta > 0 :  \widehat{\mathcal{F}}(\theta) < 0\}.}
\item Compute a modified spectrum \eqn{\widetilde{\mathcal{F}} = \widehat{\mathcal{F}}(\theta)\textbf{1}(\theta < \hat{\theta}),}
where \eqn{\textbf{1}(A)} is the indicator function over the set \eqn{A.}
\item Perform the Fourier inversion.
}
}
\section{Methods (by class)}{
\itemize{
\item \code{make_pd(CovEsts)}: Method for \code{CovEsts} objects.

\item \code{make_pd(default)}: Method for numeric vectors.

}}
\examples{
X <- c(1, 2, 3, 4)
make_pd(X)
check_pd(make_pd(X))
check_pd(make_pd(X, method.1 = FALSE))
}
\references{
Hall, P. & Patil, P. (1994). Properties of nonparametric estimators of autocovariance for stationary random fields. Probability Theory and Related Fields 99(3), 399-424. https://doi.org/10.1007/bf01199899

Hall, P., Fisher, N. I., & Hoffmann, B. (1994). On the nonparametric estimation of covariance functions. The Annals of Statistics 22(4), 2115-2134. https://doi.org/10.1214/aos/1176325774

Bilchouris, A. & Olenko, A (2025). On Nonparametric Estimation of Covariogram. Austrian Statistical Society 54(1), 112-137. https://doi.org/10.17713/ajs.v54i1.1975
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernel_regression_estimator.R
\name{rho_T1}
\alias{rho_T1}
\title{Compute \eqn{\rho(T_{1})} used in the Truncated Kernel Regression Estimator.}
\usage{
rho_T1(
  x,
  meanX,
  T1,
  b,
  xij_mat,
  kernel_name = c("gaussian", "wave", "rational_quadratic", "bessel_j"),
  kernel_params = c(),
  custom_kernel = FALSE
)
}
\arguments{
\item{x}{A vector of lags.}

\item{meanX}{The average value of \code{X}.}

\item{T1}{The first trunctation point.}

\item{b}{Bandwidth parameter, greater than 0.}

\item{xij_mat}{The matrix of pairwise covariance values.}

\item{kernel_name}{The name of the symmetric kernel (see \link{kernel_symm_ec}) function to be used. Possible values are:
gaussian, wave, rational_quadratic, and bessel_j. Alternatively, a custom kernel function can be provided, see the examples.}

\item{kernel_params}{A vector of parameters of the kernel function. See \link{kernel_symm_ec} for parameters.}

\item{custom_kernel}{If a custom kernel is to be used or not. Defaults to \code{FALSE}.}
}
\value{
The estimated autocovariance function at \eqn{T_{1}}.
}
\description{
This helper function computes \eqn{\rho(T_{1})} used in the truncated kernel regression estimator, \link{truncated_est}.
}
\details{
This function computes the following value,
\deqn{
\hat{\rho}(T_{1}) = \left( \sum_{i=1}^{N} \sum_{j=1}^{N} \check{X}_{ij} K((T_{1} - (t_{i} - t_{j})) / b) \right) \left( \sum_{i=1}^{N} \sum_{j=1}^{N} K((T_{1} - (t_{i} - t_{j}))) / b) \right)^{-1},
}
where \eqn{\check{X}_{ij} = (X(t_{i}) - \bar{X}) (X(t_{j}) - \bar{X}),}
which is then used in \link{truncated_est},
\deqn{\hat{\rho}_{1}(t) = \left\{ \begin{array}{ll}
\hat{\rho}(t) & 0 \leq t \leq T_{1} \\
\hat{\rho}(T_{1}) (T_{2} - t)(T_{2} - T_{1})^{-1} & T_{1} < t \leq T_{2} \\
0 & t > T_{2}
\end{array} \right. .}
}
\examples{
\dontrun{
X <- c(1, 2, 3, 4)
rho_T1(1:4, mean(X), 1, 0.1, Xij_mat(X, mean(X)), "gaussian", c(), FALSE)
my_kernel <- function(x, params) {
  stopifnot(params[1] > 0, length(x) >= 1)
  return(exp(-((abs(x) / params[1])^params[2])) * (2 * params[1]  * gamma(1 + 1/params[2])))
}
rho_T1(1:4, mean(X), 1, 0.1, Xij_mat(X, mean(X)), my_kernel, c(0.25), TRUE)
}
}
\references{
Hall, P. & Patil, P. (1994). Properties of nonparametric estimators of autocovariance for stationary random fields. Probability Theory and Related Fields 99(3), 399-424. https://doi.org/10.1007/bf01199899

Hall, P., Fisher, N. I., & Hoffmann, B. (1994). On the nonparametric estimation of covariance functions. The Annals of Statistics 22(4), 2115-2134. https://doi.org/10.1214/aos/1176325774
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block_bootstrap.R
\name{block_bootstrap}
\alias{block_bootstrap}
\title{Block Bootstrap}
\usage{
block_bootstrap(
  X,
  maxLag,
  x = 0:length(X),
  n_bootstrap = 100,
  l = ceiling(length(X)^(1/3)),
  estimator = standard_est,
  type = c("autocovariance", "autocorrelation"),
  alpha = 0.05,
  boot_type = c("moving", "circular"),
  parallel = FALSE,
  ncores = parallel::detectCores() - 1,
  cl_export = NULL,
  boot_seed = NULL,
  cl = NULL,
  ...
)
}
\arguments{
\item{X}{A vector representing observed values of the time series.}

\item{maxLag}{The maximum lag to compute the estimated autocovariance function at.}

\item{x}{A vector of lag indices. Defaults to the sequence \code{0:length(X)}. Must be at least as large as \code{maxLag + 1}.}

\item{n_bootstrap}{The number of times to run moving block bootstrap. Defaults to 100.}

\item{l}{The block length considered for bootstrap. Defaults to \eqn{\lceil N \rceil^{1/3}}, where \eqn{N} is the length of the observation window.}

\item{estimator}{The function name of the estimator to use. Defaults to \code{standard\_est}.}

\item{type}{Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.}

\item{alpha}{The quantile used to compute the \eqn{(1 - \alpha)\%} confidence interval. Defaults to \eqn{0.05.}}

\item{boot_type}{What type of block bootstrap should be used, either 'moving' for moving block bootstrap or 'circular' for circular block bootstrap.}

\item{parallel}{Whether or not the bootstrap computations should be done in parallel or not. Defaults to \code{FALSE}.}

\item{ncores}{The number of cores to be used in the parallel bootstrap computations. Defaults to the number cores - 1 (threads if hyperthreading is available), calculated from \code{parallel::detectCores() - 1}.}

\item{cl_export}{A vector of any additional functions or variables to export for parallel computations. This may be required if \code{estimator} is not within the package. Defaults to \code{NULL}.}

\item{boot_seed}{An integer seed for reproducibility. This is used for}

\item{cl}{An optional cluster object created by \code{parallel::makeCluster}. Defaults to \code{NULL}, which creates a temporary PSOCK cluster.}

\item{...}{Optional named arguments to the chosen estimator. See the examples.}
}
\value{
A \code{BootEsts} S3 object (list) with the following values
\describe{
\item{\code{acf_avg}}{A numeric vector containing the average autocovariance/autocorrelation bootstrap estimate.}
\item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
\item{\code{acf_orig}}{A numeric vector containing the nonbootstrapped autocovariance/autocorrelation estimate.}
\item{\code{acf_mat}}{A numeric matrix of the a matrix of all of the bootstrap estimates.}
\item{\code{conf_lower}}{A numeric vector containing the lower bounds for the estimated pointwise confidence interval.}
\item{\code{conf_upper}}{A numeric vector containing the upper bounds for the estimated pointwise confidence interval.}
\item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the argument \code{type}.}
\item{\code{est_used}}{The estimator function used, this depends on the argument \code{estimator}.}
\item{\code{boot_type}}{The type of estimate, namely 'moving' or 'circular', this depends on the argument \code{boot_type}.}
\item{\code{alpha}}{A numeric value, which is the \eqn{\alpha} value used to compute the confidence intervals, this depends on the argument \code{alpha}.}
}
}
\description{
This function performs block bootstrap (moving or circular) to obtain a \eqn{(1-\alpha)\%} confidence-interval for the autocovariance function. It will also compute average autocovariance function across all bootstrapped estimates.
}
\details{
This function performs block bootstrap to obtain a \eqn{(1-\alpha)\%} confidence-interval for the autocovariance function. It will also compute average autocovariance function across all bootstrapped estimates.

Moving block bootstrap can be described as follows.
For some times series \eqn{X(1), X(2), \dots, X(n),} construct \eqn{k \in N} overlapping blocks of the form
\eqn{B_{i} = (X(i), \dots, X(i + \ell - 1)),} where \eqn{\ell \in \{1, \dots , n\}} is the block length.
Randomly sample, with replacement, from the discrete uniform distribution with on \eqn{\{1, \dots, n - \ell + 1\}} to obtain a set of random starting locations \eqn{I_{1}, \dots, I_{k}.}
Construct a bootstrapped time series \eqn{B_{1}^{\ast}, B_{2}^{\ast}, \dots, B_{k}^{\ast},} where \eqn{B_{i}^{\ast} = B_{I_{i}}.}
The bootstrapped time series is truncated to have length \eqn{n,} and will be of the form \eqn{X^{\ast}(1), \dots , X^{\ast}(n).}
The autocovariance function is then computed for the bootstrapped time series.

An alternative to moving block bootstrap is circular block bootstrap.
Circular block bootstrap uses the time series like a circle, that is, the observation at \eqn{n + i} is the same as the observation at location \eqn{i.}
For example, for the block \eqn{B_{n - \ell + 2}}, we obtain \eqn{(X(n - \ell + 2) , \dots , X(n), X(n + 1))} is the same as \eqn{(X(n - \ell + 2) , \dots , X(n), X(1)).}
When performing random sampling to obtain starting locations, the set \eqn{\{1, \dots, n\}} is now considered.
The procedure for constructing the bootstrap time series after constructing the blocks and selecting the starting indices is the same as moving block bootstrap.

This process is repeated \code{n_bootstrap} times to obtain \code{n_boostrap} estimates of the autocovariance function using the bootstrapped time series, for which the average autocovariance function
and the \eqn{(1 - \alpha)\%} confidence intervals are constructed pointwise for each lag.

The choice of the block length, \eqn{\ell ,} depends on the degree of dependence present in the time series. If the time series has a high degree of dependence, a larger block size should be chosen to ensure the dependency structure is maintained within the block.

Any estimator of the autocovariance function can be used in this function, including a custom estimator not in the package, see the examples.

When \code{parallel = TRUE}, L'Ecuyer-CMRG is used to ensure independent random numbers across workers. The serial version uses the Mersenne-Twister algorithm.
}
\examples{
X <- c(1, 2, 3, 3, 2, 1)
block_bootstrap(X, 4, n_bootstrap = 3, l = 2, type = 'autocorrelation')
block_bootstrap(X, 4, n_bootstrap = 3, l = 2, type = 'autocovariance')
block_bootstrap(X, 4, n_bootstrap = 3, l = 2, estimator=tapered_est,
    rho = 0.5, window_name = 'blackman', window_params = c(0.16),
    type='autocorrelation')
my_cov_est <- function(X, maxLag) {
  n <- length(X)
  covVals <- rep(0, maxLag + 1)
  for(h in 0:maxLag) {
    covVals_t <- (X[1:(n-h)] - mean(X)) * (X[(1+h):n] - mean(X))
    covVals[h] <- sum(covVals_t) / (n - h)
  }
  return(covVals)
}
block_bootstrap(X, 4, n_bootstrap = 3, l = 2, estimator=my_cov_est)

plot(LakeHuron, main="Lake Huron Levels", ylab="Feet")
X <- as.vector(LakeHuron)
block_bootstrap(X, 20, n_bootstrap = 100, l = 40, type = 'autocorrelation')
block_bootstrap(X, 20, n_bootstrap = 100, l = 40, type = 'autocorrelation')
block_bootstrap(X, 20, n_bootstrap = 100, l = 40, estimator=tapered_est,
    rho = 0.5, window_name = 'blackman', window_params = c(0.16),
    type='autocorrelation')

my_cov_est <- function(X, maxLag) {
  n <- length(X)
  covVals <- rep(0, maxLag + 1)
  for(h in 0:maxLag) {
    covVals_t <- (X[1:(n-h)] - mean(X)) * (X[(1+h):n] - mean(X))
    covVals[h] <- sum(covVals_t) / (n - h)
  }
  return(covVals)
}
block_bootstrap(X, 20, n_bootstrap = 100, l = 40, estimator = my_cov_est,
    type = 'autocorrelation')
\dontrun{
library(parallel)
X <- c(1, 2, 3, 3, 2, 1)
my_cl <- makePSOCKcluster(2)
block_bootstrap(X, 4, n_bootstrap = 1000, l = 3, parallel = TRUE, cl = my_cl)
stopCluster(my_cl)
}
}
\references{
Chapters 2.5 and 2.7 in Lahiri, S. N. (2003). Resampling Methods for Dependent Data. Springer. https://doi.org/10.1007/978-1-4757-3803-2

Künsch, H. R. (1989). The Jackknife and the Bootstrap for General Stationary Observations. The Annals of Statistics 17(3), 1217-1241. https://doi.org/10.1214/aos/1176347265

Politis, D. N. & Romano, J. P. (1991). A Circular Block-Resampling Procedure for Stationary Data. In R. LePage & L. Billard, eds, Exploring the Limits of Bootstrap, Wiley, 263-270.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/standard_est.R
\name{standard_est}
\alias{standard_est}
\title{Computes the Standard Estimator of the Autocovariance Function.}
\usage{
standard_est(
  X,
  pd = TRUE,
  maxLag = length(X) - 1,
  x = 0:length(X),
  type = c("autocovariance", "autocorrelation"),
  meanX = mean(X)
)
}
\arguments{
\item{X}{A vector representing observed values of the time series.}

\item{pd}{Whether a positive-definite estimate should be used. Defaults to \code{TRUE}.}

\item{maxLag}{An optional parameter that determines the maximum lag to compute the estimated autocovariance function at. Defaults to \code{length(X) - 1}.}

\item{x}{A vector of lag indices. Defaults to the sequence \code{0:length(X)}. Must be at least as large as \code{maxLag + 1}.}

\item{type}{Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.}

\item{meanX}{The average value of \code{X}. Defaults to \code{mean(X)}.}
}
\value{
A \code{CovEsts} S3 object (list) with the following values
\describe{
\item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
\item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
\item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the \code{type} parameter.}
\item{\code{est_used}}{The estimator function used, in this case, 'standard_est'.}
}
}
\description{
This function computes the following two estimates of the autocovariance function depending on
the parameter \code{pd}, see the Details section.
}
\details{
For \code{pd = TRUE}:
\deqn{
\widehat{C}(h) = \frac{1}{N} \sum_{j=1}^{N-h} ( X(j) - \bar{X} ) ( X(j + h) - \bar{X} ) .
}

For \code{pd = FALSE}:
\deqn{
\widehat{C}(h) = \frac{1}{N - h} \sum_{j=1}^{N-h} ( X(j) - \bar{X} ) ( X(j + h) - \bar{X} ) .
}

This function will generate autocovariance values for lags \eqn{h} from the set \eqn{\{0, \dots, \mbox{maxLag}\}.}

The positive-definite estimator must be used cautiously when estimating over all lags as the sum of all values of the autocorrelation function equals to \eqn{-1/2}.
For the nonpositive-definite estimator a similar constant summation property holds.
}
\examples{
X <- c(1, 2, 3)
standard_est(X, pd = FALSE, maxLag = 2, meanX = mean(X))
}
\references{
Bilchouris, A. & Olenko, A (2025). On Nonparametric Estimation of Covariogram. Austrian Statistical Society 54(1), 112-137. https://doi.org/10.17713/ajs.v54i1.1975
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/windows.R
\name{window_symm_ec}
\alias{window_symm_ec}
\title{1D Symmetric Window Functions.}
\usage{
window_symm_ec(
  x,
  name = c("tukey", "triangular", "sine", "power_sine", "blackman", "hann_poisson",
    "welch"),
  params = c(1)
)
}
\arguments{
\item{x}{A vector or matrix of arguments of at least length 1. Each value must be between 0 and 1, inclusive.}

\item{name}{The name of the window. Options are: tukey, triangular, sine, power_sine, blackman, hann_poisson, welch.}

\item{params}{A vector of parameters for the windows. See the documentation below for the position of the parameters.}
}
\value{
A vector or matrix of values.
}
\description{
A symmetric window function in this context are traditional window functions, unlike the \link{window} functions.
Note that \code{window_symm()} is \strong{deprecated}, please use \code{window_symm_ec()} instead.
This computes one of the symmetric window functions listed below, all of which are defined for \eqn{x \in [-1, 1],} and are 0 otherwise.
}
\details{
\strong{Tukey Window}.
The Tukey window is defined as
\deqn{w(x) = \frac{1}{2} + \frac{1}{2} \cos(\pi |x|) , x \in [-1, 1].}
The \code{params} argument is empty, see the example.

\strong{Triangular Window}.
The triangular window is given by
\deqn{w(x) = 1 - |x|, x \in [-1, 1].}
The \code{params} argument is empty, see the example.

\strong{Sine Window}.
The sine window is given by
\deqn{w(x) = 1 - \sin\left(\pi |x| / 2 \right), x \in [-1, 1].}
The \code{params} argument is empty, see the example.

\strong{Power Sine Window}.
The power sine window is given by
\deqn{w(x; a) = 1 - \sin^{a}(\pi |x| / 2), x \in [-1, 1], a > 0.}
The \code{params} argument is of the form \code{c(}\eqn{a}\code{)}

\strong{Blackman Window}.
The Blackman window is defined as
\deqn{w(x; a) = 1 +  ( (a - 1) / 2) + \frac{1}{2} \cos(\pi |x|) - \frac{a}{2} \cos(2 \pi |x|), x \in [-1, 1], a \in [-0.25, 0.25] .}
The \code{params} argument is of the form \code{c(}\eqn{a}\code{)}.
The standard value of \eqn{a} for this window is \eqn{0.16.}

\strong{Hann-Poisson Window}.
The Hann-Poisson window is defined as
\deqn{w(x; a) = 1 - \frac{1}{2} (1 - \cos(\pi |x|)) \exp( - (a \left|1 - |x| \right|) ) , x \in [-1, 1], a > 0 .}
The \code{params} argument is of the form \code{c(}\eqn{a}\code{)}

\strong{Welch Window}.
The Welch window is given by
\deqn{w(x) = (|x| - 1)^2 , x \in [-1, 1] . }
The \code{params} argument is empty, see the example.
}
\examples{
x <- c(-0.6, -0.4, -0.2, 0, 0.2, 0.4, 0.6)
window_symm_ec(x, "tukey")
window_symm_ec(x, "triangular")
window_symm_ec(x, "sine")
window_symm_ec(x, "power_sine", c(0.7))
window_symm_ec(x, "blackman", c(0.16))
window_symm_ec(x, "hann_poisson", c(0.7))
window_symm_ec(x, "welch")
curve(window_symm_ec(x, "tukey"), from = -1, to = 1)
curve(window_symm_ec(x, "triangular"), from = -1, to = 1)
curve(window_symm_ec(x, "sine"), from = -1, to = 1)
curve(window_symm_ec(x, "power_sine", c(0.7)), from = -1, to = 1)
curve(window_symm_ec(x, "blackman", c(0.16)), from = -1, to = 1)
curve(window_symm_ec(x, "hann_poisson", c(0.7)), from = -1, to = 1)
curve(window_symm_ec(x, "welch"), from = -1, to = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{nearest_pd}
\alias{nearest_pd}
\alias{nearest_pd.CovEsts}
\alias{nearest_pd.default}
\title{Compute the Nearest Positive-Definite Matrix.}
\usage{
nearest_pd(X)

\method{nearest_pd}{CovEsts}(X)

\method{nearest_pd}{default}(X)
}
\arguments{
\item{X}{Either a numeric square matrix. If a vector is provided, a matrix will be created of the form found in \link{cyclic_matrix}.}
}
\value{
The closest positive-definite autocorrelation matrix.
}
\description{
This function computes the nearest positive-definite matrix to some matrix \eqn{A}.
}
\details{
This function computes the nearest positive-definite matrix to some matrix \eqn{A}.

The procedure to do so is as follows
For a matrix \eqn{X}, compute the symmetric matrix \eqn{B = (A + A^{T}) / 2.}
Let \eqn{B = UH} be the polar decomposition of B.
The nearest positive-definite matrix to \eqn{X} is \eqn{X_{F} = (B + H) / 2.}

Unlike \link{shrinking}, only an autocorrelation matrix can be returned, not an autocovariance function.

The implementation is a translation of https://au.mathworks.com/matlabcentral/fileexchange/42885-nearestspd#functions_tab .
}
\section{Methods (by class)}{
\itemize{
\item \code{nearest_pd(CovEsts)}: Method for \code{CovEsts} objects.

\item \code{nearest_pd(default)}: Method for numeric vectors.

}}
\examples{
X <- c(1, 0, -1.1)
nearest_pd(X)
check_pd(nearest_pd(X))
}
\references{
Higham, N. J. (1988). Computing a nearest symmetric positive semidefinite matrix. Linear Algebra and its Applications, 103, 103-118. https://doi.org/10.1016/0024-3795(88)90223-6

D'Errico, J. (2025). nearestSPD (https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd), MATLAB Central File Exchange. Retrieved August 2, 2025.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernel_regression_estimator.R
\name{truncated_est}
\alias{truncated_est}
\title{Compute the Truncated Kernel Regression Estimator.}
\usage{
truncated_est(
  X,
  x,
  t,
  T1,
  T2,
  b,
  kernel_name = c("gaussian", "wave", "rational_quadratic", "bessel_j"),
  kernel_params = c(),
  pd = TRUE,
  type = c("autocovariance", "autocorrelation"),
  meanX = mean(X),
  custom_kernel = FALSE,
  parallel = FALSE,
  ncores = parallel::detectCores() - 1,
  cl_export = NULL,
  cl = NULL
)
}
\arguments{
\item{X}{A vector representing observed values of the time series.}

\item{x}{A vector of lags.}

\item{t}{The arguments at which the autocovariance function is calculated at.}

\item{T1}{The first truncation point, \eqn{T_{1} > 0.}}

\item{T2}{The second truncation point, \eqn{T_{2} > T_{1} > 0.}}

\item{b}{Bandwidth parameter, greater than 0.}

\item{kernel_name}{The name of the symmetric kernel (see \link{kernel_symm_ec}) function to be used. Possible values are:
gaussian, wave, rational_quadratic, and bessel_j. Alternatively, a custom kernel function can be provided, see the examples.}

\item{kernel_params}{A vector of parameters of the kernel function. See \link{kernel_symm_ec} for parameters.}

\item{pd}{Whether a positive-definite estimate should be used. Defaults to \code{TRUE}.}

\item{type}{Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.}

\item{meanX}{The average value of \code{X}. Defaults to \code{mean(X)}.}

\item{custom_kernel}{If a custom kernel is to be used or not. Defaults to \code{FALSE}.}

\item{parallel}{Whether or not the computations should be done in parallel or not. Defaults to \code{FALSE}.}

\item{ncores}{The number of cores to be used in the parallel computations. Defaults to the number cores - 1 (threads if hyperthreading is available), calculated from \code{parallel::detectCores() - 1}.}

\item{cl_export}{A vector of any additional functions or variables to export for parallel computations. This may be required if \code{estimator} is not within the package. Defaults to \code{NULL}.}

\item{cl}{An optional cluster object created by \code{parallel::makeCluster}. Defaults to \code{NULL}, which creates a temporary PSOCK cluster.}
}
\value{
A \code{CovEsts} S3 object (list) with the following values
\describe{
\item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
\item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
\item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the \code{type} parameter.}
\item{\code{est_used}}{The estimator function used, in this case, 'truncated_est'.}
}
}
\description{
This function computes the truncated kernel regression estimator, based on the kernel regression estimator \eqn{\hat{\rho}(\cdot)}, see \link{adjusted_est}.
}
\details{
This function computes the truncated kernel regression estimator,
\deqn{\hat{\rho}_{1}(t) = \left\{ \begin{array}{ll}
\hat{\rho}(t) & 0 \leq t \leq T_{1} \\
\hat{\rho}(T_{1}) (T_{2} - t)(T_{2} - T_{1})^{-1} & T_{1} < t \leq T_{2} \\
0 & t > T_{2}
\end{array} \right. }
where \eqn{\hat{\rho}(\cdot)} is the kernel regression estimator, see \link{adjusted_est}.

Compared to \link{adjusted_est}, this function brings down the estimate to zero linearly between \eqn{T_{1}} and \eqn{T_{2}}.
In the case of short-range dependence, this may be beneficial as it can remove estimation artefacts at large lags.

To make this estimator positive-definite, the following procedure is used:
\enumerate{
\item Take the discrete Fourier cosine transform
\eqn{\widehat{\mathcal{F}}(\theta)}.
\item Find the smallest frequency where its associated value in the spectral domain is negative
\deqn{\hat{\theta} = \inf\{ \theta > 0 :  \widehat{\mathcal{F}}(\theta)) < 0\}.}
\item Set all values starting at the frequency to zero.
\item Perform the Fourier inversion.
}

If \eqn{\hat{\theta}} is a small frequency, most of the spectrum equals zero, resulting in an inaccurate estimate of the autocovariance function, see Bilchouris and Olenko (2025).
}
\examples{
X <- c(1, 2, 3, 4)
truncated_est(X, 1:4, 1:3, 1, 2, 0.1,
                  "gaussian")

my_kernel <- function(x, params) {
  stopifnot(params[1] > 0, length(x) >= 1)
  return(exp(-((abs(x) / params[1])^params[2])) * (2 * params[1] * gamma(1 + 1/params[2])))
}

truncated_est(X, 1:4, 1:3, 1, 2, 0.1, my_kernel, c(0.25), custom_kernel = TRUE)
\dontrun{
library(parallel)
X <- c(1, 2, 3, 4)
my_cl <- makePSOCKcluster(2)
truncated_est(X, 1:4, 1:3, 1, 2, 0.1, "gaussian", parallel = TRUE, cl = my_cl)
stopCluster(my_cl)
}
}
\references{
Hall, P. & Patil, P. (1994). Properties of nonparametric estimators of autocovariance for stationary random fields. Probability Theory and Related Fields 99(3), 399-424. https://doi.org/10.1007/bf01199899

Hall, P., Fisher, N. I., & Hoffmann, B. (1994). On the nonparametric estimation of covariance functions. The Annals of Statistics 22(4), 2115-2134. https://doi.org/10.1214/aos/1176325774

Bilchouris, A. & Olenko, A (2025). On Nonparametric Estimation of Covariogram. Austrian Statistical Society 54(1), 112-137. https://doi.org/10.17713/ajs.v54i1.1975
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{spectral_norm}
\alias{spectral_norm}
\alias{spectral_norm.CovEsts}
\alias{spectral_norm.default}
\title{Compute the Spectral Norm Between Estimated Functions.}
\usage{
spectral_norm(est1, est2)

\method{spectral_norm}{CovEsts}(est1, est2)

\method{spectral_norm}{default}(est1, est2)
}
\arguments{
\item{est1}{A numeric vector representing the first estimated autocovariance function.}

\item{est2}{A numeric vector of the same length as \code{est1} representing the second estimated autocovariance function}
}
\value{
The spectral norm of the differences between the two functions.
}
\description{
This function computes the spectral norm of the difference of two estimated autocovariance functions.
This function is intended for estimates over lags with a constant difference.
}
\details{
This function computes the spectral norm of the difference of two estimated autocovariance functions.
Let \eqn{D(h) = \hat{C}_{1}(h) - \hat{C}_{2}(h),}
where \eqn{\hat{C}_{1}(\cdot)} and \eqn{\hat{C}_{2}(\cdot)} are estimated autocovariance functions.

A matrix \eqn{D} is created from \eqn{D(\cdot)},
\deqn{\left[ {\begin{array}{ccccc}
D(h_{0})     & D(h_{1})     & \cdots & D(h_{n - 1}) & D(h_{n})     \\
D(h_{1})     & D(h_{0})     & \cdots & D(h_{n - 2}) & D(h_{n - 1}) \\
\vdots       & \vdots       & \ddots & \vdots       & \vdots       \\
D(h_{n - 1}) & D(h_{n - 2}) & \cdots & D(h_{0})     & D(h_{1})     \\
D(h_{n})     & D(h_{n - 1}) & \cdots & D(h_{1})     & D(h_{0})     \\
\end{array}} \right] ,
}
over a set of lags \eqn{\{h_{0}, h_{1}, \dots , h_{N} \}.}
This matrix is created by \link{cyclic_matrix}.

The spectral norm is defined as the largest eigenvalue of \eqn{D.}
}
\section{Methods (by class)}{
\itemize{
\item \code{spectral_norm(CovEsts)}: Method for \code{CovEsts} objects.

\item \code{spectral_norm(default)}: Method for numeric vectors.

}}
\examples{
x <- seq(0, 5, by=0.1)
estCov1 <- exp(-x^2)
estCov2 <- exp(-x^2.1)
spectral_norm(estCov1, estCov2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/windows.R
\name{window_ec}
\alias{window_ec}
\title{1D Window Functions.}
\usage{
window_ec(
  x,
  name = c("tukey", "triangular", "sine", "power_sine", "blackman", "hann_poisson",
    "welch"),
  params = c(1)
)
}
\arguments{
\item{x}{A vector or matrix of arguments of at least length 1. Each value must be between 0 and 1, inclusive.}

\item{name}{The name of the window. Options are: tukey, triangular, sine, power_sine, blackman, hann_poisson, welch.}

\item{params}{A vector of parameters for the windows. See the documentation below for the position of the parameters.}
}
\value{
A vector or matrix of values.
}
\description{
A window function in this context is a continuous nondecreasing function such that at 0 it is 0, and at 1, it is 1.
Note that \code{window()} is \strong{deprecated}, please use \code{window_ec()} instead.
This computes one of the window functions listed below.
}
\details{
\strong{Tukey Window}.
The Tukey window is defined as
\deqn{w(x) = \frac{1}{2} - \frac{1}{2} \cos(\pi x) , x \in [0, 1].}
The \code{params} argument is empty.

\strong{Triangular Window}.
The triangular window is given by
\deqn{w(x) = x, x \in [0, 1].}
The \code{params} argument is empty.

\strong{Sine Window}.
The sine window is given by
\deqn{w(x) = \sin\left(\pi x / 2 \right), x \in [0, 1].}
The \code{params} argument is empty.

\strong{Power Sine Window}.
The power sine window is given by
\deqn{w(x; a) = \sin^{a}(\pi x / 2), x \in [0, 1], a > 0.}
The \code{params} argument is of the form \code{c(}\eqn{a}\code{)}.

\strong{Blackman Window}.
The Blackman window is defined as
\deqn{w(x; a) = ( (1 - a) / 2) - \frac{1}{2} \cos(\pi x) + \frac{a}{2} \cos(2 \pi x), x \in [0, 1], a \in [-0.25, 0.25] .}
The \code{params} argument is of the form \code{c(}\eqn{a}\code{)}.
It is recommended that \eqn{a \in [-0.25, 0.25]} to ensure that the window is nondecreasi#'ng on \eqn{[0, 1].}

\strong{Hann-Poisson Window}.
The Hann-Poisson window is defined as
\deqn{w(x; a) = \frac{1}{2} (1 - \cos(\pi x)) \exp( - (a \left|1 - x \right|) ) , x \in [0, 1], a > 0.}
The \code{params} argument is of the form \code{c(}\eqn{a}\code{)}.

\strong{Welch Window}.
The Welch window is given by
\deqn{w(x) = 1 - (x - 1)^2 , x \in [0, 1] . }
The \code{params} argument is empty.

See the function call examples below.
}
\examples{
x <- c(0.2, 0.4, 0.6)
window_ec(x, "tukey")
window_ec(x, "triangular")
window_ec(x, "sine")
window_ec(x, "power_sine", c(0.7))
window_ec(x, "blackman", c(0.16))
window_ec(x, "hann_poisson", c(0.7))
window_ec(x, "welch")
curve(window_ec(x, "tukey"), from = 0, to = 1)
curve(window_ec(x, "triangular"), from = 0, to = 1)
curve(window_ec(x, "sine"), from = 0, to = 1)
curve(window_ec(x, "power_sine", c(0.7)), from = 0, to = 1)
curve(window_ec(x, "blackman", c(0.16)), from = 0, to = 1)
curve(window_ec(x, "hann_poisson", c(0.7)), from = 0, to = 1)
curve(window_ec(x, "welch"), from = 0, to = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/B_splines_est.R
\name{splines_est}
\alias{splines_est}
\alias{splines_est.CovEsts}
\alias{splines_est.default}
\title{Compute the Splines Estimator.}
\usage{
splines_est(
  X,
  x,
  estCov,
  p,
  m,
  maxLag = length(X) - 1,
  type = c("autocovariance", "autocorrelation"),
  initial_pars = c(),
  control = list(maxit = 1000)
)

\method{splines_est}{CovEsts}(
  X,
  x,
  estCov,
  p,
  m,
  maxLag = length(X) - 1,
  type = c("autocovariance", "autocorrelation"),
  initial_pars = c(),
  control = list(maxit = 1000)
)

\method{splines_est}{default}(
  X,
  x,
  estCov,
  p,
  m,
  maxLag = length(X) - 1,
  type = c("autocovariance", "autocorrelation"),
  initial_pars = c(),
  control = list(maxit = 1000)
)
}
\arguments{
\item{X}{A vector representing observed values of the time series.}

\item{x}{A vector of lag indices.}

\item{estCov}{An estimated autocovariance function to fit to (a vector).}

\item{p}{The order of the splines.}

\item{m}{The number of nonboundary knots.}

\item{maxLag}{An optional parameter that determines the maximum lag to compute the estimated autocovariance function at. Defaults to \code{length(X) - 1}.}

\item{type}{Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.}

\item{initial_pars}{An optional vector of parameters - can be used to fine tune the fit. By default, it is a vector of 0.5 whose length is \eqn{m+p.}}

\item{control}{An optional list of optimisation parameters used in the optimisation process, see \code{control} in \link[stats:optim]{stats::optim}.}
}
\value{
A vector whose values are the spline autocovariance estimates or a \code{CovEsts} S3 object (list) with the following values
\describe{
\item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
\item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on.}
\item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the \code{type} parameter.}
\item{\code{est_used}}{The estimator function used, in this case, 'splines_est'.}
}
If a numeric vector is given for the argument \code{estCov}, then a numeric vector output is given, and if a \code{CovEsts} S3 object is given, a \code{CovEsts} object is given as output.
}
\description{
Compute the estimated covariance function by using the method from Choi, Li & Wang (2013, pp. 614-617).
\deqn{C(\tau) = \sum_{j = 1}^{m + p} \beta_{j} f_{j}^{(p-1)}(\tau^{2}),}
where \eqn{m} is the number of nonboundary knots, \eqn{p} is the order of the splines, \eqn{\tau} is the isotropic distance, \eqn{\beta_{j}} are nonnegative weights and \eqn{f_{j}^{(p)}} are basis functions of order \eqn{p.}
For optimisation, the Nelder-Mead and L-BFGS-B methods are used, the one which selects parameters which minimises the objective function is chosen.
}
\details{
Due to the weighting scheme, the autocovariance at lag zero cannot be 1,
\deqn{
w_{i} = \frac{N - i}{1 - C(i)}
}
}
\section{Methods (by class)}{
\itemize{
\item \code{splines_est(CovEsts)}: Method for 'CovEsts' objects.

\item \code{splines_est(default)}: Method for numeric vectors.

}}
\examples{
X <- rnorm(100)
x <- seq(0, 5, by = 0.25)
maxLag <- 5
estCov <- standard_est(X, maxLag = maxLag)
estimated <- splines_est(X, x, estCov, 3, 2, maxLag = maxLag)
estimated
}
\references{
Choi, I., Li, B. & Wang, X. (2013). Nonparametric Estimation of Spatial and Space-Time Covariance Function. JABES 18, 611-630. https://doi.org/10.1007/s13253-013-0152-z
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{max_distance}
\alias{max_distance}
\alias{max_distance.CovEsts}
\alias{max_distance.default}
\title{Maximum Vertical Distance Between Estimated Functions.}
\usage{
max_distance(est1, est2, lags = c(), plot = FALSE)

\method{max_distance}{CovEsts}(est1, est2, lags = c(), plot = FALSE)

\method{max_distance}{default}(est1, est2, lags = c(), plot = FALSE)
}
\arguments{
\item{est1}{A numeric vector representing the first estimated autocovariance function.}

\item{est2}{A numeric vector of the same length as \code{est1} representing the second estimated autocovariance function}

\item{lags}{An optional vector of lags starting from 0 up until some other lag. If empty, a vector of lags is created starting from 0 until \code{len(est1) - 1}, by 1.}

\item{plot}{A boolean as to whether a plot should be created. By default, no plot is created.}
}
\value{
A numeric value representing the maximum vertical distance between the two estimated functions.
}
\description{
This function computes the maximum vertical distance between functions.
}
\details{
This function computes the maximum vertical distance between functions:
\deqn{D(\hat{C}_{1}(h), \hat{C}_{2}(h)) = \displaystyle \max_{h} \left| \hat{C}_{1}(h) - \hat{C}_{2}(h) \right| ,
}
where \eqn{\hat{C}_{1}(\cdot)} and \eqn{\hat{C}_{2}(\cdot)} are estimated autocovariance functions.
It assumes that the estimated values are given for the same set of lags.
The vectors of the function values must be of the same length.
}
\section{Methods (by class)}{
\itemize{
\item \code{max_distance(CovEsts)}: Method for \code{CovEsts} objects.

\item \code{max_distance(default)}: Method for numeric vectors.

}}
\examples{
x <- seq(0, 5, by=0.1)
estCov1 <- exp(-x^2)
estCov2 <- exp(-x^2.1)
max_distance(estCov1, estCov2, lags=x)
max_distance(estCov1, estCov2, lags=x, plot = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/B_splines_est.R
\name{generate_knots}
\alias{generate_knots}
\title{Generate Spline Knots.}
\usage{
generate_knots(m)
}
\arguments{
\item{m}{The number of nonboundary knots.}
}
\value{
A numeric vector representing the knots, including the boundary knots.
}
\description{
A helper function that generates \eqn{m + 2} spline knots of the form:
\deqn{
\kappa_{0} = 0 , \kappa_{1} = 1 / (m + 1) , \dots , \kappa_{m} = m / (m + 1) , \kappa_{m + 1} = 1 .
}
The knots are equally spaced with boundary knots \eqn{\kappa_{0} = 0} and \eqn{\kappa_{m + 1} = 1 .}
}
\examples{
\dontrun{
generate_knots(3)
}
}
\references{
Choi, I., Li, B. & Wang, X. (2013). Nonparametric Estimation of Spatial and Space-Time Covariance Function. JABES 18, 611-630. https://doi.org/10.1007/s13253-013-0152-z
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernel_regression_estimator.R
\name{idct_1d}
\alias{idct_1d}
\title{Compute 1D Inverse Discrete Cosine Transform}
\usage{
idct_1d(X)
}
\arguments{
\item{X}{A vector of values for which the discrete cosine transform is being computed.}
}
\value{
A vector with the inverse transform values.
}
\description{
This function computes the inverse of the Type-II discrete cosine transform.
}
\details{
The Type-II inverse discrete cosine transform is computed using \link[stats:fft]{stats::fft}.
For autocovariance function estimation, the spectrum is given in the input \code{X} and then an inverse FFT is applied.

The original spectrum, \code{dct_full}, from \code{X} is obtained as follows.
First, the original sample Fourier spectrum is reconstructed using \code{dct_full <- c(X, 0, -X[-1], 0, rev(X[-1]))}. After this, an inverse FFT is applied,
\code{idct <- Re(stats::fft(dct_full, inverse = TRUE)) * (2 / length(dct_full))}, which gives the original function with additional zero-values at even indices.
The zeroes are dropped, which gives the untransformed X.
}
\examples{
X <- dct_1d(c(1, 2, 3))
idct_1d(X)
}
\references{
Ochoa-Dominguez, H. & Rao, K.R. (2019). Discrete Cosine Transform, Second Edition. CRC Press. https://doi.org/10.1201/9780203729854

endolith (2013). Fast Cosine Transform via FFT. Signal Processing Stack Exchange.
https://dsp.stackexchange.com/a/10606
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{hilbert_schmidt}
\alias{hilbert_schmidt}
\alias{hilbert_schmidt.CovEsts}
\alias{hilbert_schmidt.default}
\title{Hilbert-Schmidt Norm Between Estimated Autocovariance Functions.}
\usage{
hilbert_schmidt(est1, est2)

\method{hilbert_schmidt}{CovEsts}(est1, est2)

\method{hilbert_schmidt}{default}(est1, est2)
}
\arguments{
\item{est1}{A numeric vector representing the first estimated autocovariance function.}

\item{est2}{A numeric vector of the same length as \code{est1} representing the second estimated autocovariance function}
}
\value{
A numeric value representing the estimated Hilbert-Schmidt norm between two estimated autocovariance functions.
}
\description{
This function computes the Hilbert-Schidmt norm between two estimated autocovariance functions.
}
\details{
This function computes the Hilbert-Schidmt norm between two estimated autocovariance functions.
The Hilbert-Schmidt norm of a matrix
\deqn{D = \left[(d_{i,j})_{1 \le i,j \le n}\right] = \left[ {\begin{array}{ccccc}
D(h_{0})     & D(h_{1})     & \cdots & D(h_{n - 1}) & D(h_{n})     \\
D(h_{1})     & D(h_{0})     & \cdots & D(h_{n - 2}) & D(h_{n - 1}) \\
\vdots       & \vdots       & \ddots & \vdots       & \vdots       \\
D(h_{n - 1}) & D(h_{n - 2}) & \cdots & D(h_{0})     & D(h_{1})     \\
D(h_{n})     & D(h_{n - 1}) & \cdots & D(h_{1})     & D(h_{0})     \\
\end{array}} \right] ,
}
over a set of lags \eqn{\{h_{0}, h_{1}, \dots , h_{N} \},} where \eqn{D(h) = \hat{C}_{1}(h) - \hat{C}_{2}(h),}
is defined as \deqn{{\left\Vert D \right\Vert}_{HS} = \sqrt{\sum_{i,j} d_{i, j}^{2}}.}
}
\section{Methods (by class)}{
\itemize{
\item \code{hilbert_schmidt(CovEsts)}: Method for \code{CovEsts} objects.

\item \code{hilbert_schmidt(default)}: Method for numeric vectors.

}}
\examples{
x <- seq(0, 5, by=0.1)
estCov1 <- exp(-x^2)
estCov2 <- exp(-x^2.1)
hilbert_schmidt(estCov1, estCov2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/B_splines_est.R
\name{adjusted_spline}
\alias{adjusted_spline}
\title{Compute Adjusted Splines.}
\usage{
adjusted_spline(x, j, l, p, m, taus)
}
\arguments{
\item{x}{Argument of the function.}

\item{j}{Index of basis function of order \eqn{l}.}

\item{l}{Order of function.}

\item{p}{The order of the splines.}

\item{m}{The number of nonboundary knots.}

\item{taus}{Vector of \eqn{\tau}s, see \link{get_taus}.}
}
\value{
A numeric value of the adjusted spline \eqn{f_{j}^{(l)}(x).}
}
\description{
A helper function that is an implementation of the formula from Choi, Li & Wang (2013, p. 616),
\deqn{
f_{j}^{(l)}(x) = \frac{m + 1}{l} \left( f_{j}^{(l - 1)}(x + 1) - \tau_{j - p} f_{j}^{(l - 1)}(x) + \tau_{j - p + l + 1} f_{j + 1}^{(l - 1)}(x) -  f_{j + 1}^{(l - 1)}(x + 1)  \right) ,
}
where \eqn{m} is the number of nonboundary knots, \eqn{p} is the order of the spline, \eqn{l} is the order of the adjusted spline (the function \eqn{f_{j}^{(l)}(\cdot)}) and \eqn{j = 1, 2, \dots , m + p.}
}
\examples{
\dontrun{
taus <- get_taus(3, 2)
adjusted_spline(1, 2, 1, 3, 2, taus)
}
}
\references{
Choi, I., Li, B. & Wang, X. (2013). Nonparametric Estimation of Spatial and Space-Time Covariance Function. JABES 18, 611-630. https://doi.org/10.1007/s13253-013-0152-z
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernel_regression_estimator.R
\name{Xij_mat}
\alias{Xij_mat}
\title{Compute \eqn{X_{ij}} Matrix}
\usage{
Xij_mat(X, meanX = mean(X))
}
\arguments{
\item{X}{A vector of values.}

\item{meanX}{The average value of \code{X}. Defaults to \code{mean(X)}.}
}
\value{
A matrix of size \eqn{N \times N}, where \eqn{N} is the length of the vector X.
}
\description{
This helper function computes the matrix of pairwise values, \eqn{X_{ij},} for the kernel regression estimator,
\deqn{X_{ij} = (X_{i} - \bar{X}) (X_{j} - \bar{X}) .}
}
\examples{
\dontrun{
X <- c(1, 2, 3, 4)
Xij_mat(X, mean(X))
}
}
\references{
Hall, P. & Patil, P. (1994). Properties of nonparametric estimators of autocovariance for stationary random fields. Probability Theory and Related Fields 99(3), 399-424. https://doi.org/10.1007/bf01199899

Hall, P., Fisher, N. I., & Hoffmann, B. (1994). On the nonparametric estimation of covariance functions. The Annals of Statistics 22(4), 2115-2134. https://doi.org/10.1214/aos/1176325774
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{mse}
\alias{mse}
\alias{mse.CovEsts}
\alias{mse.default}
\title{MSE Between Estimated Autocovariance Functions.}
\usage{
mse(est1, est2)

\method{mse}{CovEsts}(est1, est2)

\method{mse}{default}(est1, est2)
}
\arguments{
\item{est1}{A numeric vector representing the first estimated autocovariance function.}

\item{est2}{A numeric vector of the same length as \code{est1} representing the second estimated (or theoretical) autocovariance function}
}
\value{
A numeric value representing the MSE between two autocovariance functions (estimated or theoretical).
}
\description{
This function computes the mean-square difference/error between two autocovariance functions (estimated or theoretical).
}
\details{
This function computes the mean-square difference/error (MSE) between two estimated autocovariance functions (estimated or theoretical).
The MSE is defined as
\deqn{\frac{1}{n} \sum_{i=0}^{n} \left(\widehat{C}_{1}(h_{i}) - \widehat{C}_{2}(h_{i})\right)^{2}}
over a set of lags \eqn{\{h_{0}, h_{1} , h_{2} , \dots , h_{n} \}.}
}
\section{Methods (by class)}{
\itemize{
\item \code{mse(CovEsts)}: Method for \code{CovEsts} object.

\item \code{mse(default)}: Method for numeric vectors.

}}
\examples{
x <- seq(0, 5, by=0.1)
estCov1 <- exp(-x^2)
estCov2 <- exp(-x^2.1)
mse(estCov1, estCov2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/B_splines_est.R
\name{get_taus}
\alias{get_taus}
\title{Get all \eqn{\tau}.}
\usage{
get_taus(p, m)
}
\arguments{
\item{p}{The order of the splines.}

\item{m}{The number of nonboundary knots.}
}
\value{
A numeric vector of all \eqn{\tau_{i}, i = -p, \dots, m + p + 1.}
}
\description{
A helper function to obtain all \eqn{\tau_{i}, i=-p, \dots, m + p + 1,} where each \eqn{\tau_{i}} is as follows.
For \eqn{i = -p , -p + 1, \dots , -2, -1 , m + 2, m + 3, \dots , m + p , m + p + 1,} it is equal to \eqn{\tau_{i} = i / (m + 1)}, and for \eqn{i = 0, \dots , m + 1,} it is \eqn{\tau_{i} = \kappa_{i}.}
See Choi, Li & Wang (2013, p. 615) for details.
}
\examples{
\dontrun{
get_taus(3, 2)
}
}
\references{
Choi, I., Li, B. & Wang, X. (2013). Nonparametric Estimation of Spatial and Space-Time Covariance Function. JABES 18, 611-630. https://doi.org/10.1007/s13253-013-0152-z
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/standard_est.R
\name{normalise_acf}
\alias{normalise_acf}
\alias{normalise_acf.default}
\alias{normalise_acf.CovEsts}
\title{Normalise a CovEsts Object}
\usage{
normalise_acf(estCov)

\method{normalise_acf}{default}(estCov)

\method{normalise_acf}{CovEsts}(estCov)
}
\arguments{
\item{estCov}{A CovEst S3 object.}
}
\value{
A CovEsts S3 object with a normalised \code{acf} attribute.
}
\description{
This function normalises a CovEsts S3 object, that is, turning an autocovariance function into an autocorrelation function.
This procedure is a one-way transformation.
This function does not support any other type of argument.
}
\section{Methods (by class)}{
\itemize{
\item \code{normalise_acf(default)}: Method for other objects.

\item \code{normalise_acf(CovEsts)}: Method for 'CovEsts' objects.

}}
\examples{
normalise_acf(standard_est(c(1, 2, 3)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block_bootstrap.R
\name{bootstrap_sample}
\alias{bootstrap_sample}
\title{Block Bootstrap Sample}
\usage{
bootstrap_sample(X, l, k, boot_type = c("moving", "circular"))
}
\arguments{
\item{X}{A vector representing observed values of the time series.}

\item{l}{The block length considered for bootstrap.}

\item{k}{The number of blocks considered for bootstrap.}

\item{boot_type}{What type of block bootstrap should be used, either 'moving' for moving block bootstrap or 'circular' for circular block bootstrap.}
}
\value{
A vector of length \code{length(X)} whose values are a bootstrapped time series.
}
\description{
This function generates block bootstrap samples for either moving block bootstrap or circular bootstrap.
}
\details{
This function generates a block bootstrap sample for a time series \eqn{X.}
For the moving block bootstrap and circular bootstrap procedures see \link{block_bootstrap} and the included references.
}
\examples{
X <- c(1, 2, 3, 3, 2, 1)
bootstrap_sample(X, 2, 3)
}
\references{
Chapters 2.5 and 2.7 in Lahiri, S. N. (2003). Resampling Methods for Dependent Data. Springer. https://doi.org/10.1007/978-1-4757-3803-2

Künsch, H. R. (1989). The Jackknife and the Bootstrap for General Stationary Observations. The Annals of Statistics 17(3), 1217-1241. https://doi.org/10.1214/aos/1176347265

Politis, D. N. & Romano, J. P. (1991). A Circular Block-Resampling Procedure for Stationary Data. In R. LePage & L. Billard, eds, Exploring the Limits of Bootstrap, Wiley, 263-270.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/B_splines_est.R
\name{solve_spline}
\alias{solve_spline}
\title{Objective Function for WLS.}
\usage{
solve_spline(par, splines_df, weights)
}
\arguments{
\item{par}{A vector of initial parameters to used in the minimisation process.}

\item{splines_df}{A data frame whose structure is defined in \link{splines_df}.}

\item{weights}{A vector of weights, see the description.}
}
\value{
The value of the objective function at those parameters.
}
\description{
This is the objective function to find the weights for each basis function in the minimising spline, see Choi, Li & Wang (2013, p. 617).
The parameters must be nonnegative, so a penalty of \eqn{10^{12}} is given if any parameters are negative.
The weights are chosen as per Choi, Li & Wang (2013, p. 617).
}
\details{
Let \eqn{\mathbf{\beta} = (\beta_{0}, \dots, \beta_{m + p})^{\prime}} be a vector of model coefficients, \eqn{\{f_{1}^{(p - 1)} , \dots , f_{m + p}^{(p - 1)} \}}
be a set of completely monotone basis functions, and \eqn{\widehat{C}(\cdot)} be an estimated covariance function.
As per Choi, Li & Wang (2013, p. 617), \eqn{\mathbf{\beta}} can be estimated via weighted-least squares,
\deqn{\
\hat{\mathbf{\beta}}_{WLS} = {\arg\min}_{\beta_{j} \ge 0} \sum_{i=1}^{L} w_{i} \left(\widehat{C}(h_{i}) - \sum_{j = 1}^{m + p} \beta_{j} f_{j}^{(p - 1)}(h_{i}^{2})  \right)^{2} ,
}
where \eqn{\{h_{1} , \dots , h_{L} \}} is a set of lags and \eqn{\{w_{1}, \dots , w_{L} \}} is a set of weights.
The set of weights is calculated in \link{splines_est}, and they are of the form \eqn{w_{i} = (N - h_{i}) / ((1 - \widehat{C}(h_{i}))^{2}).}
}
\examples{
\dontrun{
taus <- get_taus(3, 2)
x <- seq(0, 2, by=0.25)
maxLag <- 4
splines_df <- splines_df(x[1:maxLag], 3, 2, taus)
splines_df$estCov <- exp(-splines_df$lags^2) + 0.001
# pars are the initial parameters used in the minimisation process.
pars <- c(0.5, 0.5, 0.5, 0.5, 0.5)
weights <- c()
X <- rnorm(50)
for(i in 0:(maxLag - 1)) {
  weights <- c(weights, (length(X) - i) / ( (1 - splines_df$estCov[i + 1])^2 ))
}
solve_spline(pars, splines_df, weights)
}
}
\references{
Choi, I., Li, B. & Wang, X. (2013). Nonparametric Estimation of Spatial and Space-Time Covariance Function. JABES 18, 611-630. https://doi.org/10.1007/s13253-013-0152-z
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{cyclic_matrix}
\alias{cyclic_matrix}
\title{Create a Cyclic Matrix for a Given Vector.}
\usage{
cyclic_matrix(v)
}
\arguments{
\item{v}{A numeric vector.}
}
\value{
A symmetric matrix.
}
\description{
This helper function creates a symmetric matrix from a given vector \eqn{v}.
}
\details{
This function creates a symmetric matrix for a given vector \eqn{v}.
If \eqn{v = \{v_{0}, v_{1} , \dots , v_{N-1} , v_{N} \},} then the symmetric matrix will has the form
\deqn{\left[ {\begin{array}{ccccc}
v_{0}     & v_{1}     & \cdots & v_{N - 1} & v_{N}     \\
v_{1}     & v_{0}     & \cdots & v_{N - 2} & v_{N - 1} \\
\vdots    & \vdots    & \ddots & \vdots    & \vdots    \\
v_{N - 1} & v_{N- 2}  & \cdots & v_{0}     & v_{1}     \\
v_{N}     & v_{N - 1} & \cdots & v_{1}     & v_{0}     \\
\end{array}} \right]
}
}
\examples{
\dontrun{
v <- c(1, 2, 3)
cyclic_matrix(v)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/corrected_standard_estimator.R
\name{kernel_est}
\alias{kernel_est}
\alias{kernel_est.CovEsts}
\alias{kernel_est.default}
\title{Kernel Correction for an Estimated Autocovariance Function.}
\usage{
kernel_est(
  estCov,
  kernel_name = c("gaussian", "exponential", "wave", "rational_quadratic", "spherical",
    "circular", "bessel_j", "matern", "cauchy"),
  kernel_params = c(),
  N_T = 0.1 * length(estCov),
  maxLag = length(estCov) - 1,
  x = 0:length(estCov),
  type = c("autocovariance", "autocorrelation"),
  custom_kernel = FALSE
)

\method{kernel_est}{CovEsts}(
  estCov,
  kernel_name = c("gaussian", "exponential", "wave", "rational_quadratic", "spherical",
    "circular", "bessel_j", "matern", "cauchy"),
  kernel_params = c(),
  N_T = 0.1 * length(estCov$acf),
  maxLag = length(estCov$acf) - 1,
  x = estCov$lags,
  type = c("autocovariance", "autocorrelation"),
  custom_kernel = FALSE
)

\method{kernel_est}{default}(
  estCov,
  kernel_name = c("gaussian", "exponential", "wave", "rational_quadratic", "spherical",
    "circular", "bessel_j", "matern", "cauchy"),
  kernel_params = c(),
  N_T = 0.1 * length(estCov),
  maxLag = length(estCov) - 1,
  x = 0:length(estCov),
  type = c("autocovariance", "autocorrelation"),
  custom_kernel = FALSE
)
}
\arguments{
\item{estCov}{A vector whose values are an estimate autocovariance function.}

\item{kernel_name}{The name of the \link{kernel_ec} function to be used. Possible values are:
gaussian, exponential, wave, rational_quadratic, spherical, circular, bessel_j, matern, cauchy.}

\item{kernel_params}{A vector of parameters of the kernel function. See \link{kernel_ec} for parameters.
In the case of gaussian, wave, rational_quadratic, spherical and circular, \code{N_T} takes the place of \eqn{\theta}.
For kernels that require parameters other than \eqn{\theta}, such as the Matern kernel, those parameters are passed.}

\item{N_T}{The range at which the kernel function vanishes at. Recommended to be \eqn{0.1 N} when considering all lags. This parameter may be large for a lag small estimation lag.}

\item{maxLag}{An optional parameter that determines the maximum lag to compute the estimated autocovariance function at. Defaults to \code{length(estCov) - 1}.}

\item{x}{A vector of lag indices. Defaults to the sequence \code{0:length(X)}. Must be at least as large as \code{maxLag + 1}.}

\item{type}{Compute either the 'autocovariance' or 'autocorrelation'. Defaults to 'autocovariance'.}

\item{custom_kernel}{If a custom kernel is to be used or not. Defaults to \code{FALSE}. See the examples of \link{corrected_est} for usage.}
}
\value{
A vector whose values are the kernel corrected autocovariance estimates or \code{CovEsts} S3 object (list) with the following values
\describe{
\item{\code{acf}}{A numeric vector containing the autocovariance/autocorrelation estimates.}
\item{\code{lags}}{A numeric vector containing the lag indices used to compute the estimates on, inherited from the argument \code{estCov}.}
\item{\code{est_type}}{The type of estimate, namely 'autocorrelation' or 'autocovariance', this depends on the argument \code{type}.}
\item{\code{est_used}}{The estimator function used, in this case, 'kernel_est'.}
}
If a numeric vector is given for the argument \code{estCov}, then a numeric vector output is given, and if a \code{CovEsts} S3 object is given, a \code{CovEsts} object is given as output.
}
\description{
This function applies kernel correction to an estimated autocovariance function,
\deqn{\widehat{C}_{T}^{(a)}(h) = \widehat{C}(h) a_{T}(h),} where \eqn{a_{T}(h) := a(h / N_{T}).}
It uses a kernel \eqn{a(\cdot)} which decays or vanishes to zero (depending on the type of kernel) where \eqn{a(0) = 1.}
The rate or value at which the kernel vanishes is \eqn{N_{T}}, which is recommended to be of order \eqn{0.1 N}, where \eqn{N} is the length of the observation window, however, one may need to play with this value.
}
\section{Methods (by class)}{
\itemize{
\item \code{kernel_est(CovEsts)}: Method for \code{CovEsts} objects.

\item \code{kernel_est(default)}: Method for numeric vectors.

}}
\examples{
X <- rnorm(1000)
Y <- c(X[1], X[2])
for(i in 3:length(X)) { Y[i] <- X[i] - 0.3*X[i - 1] - 0.6*X[i - 2] }
cov_est <- standard_est(Y)
plot(cov_est)
plot(kernel_est(cov_est,
     "bessel_j", kernel_params=c(0, 1), N_T=0.2*length(Y)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernels.R
\name{kernel_symm_ec}
\alias{kernel_symm_ec}
\title{1D Isotropic Symmetric Kernels.}
\usage{
kernel_symm_ec(
  x,
  name = c("gaussian", "wave", "rational_quadratic", "bessel_j"),
  params = c(1)
)
}
\arguments{
\item{x}{A vector or matrix of arguments of at least length 1 for which the kernel is computed at. Each value can be negative as well as positive.}

\item{name}{The name of the kernel. Options are: gaussian, wave, rational_quadratic, bessel_j.}

\item{params}{A vector of parameters for the kernel. See the documentation below for the position of the parameters. All kernels will have a scale parameter as the first value in the vector.}
}
\value{
A vector or matrix of values.
}
\description{
These functions computes values of kernels that have the properties of symmetric probability distributions. Note that \code{kernel_symm()} is \strong{deprecated}, please use \code{kernel_symm_ec()} instead.
For a kernel \eqn{a(x)}, the standardised version is \eqn{a(x) / \int_{-\infty}^{\infty} a(x) dx}, so that the integral is 1.
The symmetric kernels are different to \link{kernel} and are used in the functions \link{adjusted_est} and \link{truncated_est}.
}
\details{
\strong{Symmetric Gaussian Kernel}.
The symmetric Gaussian kernel is defined as
\deqn{a(x;\theta) = \sqrt{\pi \theta} \exp(-x^{2} / \theta) , \theta > 0.}
The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}.

\strong{Symmetric Wave Kernel}.
The wave (cardinal sine) kernel is given by
\deqn{a(x;\theta) = \left\{ \begin{array}{ll}
(\sqrt{\theta^{2}} \pi)^{-1} \frac{\theta}{x} \sin\left( \frac{x}{\theta} \right), & x \neq 0 \\
1, & x = 0
\end{array} \right.}
where \eqn{\theta > 0.}
The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}

\strong{Symmetric Rational Quadratic Kernel}.
The symmetric rational quadratic kernel is given by
\deqn{a(x;\theta) = (\pi \sqrt{\theta})^{-1} \left(\frac{\theta}{x^{2} + \theta}\right), \theta > 0.}
The \code{params} argument is of the form \code{c(}\eqn{\theta}\code{)}

\strong{Symmetric Besesel Kernel}.
The symmetric Bessel kernel, which is valid when \eqn{\nu \geq \frac{d}{2} - 1,} is given by
\deqn{a(x; \theta, \nu) = \left(\Gamma\left(\frac{1}{2} + \nu\right)/(2 \sqrt{\pi} \theta \Gamma(1 + \nu))\right) ( 2^{\nu} \Gamma(\nu + 1) J_{\nu}(x / \theta) (x / \theta)^{-\nu}), \,\theta > 0, \nu \geq \frac{d}{2} - 1,}
where \eqn{J_{\nu}(\cdot)} is the Bessel function of the first kind and \eqn{d} is the dimension.
The \code{params} argument is of the form \code{c(}\eqn{\theta, \nu, d}\code{)}.
}
\examples{
x <- c(-2, -1, 0, 1, 2)
theta <- 1
kernel_symm_ec(x, "gaussian", c(theta))
kernel_symm_ec(x, "wave", c(theta))
kernel_symm_ec(x, "rational_quadratic", c(theta))
dim <- 1
nu <- 1
kernel_symm_ec(x, "bessel_j", c(theta, nu, dim))
curve(kernel_symm_ec(x, "gaussian", c(theta)), from = -5, to = 5)
curve(kernel_symm_ec(x, "wave", c(theta)), from = -5, to = 5)
curve(kernel_symm_ec(x, "rational_quadratic", c(theta)), from = -5, to = 5)
curve(kernel_symm_ec(x, "bessel_j", c(theta, nu, dim)), from = -5, to = 5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{CovEsts-deprecated}
\alias{CovEsts-deprecated}
\alias{kernel}
\alias{kernel_symm}
\alias{window}
\alias{window_symm}
\title{Deprecated functions in package \pkg{CovEsts}.}
\usage{
kernel(x, name, params = c(1))

kernel_symm(x, name, params = c(1))

window(x, name, params = c(1))

window_symm(x, name, params = c(1))
}
\arguments{
\item{x, name, params}{Parameters passed to the updated functions.}
}
\description{
These functions are provided for compatibility with older verisons of the \pkg{CovEsts} package.
These functions may be removed entirely.
}
\section{\code{kernel}}{

For \code{kernel()}, use \code{\link{kernel_ec}} instead.
}

\section{\code{kernel_symm}}{

For \code{kernel_symm()}, use \code{\link{kernel_symm_ec}} instead.
}

\section{\code{window}}{

For \code{window()}, use \code{\link{window_ec}} instead.
}

\section{\code{window_symm}}{

For \code{window_symm()}, use \code{\link{window_symm_ec}} instead.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{area_between}
\alias{area_between}
\alias{area_between.CovEsts}
\alias{area_between.default}
\title{Area Between Estimated Autocovariance Functions.}
\usage{
area_between(est1, est2, lags = c(), plot = FALSE)

\method{area_between}{CovEsts}(est1, est2, lags = c(), plot = FALSE)

\method{area_between}{default}(est1, est2, lags = c(), plot = FALSE)
}
\arguments{
\item{est1}{A numeric vector representing the first estimated autocovariance function.}

\item{est2}{A numeric vector of the same length as \code{est1} representing the second estimated autocovariance function}

\item{lags}{An optional vector of lags starting from 0 up until some other lag. If empty, a vector of lags is created starting from 0 until \code{len(est1) - 1}, by 1.}

\item{plot}{A boolean determining whether a plot should be created. By default, no plot is created.}
}
\value{
A numeric value representing the estimated area between two estimated autocovariance functions.
}
\description{
This function estimates the area between two estimated autocovariance functions.
}
\details{
This function estimates the area between two estimated autocovariance functions over a set of lags, from 0 up to \eqn{h_{n}} defined by
\deqn{
\int_{0}^{h_{n}} \left| \hat{C}_{1}(h) - \hat{C}_{2}(h) \right| dh ,
}
where \eqn{\hat{C}_{1}(\cdot)} and \eqn{\hat{C}_{2}(\cdot)} are estimated autocovariance functions.

To approximate this integral the trapezoidal rule is used.

If \code{lags} is empty, a uniform time grid with a step of 1 will be used which may result in a different area than if lags is specified.
}
\section{Methods (by class)}{
\itemize{
\item \code{area_between(CovEsts)}: Method for \code{CovEsts} objects.

\item \code{area_between(default)}: Method for numeric vectors.

}}
\examples{
x <- seq(0, 5, by=0.1)
estCov1 <- exp(-x^2)
estCov2 <- exp(-x^2.1)
area_between(estCov1, estCov2, lags=x)
area_between(estCov1, estCov2, lags=x, plot = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernel_regression_estimator.R
\name{dct_1d}
\alias{dct_1d}
\title{Compute 1D Discrete Cosine Transform}
\usage{
dct_1d(X)
}
\arguments{
\item{X}{A vector of values for which the discrete cosine transform is being computed.}
}
\value{
A vector of discrete cosine transform values.
}
\description{
This function computes the Type-II discrete cosine transform.
}
\details{
The Type-II discrete cosine transform is obtained using \link[stats:fft]{stats::fft}.
If \eqn{X} is of length N, construct a new signal \eqn{Y} of length \eqn{4N}, with values \eqn{Y_{2n} = 0, Y_{2n + 1} = x_{n}} for \eqn{0 \le n < N},
and \eqn{Y_{2N} = 0, Y_{4N - n} = y_{n}} for \eqn{0 < n < 2N.}
After this, the Type-II discrete cosine transform is computed by \code{0.5 * Re(stats::fft(Y))[1:(length(Y) / 4)]}.
}
\examples{
X <- c(1, 2, 3)
dct_1d(X)
}
\references{
Ochoa-Dominguez, H. & Rao, K.R. (2019). Discrete Cosine Transform, Second Edition. CRC Press. https://doi.org/10.1201/9780203729854

Makhoul, J. (1980). A Fast Cosine Transform in One and Two Dimensions. IEEE Transactions on Acoustics, Speech, and Signal Processing 28(1), 27-34. https://doi.org/10.1109/TASSP.1980.1163351

Stasiński, R. (2002). DCT Computation Using Real-Valued DFT Algorithms. Proceedings of the 11th European Signal Processing Conference.
}
